#!/bin/bash
SCRIPT_NAME=$(basename $0)
if [ $# -ne 1 ]; then
	echo sample call: $SCRIPT_NAME subsystem_dir
	exit 1
fi
SUBSYSDIR=$1
if [ ! -d "$SUBSYSDIR" ]; then
	echo Directory $SUBSYSDIR does not exist
	exit 1
fi

MAIN_LOG_FILE=$SUBSYSDIR/$SCRIPT_NAME.log
if [ -f "$MAIN_LOG_FILE" ]; then
	rm $MAIN_LOG_FILE
fi

TIME_FORMAT="%F %T"
function message {
	if [ $# != 1 ]; then
		echo sample call: message message_to_send
		exit 1
	fi
	tmp=$(date +"$TIME_FORMAT")
	echo $tmp $1
	echo $tmp $1 >> $MAIN_LOG_FILE
}

if [ ! -f "$SUBSYSDIR/bk.conf" ]; then
	message "Configuration file $SUBSYSDIR/bk.conf does not exist"
	exit 1
fi
. $SUBSYSDIR/bk.conf
if [ -z "$SUBSYS" ]; then
	message "configuration parameter SUBSYS is not specified"
	exit 1
fi
if [ -z "$SOURCES" ]; then
	message "configuration parameter SOURCES is not specified"
	exit 1
fi
if [ -z "$BANK_1" ]; then
	message "configuration parameter BANK_1 is not specified"
	exit 1
fi
if [ -z "$BANK_2" ]; then
	message "configuration parameter BANK_2 is not specified"
	exit 1
fi
if [ -z "$NUM" ]; then
	message "configuration parameter NUM is not specified"
	exit 1
fi
SUCCESS_FILE=$SUBSYSDIR/SUCCESS
FAILURE_FILE=$SUBSYSDIR/FAILURE
NOHUP_OUT=$SUBSYSDIR/nohup.out

#
# Check the user whod launched this backup
#
if [ $(whoami) != "root" ]; then
	message "only user root can run this script"
	exit 1
fi

if [ ! -d "$BANK_1" ]; then
	message "Bank $BANK_1 does not exist"
	exit 1
fi
if [ ! -d "$BANK_2" ]; then
	message "Bank $BANK_2 does not exist"
	exit 1
fi
DIR_1=$BANK_1/$SUBSYS
DIR_2=$BANK_2/$SUBSYS
if [ ! -d "$DIR_1" ]; then
	mkdir $DIR_1	
fi
if [ ! -d "$DIR_2" ]; then
	mkdir $DIR_2	
fi
#
# Check for backup of this subsystem is in progress
#
sleep 2
if [ $(ps -ef | grep rsync | grep $DIR_1 | wc -w) -gt 0 ]; then
	message "backup of $SUBSYS-subsystem is already running into bank $DIR_1"
	exit 1
fi
if [ $(ps -ef | grep rsync | grep $DIR_2 | wc -w) -gt 0 ]; then
	message "backup of $SUBSYS-subsystem is already running into bank $DIR_2"
	exit 1
fi

#
# Check if script run in background
#
if [ -z "$_BG_" ]; then
	if [ -f "$NOHUP_OUT" ]; then
		rm $NOHUP_OUT
	fi
	export _BG_=BG
	nohup $0 $@ >$NOHUP_OUT &
	exit 0
fi

if [ -f "$SUCCESS_FILE" ]; then
	rm $SUCCESS_FILE
fi
if [ -f "$FAILURE_FILE" ]; then
	rm $FAILURE_FILE
fi

#
# Choose which bank (DIR_1 or DIR_2) we have to use now for backup
#
DATE_1=""
DATE_2=""
BACKUP_1_FAILED=""
BACKUP_2_FAILED=""
BACKUP_FAILED=""
CHOOSE=""
DATE=""
DIR=""
OTHER_DIR_DATE=""
OTHER_DIR=""
if [ -f $DIR_1/LAST_BACKUP_SUCCEEDED ]; then
	DATE_1="$(head -c19 $DIR_1/LAST_BACKUP_SUCCEEDED)"
fi 
if [ -f $DIR_1/LAST_BACKUP_FAILED ]; then
	DATE_1="$(head -c19 $DIR_1/LAST_BACKUP_FAILED)"
	BACKUP_1_FAILED=YES
fi 
if [ -f $DIR_2/LAST_BACKUP_SUCCEEDED ]; then
	DATE_2="$(head -c19 $DIR_2/LAST_BACKUP_SUCCEEDED)"
fi 
if [ -f $DIR_2/LAST_BACKUP_FAILED ]; then
	DATE_2="$(head -c19 $DIR_2/LAST_BACKUP_FAILED)"
	BACKUP_2_FAILED=YES
fi 

if [ -z "$DATE_1" ] || [ "$DATE_1" \< "$DATE_2" ]; then
	if [ "$BACKUP_1_FAILED" == "$BACKUP_2_FAILED" ] || [ -z "$BACKUP_2_FAILED" ]; then
		CHOOSE=1
	else
		CHOOSE=2
	fi
else
	if [ "$BACKUP_1_FAILED" == "$BACKUP_2_FAILED" ] || [ -z "$BACKUP_1_FAILED" ]; then
		CHOOSE=2
	else
		CHOOSE=1
	fi
fi
if [ "$CHOOSE" == "1" ]; then
	DATE="$DATE_1"
	DIR="$DIR_1"
	OTHER_DIR_DATE="$DATE_2"
	OTHER_DIR="$DIR_2"
	if [ ! -z "$BACKUP_1_FAILED" ]; then
		BACKUP_FAILED=YES
	fi
else
	DATE="$DATE_2"
	DIR="$DIR_2"
	OTHER_DIR_DATE="$DATE_1"
	OTHER_DIR="$DIR_1"
	if [ ! -z "$BACKUP_2_FAILED" ]; then
		BACKUP_FAILED=YES
	fi
fi

if [ -n "$DATE" ]; then
	message "Start $SUBSYS-subsystem backup into bank $DIR. Previous backup in this bank has been done on $DATE"
else
	message "Start $SUBSYS-subsystem backup into bank $DIR"
fi
if [ -n "$OTHER_DIR_DATE" ]; then
	message "Backup on alternate bank $OTHER_DIR has been done on $OTHER_DIR_DATE"
fi

#
# Now, work with chosen bank, DIR
#
CONTROL=$DIR/control
if [ -f $CONTROL ]; then
	SEQ=$(head -n1 $CONTROL)
else
	SEQ=1
fi
SUCCESS=$DIR/LAST_BACKUP_SUCCEEDED
FAILURE=$DIR/LAST_BACKUP_FAILED
LOG_FILE=$DIR/$SEQ.log
FULL_LOG_FILE=$DIR/$SEQ.fulllog
START=$(date +"$TIME_FORMAT")
START_SEC=$(date +%s)
echo "$START starting backup #$SEQ" > $LOG_FILE
echo "SEQ = $SEQ"
if [ $SEQ -gt 1 ]; then
	PREV_SEQ=$((SEQ-1))
else
	PREV_SEQ=$NUM
fi
if [ ! -d "$DIR/$SUBSYS$PREV_SEQ" ]; then
	mkdir $DIR/$SUBSYS$PREV_SEQ
fi
if [ -f $FULL_LOG_FILE ]; then
	rm $FULL_LOG_FILE
fi
if [ -f $SUCCESS ]; then
	rm $SUCCESS
fi
if [ -f $FAILURE ]; then
	rm $FAILURE
fi
if [ ! -z "$BACKUP_FAILED" ]; then
	echo "$(date +"$TIME_FORMAT") deleting $DIR/$SUBSYS$SEQ because of previous backup failure..." >> $LOG_FILE 
	rm -r $DIR/$SUBSYS$SEQ
fi

for source in $SOURCES
do
	echo "$(date +"$TIME_FORMAT") running rsync for $source ..." >> $LOG_FILE
	rsync -av --delete-after --force --link-dest=$DIR/$SUBSYS$PREV_SEQ --log-file=$FULL_LOG_FILE $source $DIR/$SUBSYS$SEQ
	EXIT_CODE=$?
	if [ $EXIT_CODE -ne 0 ] && [ $EXIT_CODE -ne 24 ]; then
		break
	fi
done
END=$(date +"$TIME_FORMAT")
END_SEC=$(date +%s)
if [ $EXIT_CODE -ne 0 ] && [ $EXIT_CODE -ne 24 ]; then
	tmp="$END backup #$SEQ abnormally terminated with exit code $EXIT_CODE"
	echo $tmp
	echo $tmp >> $MAIN_LOG_FILE
	echo $tmp >> $LOG_FILE
	touch $FAILURE_FILE
	echo "last 5 lines from $FULL_LOG_FILE are below" >> $LOG_FILE
	tail -n5 $FULL_LOG_FILE >> $LOG_FILE
	echo -e "$END backup #$SEQ abnormally terminated with exit code $EXIT_CODE\nSee files $LOG_FILE and $FULL_LOG_FILE for details" > $FAILURE
	exit $EXIT_CODE
fi
NEXT_SEQ=$((SEQ+1))
if [ $NEXT_SEQ -gt $NUM ]; then
	NEXT_SEQ=1
fi
if [ -f $CONTROL ]; then
	rm $CONTROL
fi
echo $NEXT_SEQ > $CONTROL
tmp="$END backup #$SEQ successfully finished"
echo $tmp
echo $tmp >> $MAIN_LOG_FILE
echo $tmp >> $LOG_FILE
touch $SUCCESS_FILE
echo "$END elapsed time $(($END_SEC-$START_SEC)) seconds" >> $LOG_FILE
echo "last 5 lines from $FULL_LOG_FILE are below" >> $LOG_FILE
tail -n5 $FULL_LOG_FILE >> $LOG_FILE
echo -e "$END backup #$SEQ successfully finished\nSee files $LOG_FILE and $FULL_LOG_FILE for details" > $SUCCESS
exit 0
