#!/bin/bash
# 
# Makes gbak-backup of firebird database, so that backup target storage is alternated between two targets locations/banks
#

SCRIPT_NAME=$(basename $0)
if [ $# -ne 1 ]; then
	echo sample call: $SCRIPT_NAME subsystem_dir
	exit 1
fi
CONFIG_DIR=$1
if [ ! -d "$CONFIG_DIR" ]; then
	echo Directory $CONFIG_DIR does not exist
	exit 1
fi

MAIN_LOG_FILE=$CONFIG_DIR/$SCRIPT_NAME.log
if [ -f "$MAIN_LOG_FILE" ]; then
	rm $MAIN_LOG_FILE
fi

TIME_FORMAT="%F %T"
function message {
	if [ $# != 1 ]; then
		echo sample call: message message_to_send
		exit 1
	fi
	tmp=$(date +"$TIME_FORMAT")
	echo -e "$tmp $1"
	echo -e "$tmp $1" >> $MAIN_LOG_FILE
}

if [ ! -f "$CONFIG_DIR/bk.conf" ]; then
	message "Configuration file $CONFIG_DIR/bk.conf does not exist"
	exit 1
fi

. $CONFIG_DIR/bk.conf


if [ -z "$DATABASE" ]; then
	message "configuration parameter DATABASE is not specified"
	exit 1
fi
if [ -z "$URL" ]; then
	message "configuration parameter URL is not specified"
	exit 1
fi
if [ -z "$BACKUP_FILE" ]; then
	message "configuration parameter BACKUP_FILE is not specified"
	exit 1
fi
if [ -z "$BANK_1" ]; then
	message "configuration parameter BANK_1 is not specified"
	exit 1
fi
if [ -z "$BANK_2" ]; then
	message "configuration parameter BANK_2 is not specified"
	exit 1
fi

BACKUP_HOME_DIR=/opt/backup
PFILE=$BACKUP_HOME_DIR/db/_p_

SUCCESS_FILE=$CONFIG_DIR/SUCCESS
FAILURE_FILE=$CONFIG_DIR/FAILURE
NOHUP_OUT=$CONFIG_DIR/nohup.out
#
# Check the user who launched this script
#
if [ $(whoami) != "root" ]; then
	message "only user root can run this script"
	exit 1
fi

#
# Check for both bank-containers exist
#
if [ ! -d "$BANK_1" ]; then
	message "Bank $BANK_1 does not exist"
	exit 1
fi
if [ ! -d "$BANK_2" ]; then
	message "Bank $BANK_2 does not exist"
	exit 1
fi

#
# create variables for both bank-subdirectories for this database backup
#
DIR_1=$BANK_1/$DATABASE
DIR_2=$BANK_2/$DATABASE

#
# Check for backup of this subsystem is in progress
#
sleep 2
if [ $(ps -ef | grep gbak | grep $DIR_1 | wc -w) -gt 0 ]; then
	message "backup of database $DATABASE is already running into bank $DIR_1"
	exit 1
fi
if [ $(ps -ef | grep gbak | grep $DIR_2 | wc -w) -gt 0 ]; then
	message "backup of database $DATABASE is already running into bank $DIR_2"
	exit 1
fi

#
# If bank-subdirectories for this database backup do not exist, create them
#
if [ ! -d "$DIR_1" ]; then
	mkdir $DIR_1	
fi
if [ ! -d "$DIR_2" ]; then
	mkdir $DIR_2	
fi

#
# Check if script run in background
#
if [ -z "$_BG_" ]; then
	if [ -f "$NOHUP_OUT" ]; then
		rm $NOHUP_OUT
	fi
	export _BG_=BG
	nohup $0 $@ >$NOHUP_OUT &
fi

if [ -f "$SUCCESS_FILE" ]; then
	rm $SUCCESS_FILE
fi
if [ -f "$FAILURE_FILE" ]; then
	rm $FAILURE_FILE
fi

function getFileTime {
	if [ $# != 1 ]; then
		message "Sample call: getFileName file_name"
		exit 1
	fi
	if [ -f "$1" ]; then
		local lsout=$(ls --time-style=+"%F %T" -o -g $1)
		local date=$(echo $lsout | cut -d" " -f4)
		local time=$(echo $lsout | cut -d" " -f5)
		echo "$date $time"
	else
		echo ""
	fi
}

#
# Choose which bank (DIR_1 or DIR_2) we have to use now for backup
#
DATE_1=""
DATE_2=""
BACKUP_1_FAILED=""
BACKUP_2_FAILED=""
CHOOSE=""
if [ -f $DIR_1/LAST_BACKUP_SUCCEEDED ]; then
	DATE_1=$(getFileTime $DIR_1/LAST_BACKUP_SUCCEEDED)
fi 
if [ -f $DIR_1/LAST_BACKUP_FAILED ]; then
	DATE_1=$(getFileTime $DIR_1/LAST_BACKUP_FAILED)
	BACKUP_1_FAILED=YES
fi 
if [ -f $DIR_2/LAST_BACKUP_SUCCEEDED ]; then
	DATE_2=$(getFileTime $DIR_2/LAST_BACKUP_SUCCEEDED)
fi 
if [ -f $DIR_2/LAST_BACKUP_FAILED ]; then
	DATE_2=$(getFileTime $DIR_2/LAST_BACKUP_FAILED)
	BACKUP_2_FAILED=YES
fi 

if [ -z "$DATE_1" ] || [ "$DATE_1" \< "$DATE_2" ]; then
	if [ "$BACKUP_1_FAILED" == "$BACKUP_2_FAILED" ] || [ -z "$BACKUP_2_FAILED" ]; then
		CHOOSE=1
	else
		CHOOSE=2
	fi
else
	if [ "$BACKUP_1_FAILED" == "$BACKUP_2_FAILED" ] || [ -z "$BACKUP_1_FAILED" ]; then
		CHOOSE=2
	else
		CHOOSE=1
	fi
fi

BACKUP_FAILED=""
DATE=""
DIR=""
OTHER_DIR_DATE=""
OTHER_DIR=""

if [ "$CHOOSE" == "1" ]; then
	DATE="$DATE_1"
	DIR="$DIR_1"
	OTHER_DIR_DATE="$DATE_2"
	OTHER_DIR="$DIR_2"
	if [ ! -z "$BACKUP_1_FAILED" ]; then
		BACKUP_FAILED=YES
	fi
else
	DATE="$DATE_2"
	DIR="$DIR_2"
	OTHER_DIR_DATE="$DATE_1"
	OTHER_DIR="$DIR_1"
	if [ ! -z "$BACKUP_2_FAILED" ]; then
		BACKUP_FAILED=YES
	fi
fi

if [ -n "$DATE" ]; then
	message "Start database $DATABASE backup into bank $DIR. Previous backup into this bank has been done on $DATE"
else
	message "Start database $DATABASE backup into bank $DIR"
fi
if [ -n "$OTHER_DIR_DATE" ]; then
	message "Backup into alternate bank $OTHER_DIR has been done on $OTHER_DIR_DATE"
fi

GBAK="/opt/firebird/bin/gbak -b -t -g -v "

SUCCESS=$DIR/LAST_BACKUP_SUCCEEDED
FAILURE=$DIR/LAST_BACKUP_FAILED
if [ -f "$SUCCESS" ]; then
	rm $SUCCESS
fi
if [ -f "$FAILURE" ]; then
	rm $FAILURE
fi

if [ -f $DIR/backup.log ]; then
	rm $DIR/backup.log
fi

message "/opt/firebird/bin/gbak -b -t -g -v -y $DIR/backup.log $URL $DIR/$BACKUP_FILE" 
/opt/firebird/bin/gbak -b -t -g -v  -y $DIR/backup.log -user SYSDBA -fetch $PFILE $URL $DIR/$BACKUP_FILE
EXIT_CODE=$?
END=$(date +"$TIME_FORMAT")

if [ $EXIT_CODE -ne 0 ]; then
	message "Backup of database $DATABASE into $DIR is abnormally terminated with exit code $EXIT_CODE"
	touch $FAILURE_FILE
	touch $FAILURE
	message "last 5 lines from $DIR/backup.log are below"
	tail -n5 $DIR/backup.log >> $MAIN_LOG_FILE
	echo -e "$END Backup of database $DATABASE into $DIR is abnormally terminated with exit code $EXIT_CODE\nSee file $MAIN_LOG_FILE and $DIR/backup.log for details" > $FAILURE_FILE
	exit $EXIT_CODE
fi

message "Backup of database $DATABASE into $DIR successfully finished"
touch $SUCCESS_FILE
touch $SUCCESS
message "last 5 lines from $DIR/backup.log are below"
message "$(tail -n5 $DIR/backup.log)"
echo -e "$END Backup of database $DATABASE into $DIR successfully finished\nSee file $MAIN_LOG_FILE and $DIR/backup.log for details" > $SUCCESS_FILE
