#!/bin/bash
# 
# Alternatively makes DIDB/DRIFT backup. Also, backup target storage is alternated between two targets locations/banks
#
# 2-databases(DIDB and DRIFT) 2-banks(=2-directories) backup
# Single run of this script backups one of two database, DIDB or DRIFT, alternatively, into one of
# two banks, /A or /B.
# So, after running this script 4 times, database DIDB will be backuped into subdirectory DIDB of both banks,  
# and database DRIFT will be backuped into subdirectory DRIFT of both banks.
# It is suggested, for better chances of disaster recovery to have these banks(=directories) to be mount points
# of independent devices.
# As these databases are very large and backups of its takes couple of days, it is recommended to start this
# script once per week, say each Friday night at 8PM. it can be automated using cron table by adding,
# run weekly, every Friday at 8PM. For this you can add the following entry to cron table,
# 0 20 * * 5 export _BG_=BG;/opt/backup/db/2db2bnk.sh

BACKUP_HOME_DIR=/opt/backup
SUBSYS=db
BANK_1=/A   # you might be need to adjust it
BANK_2=/B   # you might be need to adjust it
SCRIPT_NAME=$(basename $0)
PFILE=$BACKUP_HOME_DIR/$SUBSYS/_p_
MAIN_LOG_FILE=$BACKUP_HOME_DIR/$SUBSYS/$SCRIPT_NAME.log
SUCCESS_FILE=$BACKUP_HOME_DIR/$SUBSYS/SUCCESS
FAILURE_FILE=$BACKUP_HOME_DIR/$SUBSYS/FAILURE
NOHUP_OUT=$BACKUP_HOME_DIR/$SUBSYS/nohup.out
TIME_FORMAT="%F %T"

function message {
	if [ $# != 1 ]; then
		echo sample call: message message_to_send
		exit 1
	fi
	local tmp=$(date +"$TIME_FORMAT")
	echo $tmp $1
	echo $tmp $1 >> $MAIN_LOG_FILE
}

if [ -f "$MAIN_LOG_FILE" ]; then
	rm $MAIN_LOG_FILE
fi
#
# Check the user who launched this script
#
if [ $(whoami) != "root" ]; then
	message "only user root can run this script"
	exit 1
fi

#
# Check for backup of any of databases is in progress
#
sleep 2
if [ $(ps -ef | grep gbak | grep " -b " | wc -w) -gt 1 ]; then
	message "database backup is already running"
	exit 1
fi

#
# Check if script run in background
#
if [ -z "$_BG_" ]; then
	if [ -f "$NOHUP_OUT" ]; then
		rm $NOHUP_OUT
	fi
	export _BG_=BG
	nohup $0 $@ >$NOHUP_OUT &
fi

if [ -f "$SUCCESS_FILE" ]; then
	rm $SUCCESS_FILE
fi
if [ -f "$FAILURE_FILE" ]; then
	rm $FAILURE_FILE
fi

SUCCESS_FILE_NAME=LAST_BACKUP_SUCCEEDED
FAILURE_FILE_NAME=LAST_BACKUP_FAILED
if [ ! -d "$BANK_1" ]; then
	message "Bank $BANK_1 does not exist"
	exit 1
fi
if [ ! -d "$BANK_2" ]; then
	message "Bank $BANK_2 does not exist"
	exit 1
fi
DIDB_DIR_1=$BANK_1/DIDB
DIDB_DIR_2=$BANK_2/DIDB
DRIFT_DIR_1=$BANK_1/DRIFT
DRIFT_DIR_2=$BANK_2/DRIFT
if [ ! -d "$DIDB_DIR_1" ]; then
	mkdir $DIDB_DIR_1
fi
if [ ! -d "$DIDB_DIR_2" ]; then
	mkdir $DIDB_DIR_2
fi
if [ ! -d "$DRIFT_DIR_1" ]; then
	mkdir $DRIFT_DIR_1
fi
if [ ! -d "$DRIFT_DIR_2" ]; then
	mkdir $DRIFT_DIR_2
fi
DIDB_1_FAILURE=""
DIDB_2_FAILURE=""
DRIFT_1_FAILURE=""
DRIFT_2_FAILURE=""
if [ -f "$DIDB_DIR_1/$FALIURE_FILE_NAME" ]; then
	DIDB_1_FAILURE=YES
fi
if [ -f "$DIDB_DIR_2/$FALIURE_FILE_NAME" ]; then
	DIDB_2_FAILURE=YES
fi
if [ -f "$DRIFT_DIR_1/$FALIURE_FILE_NAME" ]; then
	DRIFT_1_FAILURE=YES
fi
if [ -f "$DRIFT_DIR_2/$FALIURE_FILE_NAME" ]; then
	DRIFT_2_FAILURE=YES
fi
GBAK="/opt/firebird/bin/gbak -b -t -v "
DIDB_BACKUP_FILE=d1
DRIFT_BACKUP_FILE=f1
DIDB_URL=129.63.134.212:/ext/db/ib/didb
DRIFT_URL=129.63.134.45:/ext2/db/ib/dfdb

function getFileTime {
	if [ $# != 1 ]; then
		message "Sample call: getFileName file_name"
		exit 1
	fi
	local LS_NOT_FOUND="ls:"
	local lsout=$(ls --time-style=+"%F %T" -o -g $1)
	if [ "${lsout:1:3}" != "$LS_NOT_FOUND" ]; then
		local date=$(echo $lsout | cut -d" " -f4)
		local time=$(echo $lsout | cut -d" " -f5)
		echo "$date $time"
	else
		echo ""
	fi
}

DIDB_DATE_1=$(getFileTime $DIDB_DIR_1/$DIDB_BACKUP_FILE)
DIDB_DATE_2=$(getFileTime $DIDB_DIR_2/$DIDB_BACKUP_FILE)
DRIFT_DATE_1=$(getFileTime $DRIFT_DIR_1/$DRIFT_BACKUP_FILE)
DRIFT_DATE_2=$(getFileTime $DRIFT_DIR_2/$DRIFT_BACKUP_FILE)

BACKUP_DATE=$DRIFT_DATE_2
DATABASE=DRIFT
DIR=$DRIFT_DIR_2
if [ "$DIDB_DATE_2" \< "$BACKUP_DATE" ]; then
	BACKUP_DATE=$DIDB_DATE_2
	DATABASE=DIDB
	DIR=$DIDB_DIR_2
fi
if [ "$DRIFT_DATE_1" \< "$BACKUP_DATE" ]; then
	BACKUP_DATE=$DRIFT_DATE_1
	DATABASE=DRIFT
	DIR=$DRIFT_DIR_1
fi
if [ "$DIDB_DATE_1" \< "$BACKUP_DATE" ]; then
	BACKUP_DATE=$DIDB_DATE_1
	DATABASE=DIDB
	DIR=$DIDB_DIR_1
fi

SUCCESS=$DIR/LAST_BACKUP_SUCCEEDED
FAILURE=$DIR/LAST_BACKUP_FAILED
if [ -f "$SUCCESS" ]; then
	rm $SUCCESS
fi
if [ -f "$FAILURE" ]; then
	rm $FAILURE
fi
message "Start $DATABASE backup into $DIR"

if [ "$DATABASE" == DIDB ]; then
	BACKUP_FILE=$DIDB_BACKUP_FILE
	URL=$DIDB_URL
else
	BACKUP_FILE=$DRIFT_BACKUP_FILE
	URL=$DRIFT_URL
fi

if [ -f $DIR/backup.log ]; then
	rm $DIR/backup.log
fi
message "$GBAK -y $DIR/backup.log $URL $DIR/$BACKUP_FILE" 
$GBAK -y $DIR/backup.log -user SYSDBA -fetch $PFILE $URL $DIR/$BACKUP_FILE
EXIT_CODE=$?
END=$(date +"$TIME_FORMAT")
if [ $EXIT_CODE -ne 0 ]; then
	message "Backup of $DATABASE into $DIR is abnormally terminated with exit code $EXIT_CODE"
	touch $FAILURE_FILE
	touch $FAILURE
	message "last 5 lines from $DIR/backup.log are below"
	tail -n5 $DIR/backup.log >> $MAIN_LOG_FILE
	echo -e "$END Backup of $DATABASE into $DIR is abnormally terminated with exit code $EXIT_CODE\nSee file $MAIN_LOG_FILE and $DIR/backup.log for details" > $FAILURE_FILE
	exit $EXIT_CODE
fi

message "Backup of $DATABASE into $DIR successfully finished"
touch $SUCCESS_FILE
touch $SUCCESS
message "last 5 lines from $DIR/backup.log are below"
tail -n5 $DIR/backup.log >> $MAIN_LOG_FILE
echo -e "$END Backup of $DATABASE into $DIR successfully finished\nSee file $MAIN_LOG_FILE and $DIR/backup.log for details" > $SUCCESS_FILE
