/*
 * DGU stands for (D)i(g)isonde (U)MS database
 */

CREATE DOMAIN BOOLEAN AS SMALLINT
CHECK (value is NULL or value in (0, 1));

/*
   Digisonde stations are here with there URSI codes and location coordinates 
 */
CREATE TABLE Location
(
  /* Unique Generated Identification Number for internal use only */
  Ident        SMALLINT NOT NULL,
  /* Union Radio-Scientifique Internationale code */
  UrsiCode     VARCHAR(10) NOT NULL,
  Name         VARCHAR(30) NOT NULL,
  /* Latitude */
  Lat          FLOAT NOT NULL,
  /* Longitude */
  Lon          FLOAT NOT NULL,
  Description  VARCHAR(1000),       /* Textual description of this location */
    CONSTRAINT Location_Ident_IS_PKEY PRIMARY KEY (Ident),
    CONSTRAINT Location_UrsiCode_is_unique UNIQUE (UrsiCode),
    CONSTRAINT Location_UrsiCode_is_not_empty CHECK (UrsiCode <> ''),
    CONSTRAINT Location_UrsiCode_is_caps CHECK (UrsiCode = UPPER(UrsiCode)),
    CONSTRAINT Location_Name_is_not_empty CHECK (Name <> ''),
    CONSTRAINT Location_Lat_Check CHECK (lat >=-90 AND Lat <= 90),
    CONSTRAINT Location_Lon_Check CHECK (Lon >= 0 AND lon <= 360)
);
COMMIT;

/*
   Dictionary.
   Types of Beacons (transmission stations) that are not Digisonde
   Content corresponds to enum class BeaconType
 */
CREATE TABLE BeaconType
(
    Ident   SMALLINT NOT NULL,
    Name    VARCHAR(20),
    Desc    VARCHAR(1000),
    CONSTRAINT BeaconType_Ident_IS_PKEY PRIMARY KEY (Ident),
    CONSTRAINT BeaconType_Name_is_unique UNIQUE (Name)
);

/*
 *  Transmission station that is not Digisondes. It can be Radar, Ionosonde, or may be something else  
 */
CREATE TABLE Beacon
(
    Ident   SMALLINT NOT NULL,
    TypeId  SMALLINT NOT NULL, /* 1 Radar, 2 Ionosonde, 3 others */
    /* Unique code. For Ionosondes, URSI code can be used. */
    Code    CHAR(10) NOT NULL,
    Name    CHAR(50) NOT NULL,
    /* Latitude */
    Lat          FLOAT NOT NULL,
    /* Longitude */
    Lon          FLOAT NOT NULL,
    Description  VARCHAR(1000),       /* Textual description of this location */
    CONSTRAINT Beacon_Ident_IS_PKEY PRIMARY KEY (Ident),
    CONSTRAINT Beacon_Code_is_unique UNIQUE (Code),
    CONSTRAINT Beacon_Code_is_not_empty CHECK (Code <> ''),
    CONSTRAINT Beacon_Code_is_caps CHECK (Code = UPPER(Code)),
    CONSTRAINT Beacon_Name_is_not_empty CHECK (Name <> ''),
    CONSTRAINT Beacon_Name_is_caps CHECK (Name = UPPER(Name)),
    CONSTRAINT Beacon_TypeId_is_FKey FOREIGN KEY (TypeId) REFERENCES BeaconType ON DELETE CASCADE ON UPDATE CASCADE
);
COMMIT;

/*
   Dictionary.
   Types of Tx Station (transmission stations)
   Content corresponds to enum class TxStationType
 */
CREATE TABLE TxStationType
(
    Ident   SMALLINT NOT NULL,
    Name    CHAR(20),
    Desc    VARCHAR(1000),
    CONSTRAINT TxStationType_Ident_IS_PKEY PRIMARY KEY (Ident),
    CONSTRAINT TxStationType_Name_is_caps CHECK (Name = UPPER(Name)),
    CONSTRAINT TxStationType_Name_is_unique UNIQUE (Name)
);

/*
 *  Transmission station which can be Digisondes, Radar, and may be something else  
 */
CREATE TABLE TxStation
(
    Ident       SMALLINT NOT NULL,
    TypeId      SMALLINT NOT NULL, /* 1 Digisonde, 2 Beacon */
    LocationId  SMALLINT,
    BeaconId    SMALLINT,
    CONSTRAINT TxStation_Ident_IS_PKEY PRIMARY KEY (Ident),
    CONSTRAINT TxStation_TypeId_is_FKey FOREIGN KEY (TypeId) REFERENCES TxStationType ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT TxStation_LocationId_is_FKey FOREIGN KEY (LocationId) REFERENCES Location ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT TxStation_BeaconId_is_FKey FOREIGN KEY (BeaconId) REFERENCES Beacon ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT TxStation_Exactly_One_Ref CHECK ((LocationId IS NOT NULL AND BeaconId IS NULL) OR (LocationId IS NULL AND BeaconId IS NOT NULL))
);
COMMIT;


/*
 * Disctionary
 * Mode of Operation. There are can be 4 modes of Operation/Program
 * 1. SND     SouNDing - science operation
 * 2. BIT     Built-In Test - housekeeping operation
 * 3. CEQ     Channel EQualizing - science operation
 * 4. TRKCAL  Tracker Calibration - housekeeping operation
 */
CREATE TABLE OpMode
(
    Ident   SMALLINT NOT NULL,
    Name    CHAR(15) NOT NULL,
    isHK    BOOLEAN NOT NULL, /* isHousKeeping: true = HouseKeeping, false = Science */ 
    Description VARCHAR(1000),
    CONSTRAINT OpMode_Ident_Is_PKey PRIMARY KEY (Ident),
    CONSTRAINT OpMode_Name_Is_Unique UNIQUE (Name),
    CONSTRAINT OpMode_Name_Is_Caps CHECK (Name = UPPER(Name))
);
COMMIT;

/**
 * Distionary
 * Data Processing codes.
 * For each Mode of Operation there are its own Data Procesing codes
 */
CREATE TABLE DataProc
(
    Ident       INTEGER NOT NULL,
    OpModeId    SMALLINT NOT NULL,
    DPIndex     SMALLINT NOT NULL,
    DPName      CHAR(30) NOT NULL,
    CONSTRAINT DP_Ident_Is_PKey PRIMARY KEY (Ident),
    CONSTRAINT DP_OpModeId_is_FKey FOREIGN KEY (OpModeId) REFERENCES OpMode ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT DP_OpModeId_DPIndex_is_uniq UNIQUE (OpModeId, DPIndex),
    CONSTRAINT DP_OpModeId_DPName_is_uniq UNIQUE (OpModeId, DPName)
);
COMMIT;

/**
 * Distionary
 * Data Processing Steps
 */
CREATE TABLE ProcStep
(
    Ident   INTEGER NOT NULL,
    Mnem    CHAR(10) NOT NULL,
    Name    CHAR(50) NOT NULL,
    CONSTRAINT PS_Ident_Is_PKey PRIMARY KEY (Ident),
    CONSTRAINT PS_Mnem_Is_Unique UNIQUE (Mnem),
    CONSTRAINT PS_Mnem_is_not_empty CHECK (Mnem <> ''),
    CONSTRAINT PS_Mnem_is_caps CHECK (Mnem = UPPER(Mnem)),
    CONSTRAINT PS_Name_is_not_empty CHECK (Name <> '')
);
COMMIT;

/*
 *  Dictionary
 *  Classification of Digisonde measurement program from the point of view of transmitter.
 *  Content of this dictionary is the same as corresponding enam class: 
 *   NOT_SOUNDING_OPERATION      (1, "not sounding operation"), 
 *   TX_ENABLED_VERTICAL         (2, "Tx enabled, vertical"), 
 *   TX_DISABLED_CW_LISTEN       (3, "Tx disabled, CW listening"), 
 *   TX_DISABLED_DIGI_LISTEN     (4, "Tx disabled, other Digisonde listening (D2D oblique sounding)");
 * 
 *  Content corresponds to enum class TxMode
 */
CREATE TABLE TxMode
(
    Ident   SMALLINT NOT NULL,
    Name    CHAR(20) NOT NULL,
    Desc    VARCHAR(1000),
    CONSTRAINT TxMode_Ident_Is_PKey PRIMARY KEY (Ident),
    CONSTRAINT TxMode_Name_Is_Unique UNIQUE (Name),
    CONSTRAINT TxMode_Name_is_not_empty CHECK (Name <> ''),
    CONSTRAINT TxMode_Name_is_caps CHECK (Name = UPPER(Name))
);
    
/**
 * Contains main attributes of measurement, as to start time, Operation Mode, Schedule Number, and Program Number.
 * Does not contain content of produced data, which is saved in separate table, MeasData.
 */
CREATE TABLE Measurement
(
    /* Unique Generated Identification Number for internal use only */
    Ident       INTEGER NOT NULL,
    /* Start time as UT */
    TimeUT      TIMESTAMP NOT NULL,
    /* Internal ident of Location */
    LocationId  SMALLINT NOT NULL,
    OpModeId    SMALLINT NOT NULL,
    DPIndex     INTEGER NOT NULL,
    ProgramNo   SMALLINT NOT NULL,
    ScheduleNo  SMALLINT NOT NULL,
    LowerFreqHz INTEGER,  /* NULL if not applicable for this Operation Mode */
    UpperFreqHz INTEGER,  /* NULL if not applicable for this Operation Mode */
    TxModeId    SMALLINT NOT NULL,
    TxStationId SMALLINT, /* not null for scientific sounding mode measurement with listenening of other known transmitter (can be DPS4D, DPS4, Radar, Ionosonde, and others) */ 
    CONSTRAINT Meas_Ident_Is_PKey PRIMARY KEY (Ident),
    CONSTRAINT Meas_LocID_TimeUT_is_uniq UNIQUE (LocationId, TimeUT),
    CONSTRAINT Meas_LocationID_is_FKey FOREIGN KEY (LocationId) REFERENCES Location ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT Meas_OpModeId_DPIndex_Is_FKey FOREIGN KEY (OpModeId,DPIndex) REFERENCES DataProc(OpModeId,DPIndex) ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT Meas_ProgramNo_GE1_LE256 CHECK (ProgramNo >= 1 AND ProgramNo <= 256),
    CONSTRAINT Meas_ScheduleNo_GE0_LT256 CHECK (ScheduleNo >= 0 AND ScheduleNo < 256),
    CONSTRAINT Meas_TxModeId_is_FKey FOREIGN KEY (TxModeId) REFERENCES TxMode ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT Meas_TxStationId_Is_FKey FOREIGN KEY (TxStationId) REFERENCES TxStation ON DELETE CASCADE ON UPDATE CASCADE
);
COMMIT;

/**
 * Contains content of measurements as BLOB data, that will be interpreted only by application
 */
CREATE TABLE MeasData
(
    Ident           INTEGER NOT NULL,
    MeasurementId   INTEGER NOT NULL,
    OpModeId        SMALLINT NOT NULL,
    DPIndexApp      INTEGER NOT NULL,
    Length          BIGINT NOT NULL,
    Data            BLOB NOT NULL,
    CONSTRAINT MD_Ident_Is_PKey PRIMARY KEY (Ident),
    CONSTRAINT MD_MeasurementId_Is_FKey FOREIGN KEY (MeasurementId) REFERENCES Measurement ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT MD_OpModeId_DPIndexApp_Is_FKey FOREIGN KEY (OpModeId, DPIndexApp) REFERENCES DataProc(OpModeId, DPIndex) ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT MD_Meas_OpMode_DPIndexApp_Uniq UNIQUE (MeasurementId, OpModeId, DPIndexApp),
    CONSTRAINT Meas_Length_GT_0 CHECK (Length > 0)
);
COMMIT;

/* !!!don't delete this table - it's for connection availability checking!!! */
CREATE TABLE UserReg
(
	Ident          SMALLINT  NOT NULL,
	Count_act      INTEGER   NOT NULL,
	Host_addr      CHAR(20)  NOT NULL,
	App_name       CHAR(20)  NOT NULL,
	App_version    CHAR(20)  NOT NULL,
	Connect_Time   TIMESTAMP NOT NULL,
	U_Name         CHAR(40)  NOT NULL,
	OS_Name        CHAR(20)  NOT NULL,
	OS_Version     CHAR(20)  NOT NULL,
	Host_Name      CHAR(50)  NOT NULL,
	Srv_14 TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
	CONSTRAINT UserReg_Ident_Is_PKey PRIMARY KEY (Ident),
    CONSTRAINT UserReg_H_Addr_A_Name_Is_Unique UNIQUE (Host_Addr, App_Name)
);
COMMIT;

/* And create view Measurement */
CREATE VIEW GetMeas (Ident, LocationId, UrsiCode, TimeUT, strTimeUT, OpModeId, OpModeName, isHK, DPIndex, DPName, ProgramNo, ScheduleNo, 
        LowerFreqHz, UpperFreqHz, TxModeId, TxStationId, TxUrsiCode, DPIndexApp, DPNameApp, Length)
    AS SELECT m.Ident, l.Ident, l.UrsiCode, m.TimeUT, CAST(m.TimeUT AS CHAR(24)), o.Ident, o.Name, o.isHK, p1.DPIndex, p1.DPName, m.ProgramNo, 
                m.ScheduleNo, m.LowerFreqHz, m.UpperFreqHz, m.TxModeId, lt.Ident, lt.UrsiCode, p2.DPIndex, p2.DPName, d.Length
    FROM Measurement m JOIN MeasData d ON d.MeasurementId=m.Ident 
        JOIN Location l ON l.Ident=m.LocationId 
        LEFT JOIN TxStation tx ON tx.Ident=m.TxStationId AND tx.TypeId=1 
        LEFT JOIN Location lt ON lt.Ident=tx.LocationId AND tx.TypeId=1 
        JOIN OpMode o ON m.OpModeId = o.Ident 
        JOIN DataProc p1 ON p1.OpModeId = m.OpModeId AND p1.DPIndex = m.DPIndex
        JOIN DataProc p2 ON p2.OpModeId = d.OpModeId AND p2.DPIndex = d.DPIndexApp;
COMMIT;

CREATE ROLE Loader;
CREATE ROLE Common;

/* for generation unique Ident value for new record in Location table */
CREATE GENERATOR Location_Ident_Gen;
/* for generation unique Ident value for new record in Beacon table */
CREATE GENERATOR Beacon_Ident_Gen;
/* for generation unique Ident value for new record in TxStation table */
CREATE GENERATOR TxStation_Ident_Gen;
/* for generation unique Ident value for new record in Measurement table */
CREATE GENERATOR Measurement_Ident_Gen;
/* for generation unique Ident value for new record in MeasData table */
CREATE GENERATOR MeasData_Ident_Gen;
/* for generation unique Ident value for new record in UserReg table */
CREATE GENERATOR USERREG_IDENT_GEN;


/* Create various triggers and procedures */
SET TERM !! ;

/* Insert trigger for Location table to automatically generate unique Ident value */
CREATE TRIGGER Create_Location_Ident FOR Location
    BEFORE INSERT POSITION 0
    AS BEGIN
        NEW.IDENT = GEN_ID(Location_Ident_Gen, 1); 
    END!!

/* Insert trigger for Beacon table to automatically generate unique Ident value */
CREATE TRIGGER Create_Beacon_Ident FOR Beacon
    BEFORE INSERT POSITION 0
    AS BEGIN
        NEW.IDENT = GEN_ID(Beacon_Ident_Gen, 1); 
    END!!

/* Insert trigger for TxStation table to automatically generate unique Ident value */
CREATE TRIGGER Create_TxStation_Ident FOR TxStation
    BEFORE INSERT POSITION 0
    AS BEGIN
        NEW.IDENT = GEN_ID(TxStation_Ident_Gen, 1); 
    END!!

/* Insert trigger for Measurement table to automatically generate unique Ident value */
CREATE TRIGGER Create_Measurement_Ident FOR Measurement
    BEFORE INSERT POSITION 0
    AS BEGIN
        NEW.IDENT = GEN_ID(Measurement_Ident_Gen, 1); 
    END!!

/* Insert trigger for Measurement table to automatically generate unique Ident value */
CREATE TRIGGER Create_MeasData_Ident FOR MeasData
    BEFORE INSERT POSITION 0
    AS BEGIN
        NEW.IDENT = GEN_ID(MeasData_Ident_Gen, 1); 
    END!!

/* Insert trigger for USERREG table to automatically generate unique Ident value */
CREATE TRIGGER Create_UserReg_Ident FOR UserReg
    BEFORE INSERT POSITION 0
    AS BEGIN
        NEW.IDENT = GEN_ID(UserReg_Ident_Gen, 1); 
    END!!

SET TERM ; !!
COMMIT;

/* Create indices */

/* for table Measurement */
CREATE DESCENDING INDEX DescIndOnMeasForUT ON Measurement (TimeUT);
CREATE ASCENDING INDEX AscIndOnMeasForUT ON Measurement (TimeUT);

/* !!!don't delete this grant - it's for connection availability checking!!! */
GRANT SELECT, INSERT, UPDATE ON USERREG TO UCONNECT;

GRANT SELECT, REFERENCES(Ident) ON OpMode TO Loader;
GRANT SELECT, REFERENCES(OpModeId, DPIndex) ON DataProc TO Loader;
GRANT SELECT, INSERT ON Location TO Loader;
GRANT SELECT, INSERT ON Beacon TO Loader;
GRANT SELECT, INSERT ON TxStation TO Loader;
GRANT SELECT, INSERT, UPDATE, DELETE ON Measurement TO Loader;
GRANT SELECT, INSERT, UPDATE, DELETE ON MeasData TO Loader;

GRANT SELECT ON Location TO COMMON;
GRANT SELECT ON GetMeas TO COMMON;
GRANT SELECT ON MeasData TO COMMON;
/* CREATE USER TIDLoader PASSWORD 'tidLoader' FIRSTNAME 'TID Loader' LASTNAME 'Loader'; */
GRANT Loader TO tidLoader;
GRANT COMMON TO TIDExplorer;
GRANT COMMON TO TIDReanalysis;
COMMIT;

COMMIT;
