/*******************************************************************************
******************************************************************************** 
**
**   Database: LOUI
**
********************************************************************************
********************************************************************************/


/***********************************************************
 Table: Location
   Radio-scientifique measurement stations where
   ionosonde/antenna can be located.
 **********************************************************/
CREATE TABLE Location
(
  Ident      SMALLINT NOT NULL, /* Unique Generated Identification Number for internal use only */
  UrsiCode   CHAR(10) NOT NULL, /* Union Radio-Scientifique Internationale code */
  Name       CHAR(30) NOT NULL, /* quasi-unique name */
  Lat        FLOAT NOT NULL, /* Latitude */
  Lon        FLOAT NOT NULL, /* Longitude */
  WMOID1     CHAR(10) NOT NULL,
  WMOID2     CHAR(10) NOT NULL,
  WMOID3     CHAR(10) NOT NULL,
  Description  VARCHAR(1000),       /* Textual description of this location */
    CONSTRAINT Location_Ident_IS_PKEY PRIMARY KEY( Ident ),
    CONSTRAINT Location_UrsiCode_is_unique UNIQUE( UrsiCode ),
    CONSTRAINT Location_UrsiCode_is_not_empty CHECK( UrsiCode <> '' ),
    CONSTRAINT Location_UrsiCode_is_caps CHECK( UrsiCode = UPPER( UrsiCode ) ),
    CONSTRAINT Location_Name_is_not_empty CHECK( Name <> '' ),
    CONSTRAINT Location_Name_is_caps CHECK( Name = UPPER( Name ) )
);
COMMIT;


/***********************************************************
 Table: Ionosonde
   Digisonde models.
 **********************************************************/
CREATE TABLE Ionosonde
(
  Ident SMALLINT NOT NULL,  /* Unique Identification Number for internal use only */
  Name  CHAR(30) NOT NULL,  /* Unique name/code/serial # */
  Description  VARCHAR(1000), /* Textual description of this ionosonde */
    CONSTRAINT Ionosonde_Ident_is_PKey PRIMARY KEY( Ident ),
    CONSTRAINT Ionosonde_Name_is_unique UNIQUE( Name ),
    CONSTRAINT Ionosonde_Name_is_caps CHECK( Name = UPPER( Name ) )
);
COMMIT;


/***********************************************************
 Table: AntPattern
   Antenna's Pattern (Layout).
 **********************************************************/
CREATE TABLE AntPattern
(
  Ident SMALLINT NOT NULL, /* Unique Identification Number for internal use only */
  Name  CHAR(30) NOT NULL, /* Unique name/code/serial # */
  Description  VARCHAR(1000), /* Textual description of this antenna */
    CONSTRAINT AntPattern_Ident_is_PKey PRIMARY KEY( Ident ),
    CONSTRAINT AntPattern_Name_is_unique UNIQUE( Name )
);
COMMIT;


/***********************************************************
 Table: AntRotation
   Antenna's Rotation.
 **********************************************************/
CREATE TABLE AntRotation
(
  Ident        SMALLINT NOT NULL, /* Unique Identification Number for internal use only */
  Name         CHAR(30) NOT NULL, /* Unique name/code/serial # */
  Description  VARCHAR(1000), /* Textual description of this Rotation */
    CONSTRAINT AntRotation_Ident_is_PKey PRIMARY KEY( Ident ),
    CONSTRAINT AntRotation_Name_is_unique UNIQUE( Name )
);
COMMIT;


/***********************************************************
 Table: Antenna
   Antenna's Setup.
 **********************************************************/
CREATE TABLE Antenna
(
  Ident         SMALLINT NOT NULL, /* Unique Generated Identification Number for internal use only */
  AntPatternID  SMALLINT NOT NULL, /* Internal ident of antenna's pattern */
  AntRotationID SMALLINT NOT NULL, /* Internal ident of antenna's rotation */
  TotAntennas   SMALLINT NOT NULL,
    CONSTRAINT Antenna_Ident_is_PKey PRIMARY KEY( Ident ),
    CONSTRAINT Antenna_PatID_RotID_Tot_is_Uniq UNIQUE( AntPatternID, AntRotationID, TotAntennas ),
    CONSTRAINT Antenna_AntPatternID_is_FKey FOREIGN KEY( AntPatternID ) REFERENCES AntPattern ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT Antenna_AntRotationID_is_FKey FOREIGN KEY( AntRotationID ) REFERENCES AntRotation ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT Antenna_TotAntennas_ge0 CHECK( TotAntennas >= 0 )
);
COMMIT;


/***********************************************************
 Table: AntCoord
   (x,y,z) coordinates of some Antenna's Setup.
 **********************************************************/
CREATE TABLE AntCoord
(
  AntennaID SMALLINT NOT NULL,
  SeqNumber SMALLINT NOT NULL,
  x         FLOAT NOT NULL,
  y         FLOAT NOT NULL,
  z         FLOAT NOT NULL,
    CONSTRAINT AntCoord_AntID_SeqNo_is_PKey PRIMARY KEY( AntennaID, SeqNumber),
    CONSTRAINT AntCoord_AntennaID_is_FKey FOREIGN KEY( AntennaID ) REFERENCES Antenna ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT AntCoord_SeqNumber_is_GT0_LE7 CHECK( SeqNumber > 0 AND SeqNumber <= 7 )
);
COMMIT;


/***********************************************************
 Table: Equipment
   Pairs:
   (Ionosonde, Antenna's Setup)
   This pair (Equipment) is assumed to work at some Location during
   some period of time.
 **********************************************************/
CREATE TABLE Equipment
(
  Ident          SMALLINT NOT NULL, /* Unique Generated Identification Number for internal use only */
  IonosondeID    SMALLINT NOT NULL, /* Internal (=external number)  ident of Digisonde model */
  AntennaID      SMALLINT NOT NULL, /* Internal ident of Antenna's Setup */
  PulseMult      INT NOT NULL, /* Pulse multiplier coefficient given in percents as integer */
    CONSTRAINT Equipment_Ident_is_PKey PRIMARY KEY( Ident ),
    CONSTRAINT Equipment_IonosID_AntID_is_Uniq UNIQUE( IonosondeID, AntennaID, PulseMult ),
    CONSTRAINT Equipment_IonosondeID_is_FKey FOREIGN KEY( IonosondeID ) REFERENCES Ionosonde ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT Equipment_AntennaID_is_FKey FOREIGN KEY( AntennaID ) REFERENCES Antenna ON DELETE CASCADE ON UPDATE CASCADE
);
COMMIT;


/***********************************************************
 Table: Program
   Unique set of Digisonde Program parameters ever worked on
   any Equipment (Ionosonde, Antenna's Setup).
   Using for search.
 **********************************************************/
CREATE TABLE Program
(
  Ident            SMALLINT NOT NULL,   /* Unique Identification Number for internal use only */ 
  PulseRate        INTEGER NOT NULL,    /* pulse rate */
  PhaseCode        INTEGER NOT NULL,    /* phase code */
  Delay            INTEGER NOT NULL,    /* delay */
  TotSamples       INTEGER NOT NULL,    /* N - number of samples (actual) */
  TotActSamples    INTEGER NOT NULL,    /* n - number of samples (power of 2), watch out for DGS-256 */
  CFS              DOUBLE PRECISION NOT NULL, /* coarse frequency step in MHz */
  FFS              DOUBLE PRECISION NOT NULL, /* fine frequency steps in MHz */
  NSS              INTEGER NOT NULL,    /* number of small steps in a scan */
  AntSeq           INTEGER NOT NULL,    /* antenna sequencing */

  FreqSearch       SMALLINT NOT NULL,   /* 1 = freq search enabled */
  AutoGain         SMALLINT NOT NULL,   /* 1 = auto gain enabled */
  Multiplexed      SMALLINT NOT NULL,   /* 1 = multiplexing enabled */
  PhaseSwitched    SMALLINT NOT NULL,   /* 1 = phase code with phase switched */
  O_and_X          SMALLINT NOT NULL,   /* 1 = O/X polarization option, (true=OandX, false=onlyO) */
  Alt_Antenna      SMALLINT NOT NULL,   /* 1 = alternative antenna setup */
  RadioSilent      SMALLINT NOT NULL,   /* 1 = radio silent mode */
  DubleSamples     SMALLINT NOT NULL,   /* 1 = when W=2 or 6 (DGS-256) */
  OperatingMode    INTEGER NOT NULL,
  DataFormat       INTEGER NOT NULL,    /* DPS data format from preface 1-MMM, 4-RSF, 5-SBF*/
  BaseGain         INTEGER NOT NULL,    /* base gain, from preface */
  BaseGainDB       INTEGER NOT NULL,    /* preface base gain in dB */
  ToDB             INTEGER NOT NULL,    /* echo amplitude to dB */
  AttenConstDB     INTEGER NOT NULL,    /* attenuation constant in dB */
  NDopplers        INTEGER NOT NULL,    /* total number of doppler lines */
  NDirections      INTEGER NOT NULL,    /* total number of directions */
  P1               INTEGER NOT NULL,    /* P1 - important program option */
  Z                INTEGER NOT NULL,    /* Z */
  T                INTEGER NOT NULL,    /* T */
  L                INTEGER NOT NULL,    /* L */
  PulseWidth       INTEGER NOT NULL,    /* 20 or 40 km */
  sFreq            DOUBLE PRECISION NOT NULL, /* starting frequency in MHz */
  eFreq            DOUBLE PRECISION NOT NULL, /* ending frequency in MHz */
  iFreq            DOUBLE PRECISION NOT NULL, /* increment (freq step) in MHz */
  tFreq            SMALLINT NOT NULL, /* total frequencies */
  sHeight          DOUBLE PRECISION NOT NULL, /* starting height in km */
  eHeight          DOUBLE PRECISION NOT NULL, /* ending height in km */
  iHeight          DOUBLE PRECISION NOT NULL, /* increment (height step) in km */
  tHeight          SMALLINT NOT NULL, /* total heights */
    CONSTRAINT Program_Ident_is_PKey PRIMARY KEY( Ident )
);
COMMIT;


/***********************************************************
 Table: Scaler
   "Autoscaling program" (ARTIST-4) or Expert (Operator)
   to create multirecords scaling information
 **********************************************************/
CREATE TABLE Scaler
(
  Ident        SMALLINT NOT NULL,
  LogName      CHAR(10) NOT NULL,
  FName        CHAR(20) NOT NULL,
  LName        CHAR(20) NOT NULL,
  Human        SMALLINT DEFAULT 1 NOT NULL, /* 1 - human, 0 - program */
  Rank         FLOAT NOT NULL,
    CONSTRAINT Scaler_Ident_IS_PRIMARY_KEY PRIMARY KEY( Ident ),
    CONSTRAINT Scaler_LogName_is_unique UNIQUE( LogName ),
    CONSTRAINT Scaler_LogName_is_caps CHECK( UPPER( LogName ) = LogName )
);
COMMIT;


/***********************************************************
 Table: Measurement
   The measurement (Location, Equipment)
   on some Location with some Digisonde Equipment.
 **********************************************************/
CREATE TABLE Measurement
(
  Ident          INTEGER NOT NULL, /* Unique Generated Identification Number for internal use only */
  TimeUT         TIMESTAMP NOT NULL, /* Start time as UT. Also referred as Time */
  LocationID     SMALLINT NOT NULL, /* Internal ident of Location */
  TxStationID    SMALLINT NOT NULL, /* Internal ident of Tx Station */
  EquipmentID    SMALLINT NOT NULL, /* Internal ident of Equipment */
  ProgramID      SMALLINT NOT NULL, /* Internal ident of Program */
    CONSTRAINT Measurement_Ident_is_PKey PRIMARY KEY( Ident ),
    CONSTRAINT Measurement_LocID_Time_is_uniq UNIQUE( LocationID, TimeUT ),
    CONSTRAINT Measurement_LocationID_is_FKey FOREIGN KEY( LocationID ) REFERENCES Location ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT Measurement_TxStationID_is_FKey FOREIGN KEY( TxStationID ) REFERENCES Location ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT Measurement_EquipmentID_is_FKey FOREIGN KEY( EquipmentID ) REFERENCES Equipment ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT Measurement_ProgramID_is_FKey FOREIGN KEY( ProgramID ) REFERENCES Program ON DELETE CASCADE ON UPDATE CASCADE
);
COMMIT;


/***********************************************************
 Table: Format
   Types of ionogram and scaling formats.
 **********************************************************/
CREATE TABLE Format
(
  Ident        SMALLINT NOT NULL,
  Name         CHAR(30) NOT NULL,
  Description  VARCHAR(1000) NOT NULL,
    CONSTRAINT Format_Ident_IS_PRIMARY_KEY PRIMARY KEY( Ident ),
    CONSTRAINT Format_Name_is_unique UNIQUE( Name )
);
COMMIT;


/***********************************************************
 Table: IonogramHeader
 **********************************************************/
CREATE TABLE IonogramHeader
(
  MeasurementID INTEGER NOT NULL,
  SubmissionUT  TIMESTAMP NOT NULL,  /* Time of creation/modification as UT */
  FormatID      SMALLINT NOT NULL,
  Quality       FLOAT DEFAULT 100 NOT NULL,
  Zipped        SMALLINT DEFAULT 0 NOT NULL,
  Filename      VARCHAR(64),
    CONSTRAINT IonogrmH_MeasurementID_is_PKey PRIMARY KEY( MeasurementID ),
    CONSTRAINT IonogrmH_MeasurementID_is_FKey FOREIGN KEY( MeasurementID )
               REFERENCES Measurement (Ident) ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT IonogrmH_FormatID_is_FKey FOREIGN KEY( FormatID )
               REFERENCES Format ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT Ionogram_Filename_NoBlanks CHECK (Filename IS NULL OR (POSITION(' ', Filename) = 0)),
    CONSTRAINT Ionogram_Filename_In_Lower_Case CHECK (Filename IS NULL OR Filename = LOWER(Filename))
);
COMMIT;


/***********************************************************
 Table: IonogramData
   Ionogram data (SBF, MMM, RSF).
 **********************************************************/
CREATE TABLE IonogramData
(
  MeasurementID INTEGER NOT NULL,
  Data          BLOB NOT NULL,
    CONSTRAINT IonogrmD_MeasurementID_is_PKey PRIMARY KEY( MeasurementID ),
    CONSTRAINT IonogrmD_MeasurementID_is_FKey FOREIGN KEY( MeasurementID )
               REFERENCES IonogramHeader ON DELETE CASCADE ON UPDATE CASCADE
);
COMMIT;


/***********************************************************
 Table: ScalingHeader
 **********************************************************/
CREATE TABLE ScalingHeader
(
  MeasurementID INTEGER NOT NULL,
  ScalerID      SMALLINT NOT NULL,
  SubmissionUT  TIMESTAMP NOT NULL, /* Time of creation/modification as UT */
  FormatID      SMALLINT NOT NULL,
  Quality       FLOAT NOT NULL,
  Zipped        SMALLINT DEFAULT 0 NOT NULL,
  Confidence    INTEGER,
    CONSTRAINT SclngH_Measure_Scaler_is_PKey PRIMARY KEY( MeasurementID, ScalerID ),
    CONSTRAINT SclngH_Measure_Quality_is_Uniq UNIQUE( MeasurementID, Quality ),
    CONSTRAINT SclngH_MeasurementID_is_FKey FOREIGN KEY( MeasurementID ) REFERENCES Measurement (Ident) ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT SclngH_ScalerID_is_FKey FOREIGN KEY( ScalerID ) REFERENCES Scaler(Ident) ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT SclngH_FormatID_is_FKey FOREIGN KEY( FormatID ) REFERENCES Format ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT Confidence_GE_0_LE_100 CHECK( Confidence >= 0 AND Confidence <= 100 )
);
COMMIT;


/***********************************************************
 Table: ScalingData
   Scaling data (SAO, ART).
 **********************************************************/
CREATE TABLE ScalingData
(
  MeasurementID INTEGER NOT NULL,
  ScalerID      SMALLINT NOT NULL,
  Data          BLOB NOT NULL,
    CONSTRAINT SclngD_Measure_Scaler_is_PKey PRIMARY KEY( MeasurementID, ScalerID ),
    CONSTRAINT SclngD_MeasID_ScalID_is_FKey FOREIGN KEY( MeasurementID, ScalerID )
        REFERENCES ScalingHeader ( MeasurementID, ScalerID ) ON DELETE CASCADE ON UPDATE CASCADE
);
COMMIT;


/***********************************************************
 Table: Units
   Units for table ParamCat.
 **********************************************************/
CREATE TABLE Units
(
  Ident        SMALLINT NOT NULL,
  Name         CHAR(30) NOT NULL,
  Description  VARCHAR(1000) NOT NULL,
  IsLookup     SMALLINT DEFAULT 0 NOT NULL,
    CONSTRAINT Units_Ident_IS_PRIMARY_KEY PRIMARY KEY( Ident )
);
COMMIT;


/***********************************************************
 Table: UnitsLokup
   Lokup for table Units.
 **********************************************************/
CREATE TABLE UnitsLookup
(
  UnitsID      SMALLINT NOT NULL,
  Number       SMALLINT NOT NULL,
  Name         CHAR(30) NOT NULL,
  Description  VARCHAR(1000) NOT NULL,
    CONSTRAINT UnitsLookup_Ident_is_PKey PRIMARY KEY( UnitsID, Number ),
    CONSTRAINT UnitsLookup_Name_is_unique UNIQUE( Name ),
    CONSTRAINT UnitsLookup_Number_is_FKey FOREIGN KEY( UnitsID ) REFERENCES Units ON DELETE CASCADE ON UPDATE CASCADE
);
COMMIT;


/***********************************************************
 Table: ParamCat
   Categories for numeric parameters.
 **********************************************************/
CREATE TABLE ParamCat
(
  Ident        SMALLINT NOT NULL,
  Name         CHAR(15) NOT NULL,
  Description  VARCHAR(120) NOT NULL,
  Width        SMALLINT NOT NULL,
  Fraction     SMALLINT NOT NULL,
  Exponent     SMALLINT NOT NULL,
  UnitsID      SMALLINT NOT NULL,
  MinValue     FLOAT NOT NULL,
  MaxValue     FLOAT NOT NULL,
  FillValue    FLOAT NOT NULL,
    CONSTRAINT ParamCat_Ident_IS_PRIMARY_KEY PRIMARY KEY( Ident ),
    CONSTRAINT ParamCat_Name_is_unique UNIQUE( Name ),
    CONSTRAINT ParamCat_UnitsID_is_FKey FOREIGN KEY( UnitsID )  REFERENCES Units ON DELETE CASCADE ON UPDATE CASCADE
);
COMMIT;


/***********************************************************
 Table: Param
   Values of numeric parameters.
 **********************************************************/
CREATE TABLE Param
(
  ScalerID       SMALLINT NOT NULL,
  MeasurementID  INTEGER NOT NULL,
  ParamCatID     SMALLINT NOT NULL,
  Val            FLOAT NOT NULL,
  QLetter        CHAR(1) DEFAULT '/' NOT NULL,
  DLetter        CHAR(1) DEFAULT '/' NOT NULL,
    CONSTRAINT Param_Scaler_Meas_Cat_is_Pkey  PRIMARY KEY( MeasurementID, ParamCatID ),
    CONSTRAINT Param_ScalerID_is_FKey FOREIGN KEY( ScalerID ) REFERENCES Scaler(Ident) ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT Param_MeasurementID_is_FKey FOREIGN KEY( MeasurementID ) REFERENCES Measurement (Ident) ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT Param_ParamCatID_is_FKey FOREIGN KEY( ParamCatID ) REFERENCES ParamCat ON DELETE CASCADE ON UPDATE CASCADE,
    CONSTRAINT PARAM_RATE_IS_RESTRICTED
      CHECK
      (
        (
          Val >= (SELECT MinValue FROM ParamCat WHERE Param.ParamCatID = Ident)
          AND
          Val <= (SELECT MaxValue FROM ParamCat WHERE Param.ParamCatID = Ident)
        )
        OR
        (
          Val = (SELECT FillValue FROM ParamCat WHERE Param.ParamCatID = Ident)
        )        
      )
);
COMMIT;


/* !!!don't delete this table - it's for connection availability checking!!! */
/* Table: UserReg, Owner SYSDBA */
CREATE TABLE UserReg
(
  Ident        SMALLINT  NOT NULL,
  Count_Act    INTEGER   NOT NULL,
  Host_Addr    CHAR(20)  NOT NULL,
  App_Name     CHAR(20)  NOT NULL,
  App_Version  CHAR(20)  NOT NULL,
  Time_UT      TIMESTAMP NOT NULL,
  DBUserName   CHAR(20),
  U_Name       CHAR(40)  NOT NULL,
  OS_Name      CHAR(20)  NOT NULL,
  OS_Version   CHAR(20)  NOT NULL,
  Host_Name    CHAR(50)  NOT NULL,
  Srv_Local_Time TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
    CONSTRAINT UserReg_Ident_is_PKey PRIMARY KEY ( Ident ),
    CONSTRAINT UserReg_H_Addr_A_Name_is_Unique UNIQUE( Host_Addr, App_Name )
);
COMMIT;


/* Table: Version */
CREATE TABLE Version
(
  AppName        CHAR(20)  NOT NULL,
  AppVersion     CHAR(20)  NOT NULL,
  TimeUT         TIMESTAMP NOT NULL,
  ModLevel       INT DEFAULT 1 NOT NULL,
  Incompatible   INT DEFAULT 0 NOT NULL,
  Description    VARCHAR(1000),
    CONSTRAINT Version_Name_Version_is_PKey
      PRIMARY KEY ( AppName, AppVersion ),
    CONSTRAINT Version_ModLevel_is_0_to_10
      CHECK( ModLevel > 0 AND ModLevel <= 10 ),
    CONSTRAINT Version_Incompatible_is_0_or_1
      CHECK( Incompatible = 0 OR Incompatible = 1 )
);
COMMIT;

/* Create Procedures */

SET TERM !!;

/*
 * Returns subset of table Measurement ordered by TimeUT. This subset is restricted with Location and time interval.
 * Parameters:
 *   loc     - Location Ident, SMALLINT
 *   st      - start of time interval, TIMESTAMP
 *   et      - end of time interval, TIMESTAMP
 * Returns:
 *   records which consist of columns:
 *     Ident        - (INTEGER) Unique DB-internal number for measurement.
 *     TimeUT       - (TIMESTAMP) Time (UT) of this measurement.
 *     LocationID   - (SMALLINT) Unique DB-internal number for location.
 *     EquipmentID  - (SMALLINT) Unique DB-internal number for equipment.
 */
CREATE PROCEDURE GetMeasurements (loc SMALLINT, st TIMESTAMP, et TIMESTAMP)
    RETURNS (Ident INTEGER, TimeUT TIMESTAMP, LocationID SMALLINT, EquipmentID SMALLINT)
AS
BEGIN
    FOR SELECT Ident, TimeUT, LocationID, EquipmentID FROM Measurement
        WHERE locationID = :loc AND timeUT >= :st AND timeUT <= :et 
        ORDER BY timeUT
        INTO :Ident, :timeUT, :LocationID, :equipmentID
    DO SUSPEND;
END!!


/*
 * Returns subset of table Measurement ordered by TimeUT. This subset is restricted with Station and time interval.
 * Parameters:
 *   ursi    - Location/Station ursi code, char(5)
 *   st      - start of time interval, TIMESTAMP
 *   et      - end of time interval, TIMESTAMP
 * Returns:
 *   records which consist of columns:
 *     Ident        - (INTEGER) Unique DB-internal measurement Id.
 *     TimeUT       - (TIMESTAMP) Time (UT) of this measurement.
 *     LocationID   - (SMALLINT) Unique DB-internal number for location.
 *     EquipmentID  - (SMALLINT) Unique DB-internal number for equipment.
 */
CREATE PROCEDURE GetMeasurementsAlt (ursi char(5), st TIMESTAMP, et TIMESTAMP)
    RETURNS (ident INTEGER, TimeUT TIMESTAMP, LocationID SMALLINT, EquipmentID SMALLINT)
AS
BEGIN
    FOR SELECT Ident, TimeUT, LocationID, EquipmentID FROM Measurement 
        WHERE LocationId = (SELECT Ident FROM Location WHERE UrsiCode = UPPER(:ursi)) AND timeUT >= :st AND timeUT <= :et
        ORDER BY timeUT
        INTO :ident, :timeUT, :LocationID, :equipmentID
    DO  SUSPEND;
END!!

/*
 * Returns subset of table Measurement ordered by TimeUT.
 * This subset is restricted with Location and time interval.
 * Parameters:
 *   loc     - Location Ident, SMALLINT
 *   st      - start of time interval, TIMESTAMP
 *   et      - end of time interval, TIMESTAMP
 * Returns:
 *   records which consist of columns:
 *     MeasID       - (INTEGER) Unique DB-internal measurement Id.
 *     LocID        - (SMALLINT) Unique DB-internal number for location.
 *     TimeUT       - (TIMESTAMP) Time (UT) of this measurement.
 *     strTimeUT    - (CHAR(24)) TimeUT casted to CHAR(24). Will be deleted in future
 *     SCount       - number of scaling measurements. Can be 0. Can be more than 1.
 *     IExist       - Can be either 0 or 1. 1 means ionogram exists, 0 means ionogram does not exist.
 */
CREATE PROCEDURE MeasListInfo (loc SMALLINT, st TIMESTAMP, et TIMESTAMP)
    RETURNS (MeasID INTEGER, LocID SMALLINT, timeUT TIMESTAMP, StrTimeUT CHAR(24), SCount SMALLINT, IExists SMALLINT)
AS
BEGIN
    FOR SELECT M.Ident, M.LocationID, M.TimeUT, CAST (M.TimeUT AS CHAR(24)), COUNT(S.MeasurementID), COUNT(DISTINCT I.MeasurementID)
        FROM GetMeasurements(:loc, :st, :et) M LEFT JOIN ScalingHeader S ON M.Ident = S.MeasurementID LEFT JOIN
             IonogramHeader I ON M.Ident = I.MeasurementID
        GROUP BY M.TimeUT, M.Ident, M.LocationID 
        ORDER BY TimeUT
        INTO :MeasID, :LocID, :timeUT, :StrTimeUT, :SCount, :IExists
    DO  SUSPEND;
END!!


/*
 * Returns subset of table Measurement ordered by TimeUT.
 * This subset is restricted with Location/Station and time interval.
 * Parameters:
 *   ursi    - Location/Station ursi code, char(5)
 *   st      - start of time interval, TIMESTAMP
 *   et      - end of time interval, TIMESTAMP
 * Returns:
 *   records which consist of columns:
 *     MeasID       - (INTEGER) Unique DB-internal measurement Id.
 *     LocID        - (SMALLINT) Unique DB-internal number for location.
 *     TimeUT       - (TIMESTAMP) Time (UT) of this measurement.
 *     SCount       - number of scaling measurements. Can be 0. Can be more than 1.
 *     IExist       - Can be either 0 or 1. 1 means ionogram exists, 0 means ionogram does not exist.
 */
CREATE PROCEDURE MeasListInfoAlt(ursi char(5), st TIMESTAMP, et TIMESTAMP)
    RETURNS (MeasID INTEGER, LocID SMALLINT, timeUT TIMESTAMP, SCount SMALLINT, IExists SMALLINT)
AS
BEGIN
    FOR SELECT M.Ident, M.LocationID, M.TimeUT, COUNT(S.MeasurementID), COUNT(DISTINCT I.MeasurementID)
        FROM GetMeasurementsAlt(:ursi, :st, :et) M LEFT JOIN ScalingHeader S ON M.Ident = S.MeasurementID LEFT JOIN
             IonogramHeader I ON M.Ident = I.MeasurementID
        GROUP BY M.TimeUT, M.Ident, M.LocationID 
        ORDER BY M.TimeUT
        INTO :MeasID, :LocID, :timeUT, :SCount, :IExists
    DO  SUSPEND;
END!!


/*
 * Returns exactly one row consisting of only one column named ExistFlag of SMALLINT type. Which is equal 1 if at least one
 * measurement exists for given location in given time interval, and 0, otherwise.
 * 
 * Note:
 *   Time interval, st-et, includes the start time, st, but does not include the end time, et.
 *   Use procedure ExistMeasInTimeRangeIncl if you want to include the end time, et.
 *
 * Parameters:
 *   loc - Location Ident, SMALLINT
 *   st  - start of time interval, TIMESTAMP. Included.
 *   et  - end of time interval, TIMESTAMP. Not included.
 */
CREATE PROCEDURE ExistMeasurementsInTimeRange (loc INTEGER, st TIMESTAMP, et TIMESTAMP)
    RETURNS (existFlag SMALLINT)
AS
DECLARE VARIABLE Ident INTEGER;
BEGIN
    existFlag = 0;
    ident = -1;
    SELECT First 1 Ident FROM Measurement WHERE locationID = :loc AND timeUT >= :st AND timeUT < :et INTO :ident;
    IF (ident >= 0)
    THEN existFlag = 1;
    SUSPEND;
END!!


/*
 * Returns exactly one row consisting of only one column named ExistFlag of SMALLINT type. Which is equal 1 if at least one
 * measurement exists for given location in given time interval, and 0, otherwise.
 * 
 * Note:
 *   Time interval, st-et, includes both, the start time, st, and the end time, et.
 *   Use procedure ExistMeasurementsInTimeRange if you do not want to include the end time, et.
 *
 * Parameters:
 *   loc - Location Ident, SMALLINT
 *   st  - start of time interval, TIMESTAMP. Included.
 *   et  - end of time interval, TIMESTAMP. Included.
 */
CREATE PROCEDURE ExistMeasInTimeRangeIncl (loc INTEGER, st TIMESTAMP, et TIMESTAMP)
    RETURNS (existFlag SMALLINT)
AS
DECLARE VARIABLE Ident INTEGER;
BEGIN
    existFlag = 0;
    ident = -1;
    SELECT FIRST 1 Ident FROM Measurement WHERE locationID = :loc AND timeUT >= :st AND timeUT <= :et INTO :Ident;
    IF (ident >= 0)
    THEN existFlag = 1;
    SUSPEND;
END!!


/*
 * Returns exactly one row consisting of only one column named ExistFlag of SMALLINT type. Which is equal 1 if at least one
 * measurement exists for given location in given time interval, and 0, otherwise.
 * 
 * Note:
 *   Time interval, st-et, includes the start time, st, but does not include the end time, et.
 *   Use procedure ExistMeasInTimeRangeIncl if you want to include the end time, et.
 *
 * Parameters:
 *   ursi    - Location/Station ursi code, char(5)
 *   st  - start of time interval, TIMESTAMP. Included.
 *   et  - end of time interval, TIMESTAMP. Not included.
 */
CREATE PROCEDURE ExistMeasInTimeRangeAlt (ursi char(5), st TIMESTAMP, et TIMESTAMP)
    RETURNS (existFlag SMALLINT)
AS
DECLARE VARIABLE Ident INTEGER;
BEGIN
    existFlag = 0;
    ident = -1;
    SELECT First 1 Ident FROM Measurement  
        WHERE LocationId = (SELECT Ident FROM Location WHERE UrsiCode = UPPER(:ursi)) AND timeUT >= :st AND timeUT < :et 
        INTO :ident;
    IF (ident >= 0)
    THEN existFlag = 1;
    SUSPEND;
END!!


/*
 * Returns exactly one row consisting of only one column named ExistFlag of SMALLINT type. Which is equal 1 if at least one
 * measurement exists for given location in given time interval, and 0, otherwise.
 * 
 * Note:
 *   Time interval, st-et, includes both, the start time, st, and the end time, et.
 *   Use procedure ExistMeasurementsInTimeRange if you do not want to include the end time, et.
 *
 * Parameters:
 *   ursi    - Location/Station ursi code, char(5)
 *   st  - start of time interval, TIMESTAMP. Included.
 *   et  - end of time interval, TIMESTAMP. Included.
 */
CREATE PROCEDURE ExistMeasInTimeRangeIncAlt (ursi char(5), st TIMESTAMP, et TIMESTAMP)
    RETURNS (existFlag SMALLINT)
AS
DECLARE VARIABLE Ident INTEGER;
BEGIN
    existFlag = 0;
    ident = -1;
    SELECT First 1 Ident FROM Measurement  
        WHERE LocationId = (SELECT Ident FROM Location WHERE UrsiCode = UPPER(:ursi)) AND timeUT >= :st AND timeUT <= :et 
        INTO :ident;
    IF (ident >= 0)
    THEN existFlag = 1;
    SUSPEND;
END!!


/*
 * Returns exactly one row consisting of only one column named ExistFlag of SMALLINT type. Which is equal 1 if at least one
 * measurement exists for any station in given time interval, and 0, otherwise.
 * 
 * Note:
 *   Time interval, st-et, includes the start time, st, but does not include the end time, et.
 *   Use procedure ExistAnyMeasInTimeRangeIncl if you want to include the end time, et.
 *
 * Parameters:
 *   st - start of time interval, TIMESTAMP. Included.
 *   et - end of time interval, TIMESTAMP. Not included.
 */
CREATE PROCEDURE ExistAnyMeasurementsInTimeRange (st TIMESTAMP, et TIMESTAMP)
    RETURNS (existFlag SMALLINT)
AS
DECLARE VARIABLE Ident INTEGER;
BEGIN
    existFlag = 0;
    ident = -1;
    SELECT FIRST 1 Ident FROM Measurement WHERE timeUT >= :st AND timeUT < :et INTO :Ident;
    IF (ident >= 0)
    THEN existFlag = 1;
    SUSPEND;
END!!


/*
 * Returns exactly one row consisting of only one column named ExistFlag of SMALLINT type. Which is equal 1 if at least one
 * measurement exists for any station in given time interval, and 0, otherwise.
 * 
 * Note:
 *   Time interval, st-et, includes bothe the start time, st, and the end time, et.
 *   Use procedure ExistAnyMeasurementsInTimeRange if you do not want to include the end time, et.
 *
 * Parameters:
 *   st - start of time interval, TIMESTAMP. Included.
 *   et - end of time interval, TIMESTAMP. Included.
 */
CREATE PROCEDURE ExistAnyMeasInTimeRangeIncl (st TIMESTAMP, et TIMESTAMP)
    RETURNS (existFlag SMALLINT)
AS
DECLARE VARIABLE Ident INTEGER;
BEGIN
    existFlag = 0;
    ident = -1;
    SELECT FIRST 1 Ident FROM Measurement WHERE timeUT >= :st AND timeUT <= :et INTO :Ident;
    IF (ident >= 0)
    THEN existFlag = 1;
    SUSPEND;
END!!


/*
 * Returns rows consisting of only one column named D of Timestamp type. These are the dates from given date interval
 * which contain at least one measurement for given location.
 * Note, rows are delivered by date increasing order.
 * Parameters:
 *   loc     - Location Ident, SMALLINT
 *   sd      - start of time interval, DATE
 *   ed      - end of time interval, DATE
 */
CREATE PROCEDURE ListOfDaysForPeriod (loc INTEGER, st TIMESTAMP, et TIMESTAMP)
    RETURNS (d TIMESTAMP)
AS
DECLARE VARIABLE existFlag SMALLINT;
DECLARE VARIABLE st1 TIMESTAMP;
DECLARE VARIABLE et1 TIMESTAMP;
DECLARE VARIABLE sd DATE;
DECLARE VARIABLE ed DATE;
BEGIN
    sd = CAST(st AS DATE);
    ed = CAST(et AS DATE);
    st1 = '1900.01.01';
    SELECT MIN(TimeUT) FROM Measurement WHERE LocationID = :loc INTO :st1;
    IF (st1 <> '1900.01.01')
    THEN
    BEGIN
        IF (CAST(st1 AS DATE) > sd)
        THEN sd = CAST(st1 AS DATE);
    END
    et1 = '1900.01.01';
    SELECT MAX(TimeUT) FROM Measurement WHERE LocationID = :loc INTO :et1;
    IF (et1 <> '1900.01.01')
    THEN
    BEGIN
        IF (CAST(et1 AS DATE) < ed)
        THEN ed = CAST(et1 AS DATE);
    END
    WHILE (sd <= ed)
    DO
    BEGIN
        st = CAST (sd AS TIMESTAMP);
        et = CAST ((sd+1) AS TIMESTAMP);
        SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et) INTO :existFlag;
        IF (existFlag = 1)
        THEN
        BEGIN
            d = CAST (sd AS TIMESTAMP);
            SUSPEND;
        END
        sd = sd + 1;
    END
END!!


/*
 * Returns rows consisting of only one column named D of DATE type. These are the dates from given date interval
 * which contain at least one measurement for given location.
 * Note, rows are delivered by date increasing order.
 * Parameters:
 *   ursi    - Location/Station ursi code, char(5)
 *   sd      - start of time interval, DATE
 *   ed      - end of time interval, DATE
 */
CREATE PROCEDURE ListOfDaysForPeriodAlt (ursi CHAR(5), st TIMESTAMP, et TIMESTAMP)
    RETURNS (d TIMESTAMP)
AS
BEGIN
    FOR SELECT * FROM ListOfDaysForPeriod((SELECT Ident FROM Location WHERE UrsiCode = UPPER(:ursi)),:st,:et) INTO :d
    DO SUSPEND;
END!!


/*
 * Returns rows consisting of only one column named D of Timestamp type. These are the dates from given date interval
 * which contain at least one measurement.
 * Note, rows are delivered by date increasing order.
 * Parameters:
 *   st      - start of time interval, TIMESTAMP
 *   et      - end of time interval, TIMESTAMP
 */
CREATE PROCEDURE ListOfDaysForPeriodAnyStation (st TIMESTAMP, et TIMESTAMP)
    RETURNS (d TIMESTAMP)
AS
DECLARE VARIABLE existFlag SMALLINT;
DECLARE VARIABLE st1 TIMESTAMP;
DECLARE VARIABLE et1 TIMESTAMP;
DECLARE VARIABLE sd DATE;
DECLARE VARIABLE ed DATE;
BEGIN
    sd = CAST(st AS DATE);
    ed = CAST(et AS DATE);
    st1 = '1900.01.01';
    SELECT MIN(TimeUT) FROM Measurement INTO :st1;
    IF (st1 <> '1900.01.01')
    THEN
    BEGIN
        IF (CAST(st1 AS DATE) > sd)
        THEN sd = CAST(st1 AS DATE);
    END
    et1 = '1900.01.01';
    SELECT MAX(TimeUT) FROM Measurement INTO :et1;
    IF (et1 <> '1900.01.01')
    THEN
    BEGIN
        IF (CAST(et1 AS DATE) < ed)
        THEN ed = CAST(et1 AS DATE);
    END
    WHILE (sd <= ed)
    DO
    BEGIN
        st = CAST (sd AS TIMESTAMP);
        et = CAST ((sd+1) AS TIMESTAMP);
        SELECT existFlag FROM ExistAnyMeasurementsInTimeRange(:st, :et) INTO :existFlag;
        IF (existFlag = 1)
        THEN
        BEGIN
            d = CAST (sd AS TIMESTAMP);
            SUSPEND;
        END
        sd = sd + 1;
    END
END!!


/*
 * Returns one row consisting of 12 columns of SMALLINT. 1st column corresponds to January, 2nd to Febuary and so on. The values of columns can be either 1 or 0.
 * 1 means measurements exist for corresponding month for given station and year, 0 means no measurement exists for corresponding month for given station and year.
 * Parameters:
 *   loc     - Location/station Ident, SMALLINT
 *   yr      - year, SMALLINT
 */
CREATE PROCEDURE ListOfMonthsForYear (loc INTEGER, yr SMALLINT)
    RETURNS (yJan SMALLINT, yFeb SMALLINT, yMar SMALLINT, yApr SMALLINT, yMay SMALLINT, yJun SMALLINT
           , yJul SMALLINT, yAug SMALLINT, ySep SMALLINT, yOct SMALLINT, yNov SMALLINT, yDec SMALLINT)
AS
DECLARE VARIABLE st TIMESTAMP;
DECLARE VARIABLE et TIMESTAMP;
DECLARE VARIABLE strYear CHAR(4);
BEGIN
	strYear = CAST (yr AS CHAR(4));
    st = CAST ((strYear || '.01.01') AS TIMESTAMP); 	
    et = CAST ((strYear || '.02.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et) INTO :yJan;
    st = et;     
    et = CAST ((strYear || '.03.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et) INTO :yFeb;
    st = et;     
    et = CAST ((strYear || '.04.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et) INTO :yMar;
    st = et;     
    et = CAST ((strYear || '.05.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et) INTO :yApr;
    st = et;     
    et = CAST ((strYear || '.06.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et) INTO :yMay;
    st = et;     
    et = CAST ((strYear || '.07.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et) INTO :yJun;
    st = et;     
    et = CAST ((strYear || '.08.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et) INTO :yJul;
    st = et;     
    et = CAST ((strYear || '.09.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et) INTO :yAug;
    st = et;     
    et = CAST ((strYear || '.10.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et) INTO :ySep;
    st = et;     
    et = CAST ((strYear || '.11.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et) INTO :yOct;
    st = et;     
    et = CAST ((strYear || '.12.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et) INTO :yNov;
    st = et;     
    strYear = CAST ((yr + 1) AS CHAR(4));
    et = CAST ((strYear || '.01.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et) INTO :yDec;
    SUSPEND;
END!!


/*
 * Returns one row consisting of 12 columns of SMALLINT. 1st column corresponds to January, 2nd to Febuary and so on. The values of columns can be either 1 or 0.
 * 1 means measurements exist for corresponding month for given station and year, 0 means no measurement exists for corresponding month for given station and year.
 * Parameters:
 *   ursi    - Location/Station ursi code, char(5)
 *   yr      - year, SMALLINT
 */
CREATE PROCEDURE ListOfMonthsForYearAlt (ursi CHAR(5), yr SMALLINT)
    RETURNS (yJan SMALLINT, yFeb SMALLINT, yMar SMALLINT, yApr SMALLINT, yMay SMALLINT, yJun SMALLINT
           , yJul SMALLINT, yAug SMALLINT, ySep SMALLINT, yOct SMALLINT, yNov SMALLINT, yDec SMALLINT)
AS
BEGIN
    SELECT * FROM ListOfMonthsForYear((SELECT Ident FROM Location WHERE UrsiCode = UPPER (:ursi)), :yr) 
        INTO :yJan, :yFeb, :yMar, :yApr, :yMay, :yJun, :yJul, :yAug, :ySep, :yOct, :yNov, :yDec;
    SUSPEND;
END!!


/*
 * Returns one row consisting of 12 columns of SMALLINT. 1st column corresponds to January, 2nd to Febuary and so on. The values of columns can be either 1 or 0.
 * 1 means measurements exist for corresponding month and year, 0 means no measurement exists for corresponding month and year.
 * Parameters:
 *   yr      - year, SMALLINT
 */
CREATE PROCEDURE ListOfMonthsForYearAnyStation (yr SMALLINT)
    RETURNS (yJan SMALLINT, yFeb SMALLINT, yMar SMALLINT, yApr SMALLINT, yMay SMALLINT, yJun SMALLINT
           , yJul SMALLINT, yAug SMALLINT, ySep SMALLINT, yOct SMALLINT, yNov SMALLINT, yDec SMALLINT)
AS
DECLARE VARIABLE st TIMESTAMP;
DECLARE VARIABLE et TIMESTAMP;
DECLARE VARIABLE strYear CHAR(4);
BEGIN
	strYear = CAST (yr AS CHAR(4));
    st = CAST ((strYear || '.01.01') AS TIMESTAMP); 	
    et = CAST ((strYear || '.02.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistAnyMeasurementsInTimeRange(:st, :et) INTO :yJan;
    st = et;     
    et = CAST ((strYear || '.03.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistAnyMeasurementsInTimeRange(:st, :et) INTO :yFeb;
    st = et;     
    et = CAST ((strYear || '.04.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistAnyMeasurementsInTimeRange(:st, :et) INTO :yMar;
    st = et;     
    et = CAST ((strYear || '.05.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistAnyMeasurementsInTimeRange(:st, :et) INTO :yApr;
    st = et;     
    et = CAST ((strYear || '.06.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistAnyMeasurementsInTimeRange(:st, :et) INTO :yMay;
    st = et;     
    et = CAST ((strYear || '.07.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistAnyMeasurementsInTimeRange(:st, :et) INTO :yJun;
    st = et;     
    et = CAST ((strYear || '.08.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistAnyMeasurementsInTimeRange(:st, :et) INTO :yJul;
    st = et;     
    et = CAST ((strYear || '.09.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistAnyMeasurementsInTimeRange(:st, :et) INTO :yAug;
    st = et;     
    et = CAST ((strYear || '.10.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistAnyMeasurementsInTimeRange(:st, :et) INTO :ySep;
    st = et;     
    et = CAST ((strYear || '.11.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistAnyMeasurementsInTimeRange(:st, :et) INTO :yOct;
    st = et;     
    et = CAST ((strYear || '.12.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistAnyMeasurementsInTimeRange(:st, :et) INTO :yNov;
    st = et;     
    strYear = CAST ((yr + 1) AS CHAR(4));
    et = CAST ((strYear || '.01.01') AS TIMESTAMP);  
    SELECT existFlag FROM ExistAnyMeasurementsInTimeRange(:st, :et) INTO :yDec;
    SUSPEND;
END!!


/*
 * Returns list of years where at least one measurement exists for given station.
 * This list returns records of one column which contains year. 
 */
CREATE PROCEDURE ListOfYearsForStation(loc INTEGER)
    RETURNS (yr SMALLINT)
AS
DECLARE VARIABLE existFlag SMALLINT;
DECLARE VARIABLE st TIMESTAMP;
DECLARE VARIABLE et TIMESTAMP;
DECLARE VARIABLE sd DATE;
DECLARE VARIABLE ed DATE;
DECLARE VARIABLE lastYear SMALLINT;
DECLARE VARIABLE strYear CHAR(10);
DECLARE VARIABLE strNextYear CHAR(10);
BEGIN
    sd = '2000.01.01';
    ed = '1900.01.01';
    st = '1900.01.01';
    SELECT MIN(TimeUT) FROM Measurement WHERE LocationID = :loc INTO :st;
    IF (st <> '1900.01.01')
    THEN sd = CAST (st AS DATE);
    et = '1900.01.01';
    SELECT MAX(TimeUT) FROM Measurement WHERE LocationID = :loc INTO :et;
    IF (et <> '1900.01.01')
    THEN ed = CAST( et AS DATE );
    
    yr = EXTRACT (YEAR FROM sd);
    lastYear = EXTRACT (YEAR FROM ed);
    
    WHILE (yr <= lastYear)
    DO
    BEGIN
	    strYear = CAST (yr AS CHAR(4)) || '.01.01';
        strNextYear = CAST ((yr+1) AS CHAR(4)) || '.01.01';
        st =  CAST (strYear AS TIMESTAMP);
        et = CAST (strNextYear AS TIMESTAMP);
        SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et) INTO :existFlag;
        IF( existFlag = 1 )
        THEN SUSPEND;
        yr = yr + 1;
    END
END!!


/*
 * Returns list of years where at least one measurement exists for given station.
 * This list returns records of one column which contains year. 
 */
CREATE PROCEDURE ListOfYearsForStationAlt (ursi CHAR(5))
    RETURNS (yr SMALLINT)
AS
BEGIN
    FOR SELECT yr FROM ListOfYearsForStation((SELECT Ident FROM Location WHERE UrsiCode = UPPER(:ursi))) INTO :yr
    DO SUSPEND;
END!!


/*
 * Returns list of years where at least one measurement exists of any station.
 * This list returns records of one column which contains year. 
 */
CREATE PROCEDURE ListOfYears RETURNS (yr SMALLINT)
AS
DECLARE VARIABLE existFlag SMALLINT;
DECLARE VARIABLE st TIMESTAMP;
DECLARE VARIABLE et TIMESTAMP;
DECLARE VARIABLE sd DATE;
DECLARE VARIABLE ed DATE;
DECLARE VARIABLE lastYear SMALLINT;
DECLARE VARIABLE strYear CHAR(10);
DECLARE VARIABLE strNextYear CHAR(10);
BEGIN
    sd = '2000.01.01';
    ed = '1900.01.01';
    st = '1900.01.01';
    SELECT MIN(TimeUT) FROM Measurement INTO :st;
    IF (st <> '1900.01.01')
    THEN sd = CAST (st AS DATE);
    et = '1900.01.01';
    SELECT MAX(TimeUT) FROM Measurement INTO :et;
    IF (et <> '1900.01.01')
    THEN ed = CAST( et AS DATE );
    
    yr = EXTRACT (YEAR FROM sd);
    lastYear = EXTRACT (YEAR FROM ed);
    
    WHILE (yr <= lastYear)
    DO
    BEGIN
        strYear = CAST (yr AS CHAR(4)) || '.01.01';
        strNextYear = CAST ((yr+1) AS CHAR(4)) || '.01.01';
        st =  CAST (strYear AS TIMESTAMP);
        et = CAST (strNextYear AS TIMESTAMP);
        SELECT existFlag FROM ExistAnyMeasurementsInTimeRange(:st, :et) INTO :existFlag;
        IF( existFlag = 1 )
        THEN SUSPEND;
        yr = yr + 1;
    END
END!!


/*
 * This procedure returns rows consisting of only one column named D of DATE type. These are the dates which contain at least
 * one measurement for given location.
 * Note, rows are delivered by date increasing order.
 * Parameters:
 *   loc     - Location Ident, SMALLINT
 */
CREATE PROCEDURE ListOfDays (loc INTEGER) RETURNS (d TIMESTAMP)
AS
DECLARE VARIABLE existFlag SMALLINT;
DECLARE VARIABLE st TIMESTAMP;
DECLARE VARIABLE et TIMESTAMP;
DECLARE VARIABLE sd DATE;
DECLARE VARIABLE ed DATE;
BEGIN
	sd = '2000.01.01';
	st = '1900.01.01';
	SELECT MIN(TimeUT) FROM Measurement WHERE LocationID = :loc INTO :st;
	IF (st <> '1900.01.01')
	THEN sd = CAST (st AS DATE);
    ed = '1900.01.01';
	et = '1900.01.01';
	SELECT MAX(TimeUT) FROM Measurement WHERE LocationID = :loc INTO :et;
	IF (et <> '1900.01.01')
	THEN ed = CAST (et AS DATE);
	WHILE (sd <= ed)
	DO
	BEGIN
		st = CAST (sd AS TIMESTAMP);
		et = CAST ((sd+1) AS TIMESTAMP);
		SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et) INTO :existFlag;
		IF (existFlag = 1)
		THEN
		BEGIN
		    d = CAST (sd AS TIMESTAMP);
		    SUSPEND;
		END
		sd = sd + 1;
	END
END!!


/*
 * This procedure returns rows consisting of three columns.
 * The names of columns are:
 *   YR - year (SMALLINT);
 *   MO - month (SMALLINT);
 *   QY - Quantity of days in month containing at least
 *        one measurement (SMALLINT);
 * Procedure returns exactly one row for each month that contains at least
 * one measurement for specified location <loc>.
 * Note, rows are delivered by date increasing order.
 * Parameters:
 *   loc     - Location Ident, SMALLINT
 */
CREATE PROCEDURE StationInventory (loc INTEGER)
    RETURNS (yr SMALLINT, mo SMALLINT, qy SMALLINT)
AS
DECLARE VARIABLE existFlag SMALLINT;
DECLARE VARIABLE st TIMESTAMP;
DECLARE VARIABLE et TIMESTAMP;
DECLARE VARIABLE sd DATE;
DECLARE VARIABLE ed DATE;
DECLARE VARIABLE nyr SMALLINT;
DECLARE VARIABLE nmo SMALLINT;
BEGIN
	sd = '2000.01.01';
	st = '1900.01.01';
	SELECT MIN(TimeUT) FROM Measurement WHERE LocationID = :loc INTO :st;
	IF (st <> '1900.01.01')
	THEN sd = CAST (st AS DATE);
    ed = '1900.01.01';
	et = '1900.01.01';
	SELECT MAX(TimeUT) FROM Measurement WHERE LocationID = :loc INTO :et;
	IF (et <> '1900.01.01')
	THEN ed = CAST (et AS DATE);
	
	yr = EXTRACT (YEAR FROM sd);
	mo = EXTRACT (MONTH FROM sd);
	qy = 0;
	
	WHILE (sd <= ed)
	DO
	BEGIN
		st = CAST (sd AS TIMESTAMP);
		et = CAST ((sd+1) AS TIMESTAMP);
		SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et) INTO :existFlag;
		IF (existFlag = 1)
		THEN qy = qy + 1;
		sd = sd + 1;
		nyr = EXTRACT (YEAR FROM sd);
		nmo = EXTRACT (MONTH FROM sd);
		IF ((nmo <> mo) OR (nyr <> yr))
		THEN
		BEGIN
			IF (qy > 0)
			THEN SUSPEND;
			yr = nyr;
			mo = nmo;
			qy = 0;
		END
	END
	IF (qy > 0)
	THEN SUSPEND;
END!!


/*
 * This procedure returns rows consisting of four columns.
 * The names of columns are:
 *   URSI - URSI code of Location (CHAR(10));
 *   YR   - year (SMALLINT);
 *   MO   - month (SMALLINT);
 *   QY   - Quantity of days in month containing at least
 *          one measurement (SMALLINT);
 * For each location thia procedure returns as many rows as number of months
 * that contains at least one measurement for this location.
 * Note, rows are grouped (and ordered by URSI Code) by locations and
 * ordered inside location by date (increasingly).
 */
CREATE PROCEDURE StationsInventory
    RETURNS (ursi CHAR(10), yr SMALLINT, mo SMALLINT, qy SMALLINT)
AS
DECLARE VARIABLE loc SMALLINT;
BEGIN
    FOR SELECT Ident, ursicode FROM Location ORDER BY ursicode INTO :loc, :ursi
    DO
    BEGIN
        FOR SELECT yr, mo, qy FROM StationInventory(:loc) INTO :yr, :mo, :qy
        DO SUSPEND;
    END
END!!


/*
 * This procedure is used in another procedure, StnInvAsCalendar.
 * It is here only because there is no array concept in stored procedures.
 */
CREATE PROCEDURE SetQtyForMonth( mo SMALLINT, qy SMALLINT,
    qJan SMALLINT, qFeb SMALLINT, qMar SMALLINT, qApr SMALLINT, qMay SMALLINT, qJun SMALLINT, 
    qJul SMALLINT, qAug SMALLINT, qSep SMALLINT, qOct SMALLINT, qNov SMALLINT, qDec SMALLINT )
RETURNS (nJan SMALLINT, nFeb SMALLINT, nMar SMALLINT, nApr SMALLINT, nMay SMALLINT, nJun SMALLINT,
         nJul SMALLINT, nAug SMALLINT, nSep SMALLINT, nOct SMALLINT, nNov SMALLINT, nDec SMALLINT)
AS
BEGIN
	nJan = qJan;
	nFeb = qFeb;
	nMar = qMar;
	nApr = qApr;
	nMay = qMay;
	nJun = qJun;
	nJul = qJul;
	nAug = qAug;
	nSep = qSep;
	nOct = qOct;
	nNov = qNov;
	nDec = qDec;
	IF (mo = 1)
	THEN nJan = qy;
	IF (mo = 2)
	THEN nFeb = qy;
	IF (mo = 3)
	THEN nMar = qy;
	IF (mo = 4)
	THEN nApr = qy;
	IF (mo = 5)
	THEN nMay = qy;
	IF (mo = 6)
	THEN nJun = qy;
	IF (mo = 7)
	THEN nJul = qy;
	IF (mo = 8)
	THEN nAug = qy;
	IF (mo = 9)
	THEN nSep = qy;
	IF (mo = 10)
	THEN nOct = qy;
	IF (mo = 11)
	THEN nNov = qy;
	IF (mo = 12)
	THEN nDec = qy;
	SUSPEND;
	EXIT;
END!!


/*
 * This procedure is used in another procedure, StnInvAsCalendar.
 * It is here only because there is no array concept in stored procedures.
 */
CREATE PROCEDURE SetStrQtysForMonths (nJan SMALLINT, nFeb SMALLINT, nMar SMALLINT, nApr SMALLINT, nMay SMALLINT, nJun SMALLINT,
    nJul SMALLINT, nAug SMALLINT, nSep SMALLINT, nOct SMALLINT, nNov SMALLINT, nDec SMALLINT )
RETURNS (Jan CHAR(2), Feb CHAR(2), Mar CHAR(2), Apr CHAR(2), May CHAR(2), Jun CHAR(2),
         Jul CHAR(2), Aug CHAR(2), Sep CHAR(2), Oct CHAR(2), Nov CHAR(2), "DEC" CHAR(2))
AS
BEGIN
	IF (nJan < 10)
	THEN
		IF (nJan = 0)
		THEN Jan = '  ';
		ELSE Jan = ' ' || CAST (nJan AS CHAR(1));
	ELSE
	   Jan = CAST (nJan AS CHAR(2));
	IF (nFeb < 10)
	THEN
		IF (nFeb = 0)
		THEN Feb = '  ';
		ELSE Feb = ' ' || CAST (nFeb AS CHAR(1));
	ELSE
	    Feb = CAST (nFeb AS CHAR(2));
	IF (nMar < 10)
	THEN
		IF (nMar = 0)
		THEN Mar = '  ';
		ELSE Mar = ' ' || CAST (nMar AS CHAR(1));
	ELSE
	    Mar = CAST (nMar AS CHAR(2));
	IF (nApr < 10)
	THEN
		IF (nApr = 0)
		THEN Apr = '  ';
		ELSE Apr = ' ' || CAST (nApr AS CHAR(1));
	ELSE
	    Apr = CAST (nApr AS CHAR(2));
	IF (nMay < 10)
	THEN
		IF (nMay = 0)
		THEN May = '  ';
		ELSE May = ' ' || CAST (nMay AS CHAR(1));
	ELSE
	    May = CAST (nMay AS CHAR(2));
	IF (nJun < 10)
	THEN
		IF (nJun = 0)
		THEN Jun = '  ';
		ELSE Jun = ' ' || CAST (nJun AS CHAR(1));
	ELSE
	    Jun = CAST (nJun AS CHAR(2));
	IF (nJul < 10)
	THEN
		IF (nJul = 0)
		THEN Jul = '  ';
		ELSE Jul = ' ' || CAST (nJul AS CHAR(1));
	ELSE
	    Jul = CAST (nJul AS CHAR(2));
	IF (nAug < 10)
	THEN
		IF (nAug = 0)
		THEN Aug = '  ';
		ELSE Aug = ' ' || CAST (nAug AS CHAR(1));
	ELSE
	    Aug = CAST (nAug AS CHAR(2));
	IF (nSep < 10)
	THEN
		IF (nSep = 0)
		THEN Sep = '  ';
		ELSE Sep = ' ' || CAST (nSep AS CHAR(1));
	ELSE
	    Sep = CAST (nSep AS CHAR(2));
	IF (nOct < 10)
	THEN
		IF (nOct = 0)
		THEN Oct = '  ';
		ELSE Oct = ' ' || CAST (nOct AS CHAR(1));
	ELSE
	    Oct = CAST (nOct AS CHAR(2));
	IF( nNov < 10 )
	THEN
		IF (nNov = 0)
		THEN Nov = '  ';
		ELSE Nov = ' ' || CAST (nNov AS CHAR(1));
	ELSE
	    Nov = CAST (nNov AS CHAR(2));
	IF (nDec < 10)
	THEN
		IF (nDec = 0)
		THEN "DEC" = '  ';
		ELSE "DEC" = ' ' || CAST (nDec AS CHAR(1));
	ELSE
	    "DEC" = CAST (nDec AS CHAR(2));
	SUSPEND;
	EXIT;
END!!


/*
 * This procedure returns rows consisting of 13 columns.
 * The names of columns are:
 *   "YEAR" - year (CHAR(4));
 *   JAN    - Quantity of days in January containing  at least one measurement for given location (CHAR(2));
 *   FEB    - Quantity of days in February containing  at least one measurement for given location (CHAR(2));
 *   MAR    - Quantity of days in March containing  at least one measurement for given location (CHAR(2));
 *   .............................................................
 *        ... and so on ...
 *   .............................................................
 *   "DEC"  - Quantity of days in December containing  at least one measurement for given location (CHAR(2));
 *
 * Procedure returns exactly one row for each year that contains at least one measurement for specified location <loc>.
 * Note, rows are delivered by date increasing order.
 * Parameters:
 *   loc     - Location Ident, SMALLINT
 */
CREATE PROCEDURE StnInvAsCalendar (loc SMALLINT)
    RETURNS ("YEAR" CHAR(4), Jan CHAR(2), Feb CHAR(2), Mar CHAR(2), Apr CHAR(2), May CHAR(2), Jun CHAR(2),
             Jul CHAR(2), Aug CHAR(2), Sep CHAR(2), Oct CHAR(2), Nov CHAR(2), "DEC" CHAR(2) )
AS
DECLARE VARIABLE existFlag SMALLINT;
DECLARE VARIABLE st TIMESTAMP;
DECLARE VARIABLE et TIMESTAMP;
DECLARE VARIABLE sd DATE;
DECLARE VARIABLE ed DATE;
DECLARE VARIABLE yr SMALLINT;
DECLARE VARIABLE mo SMALLINT;
DECLARE VARIABLE qy SMALLINT;
DECLARE VARIABLE nyr SMALLINT;
DECLARE VARIABLE nmo SMALLINT;
DECLARE VARIABLE nJan SMALLINT;
DECLARE VARIABLE nFeb SMALLINT;
DECLARE VARIABLE nMar SMALLINT;
DECLARE VARIABLE nApr SMALLINT;
DECLARE VARIABLE nMay SMALLINT;
DECLARE VARIABLE nJun SMALLINT;
DECLARE VARIABLE nJul SMALLINT;
DECLARE VARIABLE nAug SMALLINT;
DECLARE VARIABLE nSep SMALLINT;
DECLARE VARIABLE nOct SMALLINT;
DECLARE VARIABLE nNov SMALLINT;
DECLARE VARIABLE nDec SMALLINT;
BEGIN
	nJan = 0;
	nFeb = 0;
	nMar = 0;
	nApr = 0;
	nMay = 0;
	nJun = 0;
	nJul = 0;
	nAug = 0;
	nSep = 0;
	nOct = 0;
	nNov = 0;
	nDec = 0;
	Jan  = '??';
	Feb  = '??';
	Mar  = '??';
	Apr  = '??';
	May  = '??';
	Jun  = '??';
	Jul  = '??';
	Aug  = '??';
	Sep  = '??';
	Oct  = '??';
	Nov  = '??';
	"DEC" = '??';
	"YEAR" = '????';
	sd = '2000.01.01';
	st = '1900.01.01';
	SELECT MIN(TimeUT) FROM Measurement WHERE LocationID = :loc INTO :st;
	IF (st <> '1900.01.01')
	THEN sd = CAST (st AS DATE);
    ed = '1900.01.01';
	et = '1900.01.01';
	SELECT MAX(TimeUT) FROM Measurement WHERE LocationID = :loc INTO :et;
	IF (et <> '1900.01.01')
	THEN ed = CAST (et AS DATE);
	
	yr = EXTRACT (YEAR FROM sd);
	mo = EXTRACT (MONTH FROM sd);
	qy = 0;
	
	WHILE (sd <= ed)
	DO
	BEGIN
		st = CAST (sd AS TIMESTAMP);
		et = CAST ((sd+1) AS TIMESTAMP);
		SELECT existFlag FROM ExistMeasurementsInTimeRange(:loc, :st, :et)
		   INTO :existFlag;
		IF (existFlag = 1)
		THEN qy = qy + 1;
		sd = sd + 1;
		nyr = EXTRACT (YEAR FROM sd);
		nmo = EXTRACT (MONTH FROM sd);
		
		IF (nmo <> mo)
		THEN
		BEGIN
			IF (qy > 0)
			THEN
			BEGIN
				SELECT * FROM SetQtyForMonth(:mo,:qy,:nJan,:nFeb,:nMar,:nApr,:nMay,:nJun,:nJul,:nAug,:nSep,:nOct,:nNov,:nDec)
				    INTO :nJan,:nFeb,:nMar,:nApr,:nMay,:nJun,:nJul,:nAug,:nSep,:nOct,:nNov,:nDec;
				qy = 0;
			END
			mo = nmo;
		END
		IF (nyr <> yr)
		THEN
		BEGIN
			IF (nJan>0 OR nFeb>0 OR nMar>0 OR nApr>0 OR nMay>0 OR nJun>0 OR nJul>0 OR nAug>0 OR nSep>0 OR nOct>0 OR nNov>0 OR nDec>0)
			THEN
			BEGIN
				SELECT * FROM SetStrQtysForMonths(:nJan,:nFeb,:nMar,:nApr,:nMay,:nJun,:nJul,:nAug,:nSep,:nOct,:nNov,:nDec)
				    INTO :Jan,:Feb,:Mar,:Apr,:May,:Jun,:Jul,:Aug,:Sep,:Oct,:Nov,:"DEC";
				"YEAR" = CAST (yr AS CHAR(4));
				SUSPEND;
				nJan = 0;
				nFeb = 0;
				nMar = 0;
				nApr = 0;
				nMay = 0;
				nJun = 0;
				nJul = 0;
				nAug = 0;
				nSep = 0;
				nOct = 0;
				nNov = 0;
				nDec = 0;
			END
			yr = nyr;
		END
	END
	IF (qy > 0)
	THEN
	BEGIN
		SELECT * FROM SetQtyForMonth(:mo,:qy,:nJan,:nFeb,:nMar,:nApr,:nMay,:nJun,:nJul,:nAug,:nSep,:nOct,:nNov,:nDec)
		    INTO :nJan,:nFeb,:nMar,:nApr,:nMay,:nJun,:nJul,:nAug,:nSep,:nOct,:nNov,:nDec;
	END
	IF (nJan>0 OR nFeb>0 OR nMar>0 OR nApr>0 OR nMay>0 OR nJun>0 OR nJul>0 OR nAug>0 OR nSep>0 OR nOct>0 OR nNov>0 OR nDec>0)
	THEN
	BEGIN
		SELECT * FROM SetQtyForMonth(:mo,:qy,:nJan,:nFeb,:nMar,:nApr,:nMay,:nJun,:nJul,:nAug,:nSep,:nOct,:nNov,:nDec)
		    INTO :nJan,:nFeb,:nMar,:nApr,:nMay,:nJun,:nJul,:nAug,:nSep,:nOct,:nNov,:nDec;
		SELECT * FROM SetStrQtysForMonths(:nJan,:nFeb,:nMar,:nApr,:nMay,:nJun,:nJul,:nAug,:nSep,:nOct,:nNov,:nDec)
		    INTO :Jan,:Feb,:Mar,:Apr,:May,:Jun,:Jul,:Aug,:Sep,:Oct,:Nov,:"DEC";
		"YEAR" = CAST (yr AS CHAR(4));
		SUSPEND;
	END
END!!


/*
 * This procedure returns rows consisting of 14 columns.
 * The names of columns are:
 *   URSI   - URSI code of location (CHAR(10));
 *   "YEAR" - year (CHAR(4));
 *   JAN    - Quantity of days in January containing  at least one measurement for given location (CHAR(2));
 *   FEB    - Quantity of days in February containing  at least one measurement for given location (CHAR(2));
 *   MAR    - Quantity of days in March containing  at least one measurement for given location (CHAR(2));
 *   .............................................................
 *        ... and so on ...
 *   .............................................................
 *   "DEC"  - Quantity of days in December containing  at least one measurement for given location (CHAR(2));
 *
 * Procedure returns exactly one row for each year that contains at least one measurement for specified location <loc>.
 * Note, rows are delivered by date increasing order.
 * Parameters:
 *   loc     - Location Ident, SMALLINT
 */
CREATE PROCEDURE InvAsCalendar
    RETURNS (Ursi CHAR(10), "YEAR" CHAR(4), 
            Jan CHAR(2), Feb CHAR(2), Mar CHAR(2), Apr CHAR(2), May CHAR(2), Jun CHAR(2),
            Jul CHAR(2), Aug CHAR(2), Sep CHAR(2), Oct CHAR(2), Nov CHAR(2), "DEC" CHAR(2))
AS
DECLARE VARIABLE loc SMALLINT;
BEGIN
    FOR SELECT Ident, ursicode FROM Location ORDER BY ursicode INTO :loc, :Ursi
    DO
    BEGIN
        FOR SELECT "YEAR", JAN, FEB, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, "DEC"
            FROM StnInvAsCalendar(:loc)
            INTO :"YEAR", :JAN, :FEB, :MAR, :APR, :MAY, :JUN, :JUL, :AUG, :SEP, :OCT, :NOV, :"DEC"
        DO SUSPEND;
    END
END!!


/*
 * Create Stations list. Returned columns are:
 *   1) ident   - Location ID; (SMALLINT)
 *   2) ursi    - URSI code of this location; (CHAR(10))
 *   3) name    - Name of this locationl (CHAR(30))
 *   4) minTime - Minimum measurement time for this location; (TIMESTAMP)
 *   5) maxTime - Maximum measurement time for this location; (TIMESTAMP)
 */
CREATE PROCEDURE StationsList
    RETURNS (ident SMALLINT, ursi CHAR(10), name CHAR(30), minTime TIMESTAMP, maxTime TIMESTAMP)
AS
BEGIN
    FOR SELECT Ident, ursicode, name FROM Location ORDER BY ursicode INTO :ident, :ursi, :name
    DO
    BEGIN
        SELECT MIN(timeUT) FROM Measurement WHERE LocationID = :ident INTO :minTime;
        SELECT MAX(timeUT) FROM Measurement WHERE LocationID = :ident INTO :maxTime;
        SUSPEND;
    END
END!!


/*
 * Returns current time stamp of DB server. 
 * Caution: it returns local time of DB server
 */
CREATE PROCEDURE CurrentTimestamp
    RETURNS (currTS TIMESTAMP)
AS
BEGIN
    currTS = CURRENT_TIMESTAMP;
    SUSPEND;
END!!


SET TERM ; !!

/* Create Views */

/* Ionogram */
CREATE VIEW Ionogram (MeasurementID, SubmissionUT, FormatID, Data, Zipped)
AS
    SELECT H.MeasurementID, H.SubmissionUT, H.FormatID, D.Data, H.Zipped FROM IonogramHeader H, IonogramData D
        WHERE H.MeasurementID = D.MeasurementID;

/* Scaling */
CREATE VIEW Scaling (MeasurementID, ScalerID, SubmissionUT, FormatID, Data, Quality, Zipped, Confidence)
AS
    SELECT H.MeasurementID, H.ScalerID, H.SubmissionUT, H.FormatID, D.Data, H.Quality, H.Zipped, H.Confidence
        FROM ScalingHeader H, ScalingData D
        WHERE H.MeasurementID = D.MeasurementID AND H.ScalerID = D.ScalerID;

/* List of active Scalers for given Measurement */
CREATE VIEW ActiveScalers (MeasID, Ident, LogName, Rank, FName, LName, Human)
AS
    SELECT M.Ident, C.Ident, C.LogName, C.Rank, C.FName, C.LName, C.Human
        FROM Measurement M LEFT JOIN ScalingHeader S ON M.Ident = S.MeasurementID LEFT JOIN Scaler C ON S.ScalerID = C.Ident;

/* The best Scaler's scaling  - ScalerID == NULL if no scaling */
CREATE VIEW BestScaler (MeasID, ScalerID) 
AS
    SELECT A.MeasID, A.Ident FROM ActiveScalers A
    WHERE Rank = (SELECT MAX(C.Rank) FROM ActiveScalers C WHERE C.MeasID = A.MeasID)
    UNION
    SELECT MeasID, Ident FROM ActiveScalers WHERE Ident IS NULL;

/* Distribution number of days per month where data exist */
CREATE VIEW StnInv (stn, yr, mo, Ndays) 
AS
    SELECT DISTINCT UrsiCode, EXTRACT(YEAR FROM timeUT), EXTRACT(MONTH FROM timeUT),
      (
          SELECT COUNT( DISTINCT EXTRACT(DAY FROM timeUT) )
          FROM Measurement M1, Location L1
          WHERE M1.LocationID = L1.Ident AND
              EXTRACT( YEAR FROM M1.timeUT ) = EXTRACT( YEAR FROM M.timeUT ) AND
              EXTRACT( MONTH FROM M1.timeUT ) = EXTRACT( MONTH FROM M.timeUT )
      )
    FROM Measurement M, Location L
    WHERE M.LocationID = L.Ident;

COMMIT;


/* Create indices */

/* for table Measurement */
CREATE DESCENDING INDEX DescIndOnMeasForLocTimeUT ON Measurement (LocationID,TimeUT);
CREATE DESCENDING INDEX DescIndOnMeasurementForTime_UT ON Measurement (TimeUT);
CREATE ASCENDING  INDEX AscIndOnMeasurementForTime_UT ON Measurement (TimeUT);
CREATE DESCENDING INDEX DescIndOnMeasurementForIdent ON Measurement (Ident);

/* for table IonogramHeader */
CREATE DESCENDING INDEX DescIndOnIonHeadForMeasID ON IonogramHeader (MeasurementID);

/* for table IonogramHeader */
CREATE ASCENDING INDEX AscIndOnScHeadForSubmissUT ON ScalingHeader (SubmissionUT);


/* Create various generators */

/* for generation unique Ident value for new record in Measurement table */
CREATE GENERATOR Measurement_Ident_Gen;
/* for generation unique Ident value for new record in Antenna table */
CREATE GENERATOR Antenna_Ident_Gen;
/* for generation unique Ident value for new record in Equipment table */
CREATE GENERATOR Equipment_Ident_Gen;
/* for generation unique Ident value for new record in Location table */
CREATE GENERATOR Location_Ident_Gen;
/* for generation unique Ident value for new record in Program table */
CREATE GENERATOR Program_Ident_Gen;
/* for generation unique Ident value for new record in Scaler table */
CREATE GENERATOR Scaler_Ident_Gen;
/* for generation unique IDENT value for new record in UserReg table */
CREATE GENERATOR USERREG_IDENT_GEN;
COMMIT;


/* Create various triggers and procedures */

SET TERM !! ;

/* Insert trigger for Location table to automatically generate unique Ident value */
CREATE TRIGGER Create_Location_Ident FOR Location
	BEFORE INSERT POSITION 0
	AS BEGIN
		NEW.Ident = GEN_ID( Location_Ident_Gen, 1 );
	END!!

/* Insert trigger for Antenna table to automatically generate unique Ident value */
CREATE TRIGGER Create_Antenna_Ident FOR Antenna
        BEFORE INSERT POSITION 0
	AS BEGIN
		NEW.Ident = GEN_ID( Antenna_Ident_Gen, 1 );
	END!!

/* Insert trigger for Equipment table to automatically generate unique Ident value */
CREATE TRIGGER Create_Equipment_Ident FOR Equipment
        BEFORE INSERT POSITION 0
	AS BEGIN
		NEW.Ident = GEN_ID( Equipment_Ident_Gen, 1 );
	END!!

/* Insert trigger for Program table to automatically generate unique Ident value */
CREATE TRIGGER Create_Program_Ident FOR Program
        BEFORE INSERT POSITION 0
	AS BEGIN
		NEW.Ident = GEN_ID( Program_Ident_Gen, 1 );
	END!!

/* Insert trigger for Measurement table to automatically generate unique Ident value */
CREATE TRIGGER Create_Measurement_Ident FOR Measurement
	BEFORE INSERT POSITION 0
	AS BEGIN
		NEW.Ident = GEN_ID( Measurement_Ident_Gen, 1 );
	END!!

/* Insert trigger for Scaler table to automatically generate unique Ident value */
CREATE TRIGGER Create_Scaler_Ident FOR Scaler
	BEFORE INSERT POSITION 0
	AS BEGIN
		NEW.Ident = GEN_ID( Scaler_Ident_Gen, 1 );
	END!!

/* Insert trigger for UserReg table to automatically generate unique Ident value */
CREATE TRIGGER Create_UserReg_Ident FOR UserReg
	BEFORE INSERT POSITION 0
	AS BEGIN
		NEW.IDENT = GEN_ID(UserReg_Ident_Gen, 1);
	END!!

/* Insert trigger for Location table to automatically convert 
   Name and UrsiCode fields to upper case                      */
CREATE TRIGGER Ins_Convert_Location_Upper FOR Location
	BEFORE INSERT POSITION 0
	AS BEGIN
		NEW.Name     = UPPER( NEW.Name );
		NEW.UrsiCode = UPPER( NEW.UrsiCode );
	END!!

/* Update trigger for Location table to automatically convert 
   Name and UrsiCode fields to upper case */
CREATE TRIGGER Upd_Convert_Location_Upper FOR Location
	BEFORE UPDATE POSITION 0
	AS BEGIN
		NEW.Name = UPPER( NEW.Name );
		NEW.UrsiCode = UPPER( NEW.UrsiCode );
	END!!

/* Insert trigger for Ionosonde table to automatically convert Name field to upper case */
CREATE TRIGGER Ins_Convert_Ionosonde_Upper FOR Ionosonde
	BEFORE INSERT POSITION 0
	AS BEGIN
		NEW.Name = UPPER( NEW.Name );
	END!!

/* Update trigger for Ionosonde table to automatically convert Name field to upper case */
CREATE TRIGGER Upd_Convert_Ionosonde_Upper FOR Ionosonde
	BEFORE UPDATE POSITION 0
	AS BEGIN
		NEW.Name = UPPER( NEW.Name );
	END!!

/* Insert trigger for Scaler table to automatically convert LogName field to upper case */
CREATE TRIGGER Ins_Convert_Scaler_Upper FOR Scaler
	BEFORE INSERT POSITION 0
	AS BEGIN
		NEW.LogName = UPPER( NEW.LogName );
	END!!

/* Update trigger for Scaler table to automatically convert LogName field to upper case */
CREATE TRIGGER Upd_Convert_Scaler_Upper FOR Scaler
	BEFORE UPDATE POSITION 0
	AS BEGIN
		NEW.LogName = UPPER( NEW.LogName );
	END!!

/* Update trigger for UserReg table to automatically increase COUNT field value */
CREATE TRIGGER Upd_UserReg_Inc_Count FOR UserReg
	BEFORE UPDATE POSITION 0
	AS BEGIN
		NEW.Count_Act = OLD.Count_Act + 1;
		NEW.Srv_Local_Time = CURRENT_TIMESTAMP;
	END!!

SET TERM ; !!

/* Create Roles */
CREATE ROLE Common;   /* regular user role */
CREATE ROLE Expert;   /* expert role */
CREATE ROLE Artist;   /* Autoscaling program role */
COMMIT;

/* Create Grants */

/* !!!don't delete this grant - it's for connection availability checking!!! */
GRANT SELECT, INSERT, UPDATE ON USERREG TO UCONNECT;

GRANT SELECT ON Location     TO Common, Expert, Artist;
GRANT SELECT ON Ionosonde    TO Common, Expert, Artist;
GRANT SELECT ON AntPattern   TO Common, Expert, Artist;
GRANT SELECT ON AntRotation  TO Common, Expert, Artist;
GRANT SELECT ON Antenna      TO Common, Expert, Artist;
GRANT SELECT ON AntCoord     TO Common, Expert, Artist;
GRANT SELECT ON Equipment    TO Common, Expert, Artist;
GRANT SELECT ON Program      TO Common, Expert, Artist;
GRANT SELECT ON Measurement  TO Common, Expert, Artist;
GRANT SELECT ON Format       TO Common, Expert, Artist;
GRANT SELECT ON IonogramHeader TO Common, Expert, Artist;
GRANT SELECT ON IonogramData   TO Common, Expert, Artist;
GRANT SELECT ON ScalingHeader  TO Common, Expert, Artist;
GRANT SELECT ON ScalingData    TO Common, Expert, Artist;
GRANT INSERT, UPDATE, DELETE ON ScalingHeader TO Expert, Artist;
GRANT INSERT, UPDATE, DELETE ON ScalingData   TO Expert, Artist;
GRANT SELECT ON Scaler       TO Common, Expert, Artist;
GRANT INSERT ON Scaler       TO Artist;
GRANT SELECT ON Units        TO Common, Expert, Artist;
GRANT SELECT ON UnitsLookup  TO Common, Expert, Artist;
GRANT SELECT ON ParamCat     TO Common, Expert, Artist;
GRANT SELECT ON Param        TO Common, Expert, Artist;
GRANT INSERT, UPDATE, DELETE ON Param     TO Expert, Artist;
GRANT SELECT ON Version      TO Common, Expert, Artist;

GRANT EXECUTE ON PROCEDURE GetMeasurements TO COMMON, EXPERT, Artist;
GRANT SELECT  ON Measurement TO PROCEDURE GetMeasurements;

GRANT EXECUTE ON PROCEDURE GetMeasurementsAlt TO COMMON, EXPERT, Artist;
GRANT SELECT  ON Location TO PROCEDURE GetMeasurementsAlt;
GRANT SELECT  ON Measurement TO PROCEDURE GetMeasurementsAlt;

GRANT EXECUTE ON PROCEDURE MeasListInfo    TO COMMON, EXPERT, Artist;
GRANT EXECUTE ON PROCEDURE GetMeasurements TO PROCEDURE MeasListInfo;
GRANT SELECT  ON ScalingHeader  TO PROCEDURE MeasListInfo;
GRANT SELECT  ON IonogramHeader TO PROCEDURE MeasListInfo;

GRANT EXECUTE ON PROCEDURE MeasListInfoAlt TO COMMON, EXPERT, Artist;
GRANT EXECUTE ON PROCEDURE GetMeasurementsAlt TO PROCEDURE MeasListInfoAlt;
GRANT SELECT  ON ScalingHeader  TO PROCEDURE MeasListInfoAlt;
GRANT SELECT  ON IonogramHeader TO PROCEDURE MeasListInfoAlt;

GRANT EXECUTE ON PROCEDURE ExistMeasurementsInTimeRange TO COMMON, EXPERT, Artist;
GRANT SELECT  ON Measurement TO PROCEDURE ExistMeasurementsInTimeRange;

GRANT EXECUTE ON PROCEDURE ExistMeasInTimeRangeAlt TO COMMON, EXPERT, Artist;
GRANT SELECT  ON Location TO PROCEDURE ExistMeasInTimeRangeAlt;
GRANT SELECT  ON Measurement TO PROCEDURE ExistMeasInTimeRangeAlt;

GRANT EXECUTE ON PROCEDURE ExistMeasInTimeRangeIncl TO COMMON, EXPERT, Artist;
GRANT SELECT  ON Measurement TO PROCEDURE ExistMeasInTimeRangeIncl;

GRANT EXECUTE ON PROCEDURE ExistMeasInTimeRangeIncAlt TO COMMON, EXPERT, Artist;
GRANT SELECT  ON Location TO PROCEDURE ExistMeasInTimeRangeIncAlt;
GRANT SELECT  ON Measurement TO PROCEDURE ExistMeasInTimeRangeIncAlt;

GRANT EXECUTE ON PROCEDURE ExistAnyMeasurementsInTimeRange TO COMMON, EXPERT, Artist;
GRANT SELECT  ON Measurement TO PROCEDURE ExistAnyMeasurementsInTimeRange;

GRANT EXECUTE ON PROCEDURE ExistAnyMeasInTimeRangeIncl TO COMMON, EXPERT, Artist;
GRANT SELECT  ON Measurement TO PROCEDURE ExistAnyMeasInTimeRangeIncl;

GRANT EXECUTE ON PROCEDURE ListOfDaysForPeriod TO COMMON, EXPERT, Artist;
GRANT EXECUTE ON PROCEDURE ExistMeasurementsInTimeRange TO PROCEDURE ListOfDaysForPeriod;
GRANT SELECT  ON Measurement TO PROCEDURE ListOfDaysForPeriod;

GRANT EXECUTE ON PROCEDURE ListOfDaysForPeriodAlt TO COMMON, EXPERT, Artist;
GRANT SELECT  ON Location TO PROCEDURE ListOfDaysForPeriodAlt;
GRANT EXECUTE ON PROCEDURE ListOfDaysForPeriod TO PROCEDURE ListOfDaysForPeriodAlt;

GRANT EXECUTE ON PROCEDURE ListOfDaysForPeriodAnyStation TO COMMON, EXPERT, Artist;
GRANT EXECUTE ON PROCEDURE ExistAnyMeasurementsInTimeRange TO PROCEDURE ListOfDaysForPeriodAnyStation;
GRANT SELECT  ON Measurement TO PROCEDURE ListOfDaysForPeriodAnyStation;

GRANT EXECUTE ON PROCEDURE ListOfMonthsForYear TO COMMON, EXPERT, Artist;
GRANT EXECUTE ON PROCEDURE ExistMeasurementsInTimeRange TO PROCEDURE ListOfMonthsForYear;
GRANT SELECT  ON Measurement TO PROCEDURE ListOfMonthsForYear;

GRANT EXECUTE ON PROCEDURE ListOfMonthsForYearAlt TO COMMON, EXPERT, Artist;
GRANT SELECT  ON Location TO PROCEDURE ListOfMonthsForYearAlt;
GRANT EXECUTE ON PROCEDURE ListOfMonthsForYear TO PROCEDURE ListOfMonthsForYearAlt;

GRANT EXECUTE ON PROCEDURE ListOfMonthsForYearAnyStation TO COMMON, EXPERT, Artist;
GRANT EXECUTE ON PROCEDURE ExistAnyMeasurementsInTimeRange TO PROCEDURE ListOfMonthsForYearAnyStation;
GRANT SELECT  ON Measurement TO PROCEDURE ListOfMonthsForYearAnyStation;

GRANT EXECUTE ON PROCEDURE ListOfYearsForStation TO COMMON, EXPERT, Artist;
GRANT EXECUTE ON PROCEDURE ExistMeasurementsInTimeRange TO PROCEDURE ListOfYearsForStation;
GRANT SELECT  ON Measurement TO PROCEDURE ListOfYearsForStation;

GRANT EXECUTE ON PROCEDURE ListOfYearsForStationAlt TO COMMON, EXPERT, Artist;
GRANT SELECT  ON Location TO PROCEDURE ListOfYearsForStationAlt;
GRANT EXECUTE ON PROCEDURE ListOfYearsForStation TO PROCEDURE ListOfYearsForStationAlt;

GRANT EXECUTE ON PROCEDURE ListOfYears TO COMMON, EXPERT, Artist;
GRANT EXECUTE ON PROCEDURE ExistMeasurementsInTimeRange TO PROCEDURE ListOfYears;
GRANT SELECT  ON Measurement TO PROCEDURE ListOfYears;

GRANT EXECUTE ON PROCEDURE ListOfDays          TO COMMON, EXPERT, Artist;
GRANT EXECUTE ON PROCEDURE ExistMeasurementsInTimeRange TO PROCEDURE ListOfDays;
GRANT SELECT  ON Measurement TO PROCEDURE ListOfDays;

GRANT EXECUTE ON PROCEDURE StationInventory    TO COMMON, EXPERT, Artist;
GRANT EXECUTE ON PROCEDURE ExistMeasurementsInTimeRange TO PROCEDURE StationInventory;
GRANT SELECT  ON Measurement TO PROCEDURE StationInventory;

GRANT EXECUTE ON PROCEDURE StationsInventory   TO COMMON, EXPERT, Artist;
GRANT EXECUTE ON PROCEDURE StationInventory    TO PROCEDURE StationsInventory;
GRANT SELECT  ON Location  TO PROCEDURE StationsInventory;

GRANT EXECUTE ON PROCEDURE StnInvAsCalendar    TO COMMON, EXPERT, Artist;
GRANT EXECUTE ON PROCEDURE ExistMeasurementsInTimeRange TO PROCEDURE StnInvAsCalendar;
GRANT EXECUTE ON PROCEDURE SetQtyForMonth      TO PROCEDURE StnInvAsCalendar;
GRANT EXECUTE ON PROCEDURE SetStrQtysForMonths TO PROCEDURE StnInvAsCalendar;
GRANT SELECT  ON Measurement TO PROCEDURE StnInvAsCalendar;

GRANT EXECUTE ON PROCEDURE InvAsCalendar       TO COMMON, EXPERT, Artist;
GRANT EXECUTE ON PROCEDURE StnInvAsCalendar    TO PROCEDURE InvAsCalendar;
GRANT SELECT  ON Location    TO PROCEDURE InvAsCalendar;

GRANT EXECUTE ON PROCEDURE StationsList        TO COMMON, EXPERT, Artist;
GRANT SELECT  ON Location    TO PROCEDURE StationsList;
GRANT SELECT  ON Measurement TO PROCEDURE StationsList;

GRANT SELECT ON Ionogram       TO Common, Expert, Artist;
GRANT SELECT ON Scaling        TO Common, Expert, Artist;
GRANT SELECT ON BestScaler     TO COMMON, EXPERT, Artist;
GRANT SELECT ON StnInv         TO COMMON, EXPERT, Artist;
GRANT SELECT ON ActiveScalers  TO COMMON, EXPERT, Artist;

COMMIT;
