/*
CREATE DATABASE 'C:\projects\db\test' USER 'SYSDBA' PASSWORD 'mypassword'
       PAGE_SIZE = 16384
       DEFAULT CHARACTER SET UTF8 COLLATION UNICODE_CI_AI;
*/

/*
 * 1 - ordinary
 * 2 - extraordinary
 */
CREATE DOMAIN Polarization 
	AS CHAR(1)
	CHECK (VALUE IS NULL OR VALUE IN (1,2));

/***********************************************************
 Table: Station, Owner SYSDBA
   Digisonde stations table
 **********************************************************/
CREATE TABLE Station
(
	/* Unique Generated Identification Number for internal use only */
	Ident		SMALLINT NOT NULL,
	/* Union Radio-Scientifique Internationale code */
	UrsiCode	VARCHAR(20) NOT NULL,
	/* Name of station */
	Name		VARCHAR(50) NOT NULL,
	/* Latitude in degrees */
	Lat			FLOAT NOT NULL,
	/* Longitude in degrees */
	Lon			FLOAT NOT NULL,
	/* Textual description of this station */
	Description  VARCHAR(1000),       
		CONSTRAINT Station_Ident_is_PKey PRIMARY KEY (Ident),
		CONSTRAINT Station_UrsiCode_is_Not_Empty CHECK (UrsiCode <> ''),
		CONSTRAINT Station_UrsiCode_is_NoSpaces CHECK (CHAR_LENGTH(UrsiCode) = CHAR_LENGTH(TRIM(UrsiCode))),
		CONSTRAINT Station_UrsiCode_is_Caps CHECK (UrsiCode = UPPER(UrsiCode)),
		CONSTRAINT Station_UrsiCode_is_Unique UNIQUE (UrsiCode),
		CONSTRAINT Station_Name_is_Not_Empty CHECK (Name <> ''),
		CONSTRAINT Station_Name_is_NoSpaces CHECK (CHAR_LENGTH(Name) = CHAR_LENGTH(TRIM(Name))),
		CONSTRAINT Station_Lat_Check CHECK (Lat >= -90 AND Lat <= 90),
		CONSTRAINT Station_Lon_Check CHECK (Lon >= 0 AND Lon <= 360)
);
COMMIT;

/*******************************************************************
   Dictionary.
   Types of Beacons (transmission stations) that are not Digisonde
   Content corresponds to enum class BeaconType
 ******************************************************************/
CREATE TABLE BeaconType
(
    Ident   SMALLINT NOT NULL,
    Name    VARCHAR(50) NOT NULL,
    Desc    VARCHAR(1000),
    	CONSTRAINT BeaconType_Ident_is_PKey PRIMARY KEY (Ident),
		CONSTRAINT BeaconType_Name_is_Not_Empty CHECK (Name <> ''),
		CONSTRAINT BeaconType_Name_is_NoSpaces CHECK (CHAR_LENGTH(Name) = CHAR_LENGTH(TRIM(Name))),
    	CONSTRAINT BeaconType_Name_is_Unique UNIQUE (Name)
);
COMMIT;

/******************************************************************
   Transmission station that is not Digisondes. It can be Radar, Ionosonde, or may be something else  
 ******************************************************************/
CREATE TABLE Beacon
(
    Ident   SMALLINT NOT NULL,
    /* 1 Radar, 2 Ionosonde, 3 others */
    TypeId  SMALLINT NOT NULL, 
    /* Unique code for any particular TypeId value. For Ionosondes, URSI code can be used. */
    Code    VARCHAR(20) NOT NULL,
    Name    VARCHAR(50) NOT NULL,
    /* Latitude */
    Lat          FLOAT NOT NULL,
    /* Longitude */
    Lon          FLOAT NOT NULL,
    Description  VARCHAR(1000),       /* Textual description of this location */
    	CONSTRAINT Beacon_Ident_is_PKey PRIMARY KEY (Ident),
    	CONSTRAINT Beacon_TypeId_is_FKey FOREIGN KEY (TypeId) REFERENCES BeaconType ON DELETE CASCADE ON UPDATE CASCADE,
    	CONSTRAINT Beacon_dCode_is_Not_Empty CHECK (Code <> ''),
		CONSTRAINT Beacon_Code_is_NoSpaces CHECK (CHAR_LENGTH(Code) = CHAR_LENGTH(TRIM(Code))),
    	CONSTRAINT Beacon_TypeId_Code_is_Unique UNIQUE (TypeId,Code),
    	CONSTRAINT Beacon_Name_is_Not_Empty CHECK (Name <> ''),
		CONSTRAINT Beacon_Name_is_NoSpaces CHECK (CHAR_LENGTH(Name) = CHAR_LENGTH(TRIM(Name)))
);
COMMIT;

/***********************************************************************
   Dictionary.
   Types of Tx Station (transmission stations)
   Content corresponds to enum class TxStationType
 ***********************************************************************/
CREATE TABLE TxStationType
(
    Ident   SMALLINT NOT NULL,
    Name    VARCHAR(20) NOT NULL,
    Desc    VARCHAR(1000),
    	CONSTRAINT TxStationType_Ident_is_PKey PRIMARY KEY (Ident),
    	CONSTRAINT TxStationType_Name_is_Not_Empty CHECK (Name <> ''),
		CONSTRAINT TxStationType_Name_is_NoSpaces CHECK (CHAR_LENGTH(Name) = CHAR_LENGTH(TRIM(Name))),
    	CONSTRAINT TxStationType_Name_is_Unique UNIQUE (Name)
);
COMMIT;

/***********************************************************************
   Transmission station which can be Digisondes, Radar, and may be something else  
 ***********************************************************************/
CREATE TABLE TxStation
(
    Ident       SMALLINT NOT NULL,
    TypeId      SMALLINT NOT NULL, /* 1 Digisonde, 2 Beacon */
    StationId   SMALLINT,
    BeaconId    SMALLINT,
    	CONSTRAINT TxStation_Ident_is_PKey PRIMARY KEY (Ident),
    	CONSTRAINT TxStation_TypeId_is_FKey FOREIGN KEY (TypeId) REFERENCES TxStationType ON DELETE CASCADE ON UPDATE CASCADE,
    	CONSTRAINT TxStation_StationId_is_FKey FOREIGN KEY (StationId) REFERENCES Station ON DELETE CASCADE ON UPDATE CASCADE,
    	CONSTRAINT TxStation_BeaconId_is_FKey FOREIGN KEY (BeaconId) REFERENCES Beacon ON DELETE CASCADE ON UPDATE CASCADE,
    	CONSTRAINT TxStation_TypeId_1_OR_2 CHECK (TypeId IN(1,2)),
    	CONSTRAINT TxStation_Exactly_One_Ref CHECK ((TypeId=1 AND StationId IS NOT NULL AND BeaconId IS NULL) 
    	               OR (TypeId=2 AND StationId IS NULL AND BeaconId IS NOT NULL))
);
COMMIT;

/***********************************************************************
   Dictionary.
   TID Processors.
   Note, that Processor has Name and Version, so it can be several Processors with the same name but different versions in this table.
 ***********************************************************************/
CREATE TABLE Processor
(
    Ident      SMALLINT NOT NULL,
    Name       VARCHAR(50) NOT NULL,
    Version    VARCHAR(20) NOT NULL,
    Rank       INTEGER NOT NULL,
    Desc       VARCHAR(1000),
    	CONSTRAINT Processor_Ident_id_PKey PRIMARY KEY (Ident),
    	CONSTRAINT Processor_Name_is_Not_Empty CHECK (Name <> ''),
		CONSTRAINT Processor_Name_is_NoSpaces CHECK (CHAR_LENGTH(Name) = CHAR_LENGTH(TRIM(Name))),
    	CONSTRAINT Processor_Version_is_Not_Empty CHECK (Version <> ''),
    	CONSTRAINT Processor_Version_is_NoSpaces CHECK (CHAR_LENGTH(Version) = CHAR_LENGTH(TRIM(Version))),
    	CONSTRAINT Processor_Name_Ver_is_Unique UNIQUE (Name, Version),
    	CONSTRAINT Processor_Rank_is_Gt0 CHECK (Rank > 0),
    	CONSTRAINT Processor_Rank_is_Unique UNIQUE (Rank)
);
COMMIT;

/***********************************************************
   Table: OptionsGroup, Owner SYSDBA
   Optionsa are divided into Groups with each Group uniquely identified by its Name. So, this table contains names of all Groups of options.      
 **********************************************************/
CREATE TABLE OptionsGroup (
	/* Internal unique identifier for Options Group. Used as Primary Key */
	Ident	INTEGER NOT NULL,
	/* Name of Options' Group Name, case sensitive */
	Name	VARCHAR(120) NOT NULL,
	/* description of group */
	Desc	VARCHAR(1000),
		CONSTRAINT OptionsGroup_Ident_is_PKey PRIMARY KEY (Ident),
    	CONSTRAINT OptionsGroup_Name_is_Unique UNIQUE (Name),
    	CONSTRAINT OptionsGroup_Name_is_Not_Empty CHECK (Name <> ''),
		CONSTRAINT OptionsGroup_Name_is_NoSpaces CHECK (CHAR_LENGTH(Name) = CHAR_LENGTH(TRIM(Name)))
);
COMMIT;

/***********************************************************
   Table: OptionSetHeader, Owner SYSDBA
   This table does not contain the content of options' set (i.e. pairs name-value), but only unique identifier number plus hash code of set,
   plus number of pairs name-value in t it.
 **********************************************************/
CREATE TABLE OptionSetHeader
(
	/* Internal unique identifier for unique set of options. Used as Primary Key */
	Ident		INTEGER NOT NULL,
	/* Hash Code for this set of options */
	HashCode	INTEGER NOT NULL,
	/* Number of options in this set */
	SetSize		INTEGER NOT NULL,
		CONSTRAINT OpHashCode_Ident_is_PKey PRIMARY KEY (Ident),
		CONSTRAINT OpHashCode_SetSize_is_Gt_0 CHECK (SetSize > 0)
);
COMMIT;

/***********************************************************************
   Table: OptionSet, Owner SYSDBA
   When we insert new TID measurement, we try first to find whether the same set of Proc Options already exists in this table,
   and if exists we use it. Otherwice we add the header of new set to table OptionSetHeader, and add all individual triplets 
   OptionsGroupId-Name-Value to this table. 
   Note, that for acceleration "whether set of options exists in this table" we calculate hash code of set and search it in table
   OptionSetHeader, so if given hash code with given SetSize is not found then, obviously, set of options does not exists in the table OptionSet, 
   otherwice we try to find this set in this table checking only sets that have given hash code and SetSize.  
 ********************************* **************************************/
CREATE TABLE OptionSet
(
	OptionSetId	INTEGER NOT NULL,
	OptionsGroupId	INTEGER NOT NULL,
	/* Option's name, case sensitive */
	Name		VARCHAR(50) NOT NULL COLLATE UNICODE,
	/* Option's value, case sensitive */
	Val			VARCHAR(50) NOT NULL COLLATE UNICODE,
    	CONSTRAINT OptionSet_SetIdGrpName_is_PKey PRIMARY KEY (OptionSetId, OptionsGroupId, Name),
    	CONSTRAINT OptionSet_OptsGroupId_is_FKey FOREIGN KEY (OptionsGroupId) REFERENCES OptionsGroup ON DELETE CASCADE ON UPDATE CASCADE,
    	CONSTRAINT OptionSet_Name_is_Not_Empty CHECK (Name <> ''),
		CONSTRAINT OptionSet_Name_is_NoSpaces CHECK (CHAR_LENGTH(Name) = CHAR_LENGTH(TRIM(Name))),
		CONSTRAINT OptionSet_Val_is_NoSpaces CHECK (CHAR_LENGTH(Val) = CHAR_LENGTH(TRIM(Val)))
);
COMMIT;

/*
 * Global Tempoprary Table: GTT_OptionSet, Owner SYSDBA
 * The instance of content is per transaction.
 * This Global Temporary Table is used in procedure that insert calculated TID characteristics (whic is also includes Options used in 
 * Calculations/Algorithms/Processors) 
 */
CREATE GLOBAL TEMPORARY TABLE GTT_OptionSet
(
	/* Name of Options' Group Name, case sensitive */
	GName		VARCHAR(120) NOT NULL COLLATE UNICODE,
	/* Option's name, case sensitive */
	Name		VARCHAR(50) NOT NULL COLLATE UNICODE,
	/* Option's value, case sensitive */
	Val			VARCHAR(50) NOT NULL COLLATE UNICODE,
    	CONSTRAINT GTT_OpSet_GroupName_is_Uniq UNIQUE (GName, Name),
    	CONSTRAINT GTT_OpSet_GName_is_Not_Empty CHECK (GName <> ''),
		CONSTRAINT GTT_OpSet_GName_is_NoSpaces CHECK (CHAR_LENGTH(GName) = CHAR_LENGTH(TRIM(GName))),
    	CONSTRAINT GTT_OpSet_Name_is_Not_Empty CHECK (Name <> ''),
		CONSTRAINT GTT_OpSet_Name_is_NoSpaces CHECK (CHAR_LENGTH(Name) = CHAR_LENGTH(TRIM(Name))),
		CONSTRAINT GTT_OpSet_Val_is_NoSpaces CHECK (CHAR_LENGTH(Val) = CHAR_LENGTH(TRIM(Val)))
);
COMMIT;

/***********************************************************
   Table: Observation, Owner SYSDBA
   Contains calculated TID Observation header. 
   Each TID Observation is derived from series of DOP measurements.  
 **********************************************************/
CREATE TABLE Observation
(
	Ident			INTEGER NOT NULL,
    /* Reference time as UT */
    RefTimeUT   	TIMESTAMP NOT NULL,
    /* Internal ident of Digisonde Station */
    ReceiverId  	SMALLINT NOT NULL,
    /* Internal ident of transmitter, which can be either Digisonde or Beacon */
    TransmitterId  	SMALLINT NOT NULL,
    	CONSTRAINT Observ_Ident_is_PKey PRIMARY KEY (Ident),
    	CONSTRAINT Observ_ReceiverId_is_FKey FOREIGN KEY (ReceiverId) REFERENCES Station ON DELETE CASCADE ON UPDATE CASCADE,
		CONSTRAINT Observ_TimeReceiver_is_unique UNIQUE (RefTimeUT,ReceiverId),
    	CONSTRAINT Observ_TransmitterId_is_FKey FOREIGN KEY (TransmitterId) REFERENCES TxStation ON DELETE CASCADE ON UPDATE CASCADE
);
COMMIT;

/***********************************************************
   Table: ObsProcOpt, Owner SYSDBA
   Bundles three references, refernece to Observation, reference to Processor, and reference to OptionsSet. So any particular Observation 
   may be processed by several different Proccesors and even by different Option sets within the same Processor. The possibility to play
   with Options is for tuning options, as well as for improving improving Processing itself.
   Also contains stuff derived from location and Processor (without using DOP or other "raw" measurement) 
 **********************************************************/
CREATE TABLE ObsProcOpt (
    /* Internal id of Observation Processor */
	Ident		INTEGER NOT NULL,
    /* Internal id of Observation */
	ObservId	INTEGER NOT NULL,
    /* Internal id of processor/algorithm */
    ProcId		SMALLINT NOT NULL,
    /* Id of Options set. So, we do not duplicate the same set of options in table OptionSetHeader */
    OptionSetId	INTEGER NOT NULL,
    SubmissionUT  TIMESTAMP NOT NULL,  /* Time of submitting as UT */
	ELayerGroupPath_km	SMALLINT,
	ELayerGPathFactor	FLOAT,
    	CONSTRAINT ObsProc_Ident_is_PKey PRIMARY KEY (Ident),
    	CONSTRAINT ObsProc_ObservId_is_FKey FOREIGN KEY (ObservId) REFERENCES Observation ON DELETE CASCADE ON UPDATE CASCADE,
    	CONSTRAINT ObsProc_ProcId_is_FKey FOREIGN KEY (ProcId) REFERENCES Processor ON DELETE CASCADE ON UPDATE CASCADE,
		CONSTRAINT ObsProc_ObsProcOpt_is_Unique UNIQUE (ObservId,ProcId,OptionSetId),
		CONSTRAINT ObsProc_OptionSetId_is_FKey FOREIGN KEY (OptionSetId) REFERENCES OptionSetHeader ON DELETE CASCADE ON UPDATE CASCADE,
		CONSTRAINT ObsProc_ELayerGroupPath_km_Ge0 CHECK (ELayerGroupPath_km IS NULL OR ELayerGroupPath_km > 0),
		CONSTRAINT ObsProc_ELayerGPathFactor_Ge1 CHECK (ELayerGPathFactor >= 1)
);
COMMIT;

/***********************************************************
   Table: PolFreq, Owner SYSDBA
   Bundles Polarization and frequency values with reference to ObsProcOpt table. So it might be more than one polarization, frequency per 
   one ObsProcOpt entry/row.
 **********************************************************/
CREATE TABLE PolFreq (
	Ident			INTEGER NOT NULL,
	ObsProcOptId	INTEGER NOT NULL,
	/* 1 - ordinary, 2 - extraordinary */
	Polarization	Polarization NOT NULL,
	Freq_kHz		SMALLINT NOT NULL,
    	CONSTRAINT PolFreq_Ident_is_PKey PRIMARY KEY (Ident),
    	CONSTRAINT PolFreq_ObsProcOptId_is_FKey FOREIGN KEY (ObsProcOptId) REFERENCES ObsProcOpt ON DELETE CASCADE ON UPDATE CASCADE,
    	CONSTRAINT PolFreq_ObsProcOptPolFreqUnique UNIQUE (ObsProcOptId, Polarization, Freq_kHz),
		CONSTRAINT PolFreq_Freq_is_Gt0 CHECK (Freq_kHz > 0)
);
COMMIT;

/***********************************************************
   Table: TrackHeader, Owner SYSDBA
   Contains Track Header data used, together with track signals itself (table Signal), for calculation TID parameters. 
 **********************************************************/
CREATE TABLE TrackHeader
(
	/* unique internal identification number for track */
	Ident			INTEGER NOT NULL,
	PolFreqId		INTEGER NOT NULL,
	/* Time UT of the first signal of this track */
	StartUT			TIMESTAMP NOT NULL,
	Cadence_s		SMALLINT NOT NULL,
	/* as number of points */
	TrackLength 	SMALLINT NOT NULL,
	/* number of interpolated points */
	NumGaps			SMALLINT NOT NULL,
	/* Median Group Path of the track signals */
	MedianPath_km	SMALLINT NOT NULL,
	TIDCharsId		INTEGER,
    AvgFitErr_deg	FLOAT NOT NULL,
    AvgSNR_dB		FLOAT NOT NULL,
    AvgAmp_dB		FLOAT NOT NULL,
    	CONSTRAINT TrackHeader_Ident_is_PKey PRIMARY KEY (Ident),
    	CONSTRAINT TrackHeader_FreqId_is_FKey FOREIGN KEY (PolFreqId) REFERENCES PolFreq ON DELETE CASCADE ON UPDATE CASCADE,
		CONSTRAINT TrackHeader_Cadence_is_Gt0 CHECK (Cadence_s > 0),
		CONSTRAINT TrackHeader_TrackLength_is_Gt0 CHECK (TrackLength > 0),
		CONSTRAINT TrackHeader_NumGaps CHECK (NumGaps >= 0 AND NumGaps < TrackLength - 1),
		CONSTRAINT TrackHeader_MedianPath_km_Ge0 CHECK (MedianPath_km >= 0),
		CONSTRAINT TrackHeader_AvgFitErr_deg_Ge0 CHECK (AvgFitErr_deg >= 0),
		CONSTRAINT TrackHeader_AvgSNR_dB_Ge0 CHECK (AvgSNR_dB >= 0),
		CONSTRAINT TrackHeader_AvgAmp_dB_Gt0 CHECK (AvgAmp_dB > 0)
);
COMMIT;

/***********************************************************
   Table: TIDChars, Owner SYSDBA
   Contains TID characteristics calculated from one Track Signals
 **********************************************************/
CREATE TABLE TIDChars (
	/* unique internal identification number for Track TID Parameters */
	Ident		INTEGER NOT NULL,
	TrackId		INTEGER NOT NULL,
    /* Refence latitude, in degrees, of TID characteristics. It calculated from signal track and usually expected around of mid point */
    RefLat_deg		FLOAT NOT NULL,
    /* Refence longitude, in degrees, of TID characteristics. It calculated from signal track and usually expected around of mid point */
    RefLon_deg		FLOAT NOT NULL,
    /* Refence altitude, in km, of TID characteristics. It calculated from signal track */
    RefAlt_km		FLOAT NOT NULL,
    /* wave period, in seconds */
    WavePeriod_s	FLOAT NOT NULL,
    /* wave length, in m */
    WaveLength_m	FLOAT NOT NULL,
    /* amplitude as density change factor. 0.1 means density change by 10 percents */
    Amplitude_fact	FLOAT NOT NULL,
    /* given as azimuth CW, in degrees */
    WaveDir_deg		FLOAT NOT NULL,
    /* Initial hpase of TID wave at the reference point and reference time */
    InitPhase_deg	FLOAT NOT NULL,
    /* 0 - 100, in % */
    Confidence		FLOAT NOT NULL,
    /* 0 - 100, in % */
    Uncertainty		FLOAT NOT NULL,
    	CONSTRAINT TrkTIDChar_Ident_is_PKey PRIMARY KEY (Ident),
    	CONSTRAINT TrkTIDChar_TrackId_is_FKey FOREIGN KEY (TrackId) REFERENCES TrackHeader ON DELETE CASCADE ON UPDATE CASCADE,
		CONSTRAINT TrkTIDChar_RefLat_deg_Check CHECK (RefLat_deg >= -90 AND RefLat_deg <= 90),
		CONSTRAINT TrkTIDChar_RefLon_deg_Check CHECK (RefLon_deg >= 0 AND RefLon_deg <= 360),
		CONSTRAINT TrkTIDChar_RefAlt_km_Check CHECK (RefAlt_km >= 0 AND RefAlt_km <= 1000),
		CONSTRAINT TrkTIDChar_WavePeriod_s_Ge0 CHECK (WavePeriod_s >= 0),
		CONSTRAINT TrkTIDChar_WaveLength_m_Ge0 CHECK (WaveLength_m >= 0),
		CONSTRAINT TrkTIDChar_Amplitude_fact_Ge1 CHECK (Amplitude_fact >= 1),
		CONSTRAINT TrkTIDChar_WaveDir_deg_Ge0Le360 CHECK (WaveDir_deg >= 0 AND WaveDir_deg <=360),
		CONSTRAINT TrkTIDChar_InitPh_deg_Ge0Le360 CHECK (InitPhase_deg >= 0 AND InitPhase_deg <=360),
		CONSTRAINT TrkTIDChar_Confidence_Check CHECK (Confidence >= 0 AND Confidence <= 100),
		CONSTRAINT TrkTIDChar_Uncertainty_Check CHECK (Uncertainty >= 0 AND Uncertainty <= 100)
);
COMMIT;

ALTER TABLE TrackHeader 
    ADD CONSTRAINT TrackHeader_TIDCharsId_is_FKey FOREIGN KEY (TIDCharsId) REFERENCES TIDChars ON DELETE CASCADE ON UPDATE CASCADE;
	
/***********************************************************
   Table: Signal, Owner SYSDBA
   Contains Tracks' Signals that were used for calculation of TID characteristics
   Each row is the signal that belongs to particular track 
 **********************************************************/
CREATE TABLE Signal
(
	/* unique internal identification number for track this signal belongs to */
	TrackId			INTEGER NOT NULL,
	/* starts from 1 to number of signals in track */
	SeqNumber		SMALLINT NOT NULL,
    Path_km			FLOAT NOT NULL,
    DoppShift_Hz	FLOAT NOT NULL,
    Elev_deg		FLOAT NOT NULL,
    Azim_deg		FLOAT NOT NULL,
    FitErr_deg		FLOAT NOT NULL,
    SNR_dB			FLOAT NOT NULL,
    Amp_dB			FLOAT NOT NULL,
    IsInterpolated	SMALLINT NOT NULL,
    	CONSTRAINT Signal_TrkId_SeqNum_is_PKey PRIMARY KEY (TrackId, SeqNumber),
    	CONSTRAINT Signal_TrkId_is_FKey FOREIGN KEY (TrackId) REFERENCES TrackHeader ON DELETE CASCADE ON UPDATE CASCADE,
    	CONSTRAINT Signal_SeqNumber_is_InRange CHECK (SeqNumber >= 1 AND SeqNumber <= (SELECT T.TrackLength FROM TrackHeader T WHERE T.Ident=TrackId)),
    	CONSTRAINT Signal_Path_km_is_Positive CHECK (Path_km > 0),
    	CONSTRAINT Signal_Elev_deg_Ge0_Le90 CHECK (Elev_deg >= 0 AND Elev_deg <= 90),
    	CONSTRAINT Signal_Azim_deg_Ge0_Le360 CHECK (Azim_deg >= 0 AND Azim_deg <= 360),
    	CONSTRAINT Signal_FitErr_deg_Ge0 CHECK (FitErr_deg >= 0),
    	CONSTRAINT Signal_Interpolated_is_bool CHECK (IsInterpolated = 0 OR IsInterpolated = 1)
);
COMMIT;

/* !!!don't delete this table - it's for connection availability checking!!! */
CREATE TABLE UserReg
(
	Ident          SMALLINT  NOT NULL,
	Count_act      INTEGER   NOT NULL,
	Host_addr      CHAR(20)  NOT NULL,
	App_name       CHAR(20)  NOT NULL,
	App_version    CHAR(20)  NOT NULL,
	Connect_Time   TIMESTAMP NOT NULL,
	U_Name         CHAR(40)  NOT NULL,
	OS_Name        CHAR(20)  NOT NULL,
	OS_Version     CHAR(20)  NOT NULL,
	Host_Name      CHAR(50)  NOT NULL,
	Srv_14 TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
	CONSTRAINT UserReg_Ident_Is_PKey PRIMARY KEY (Ident),
    CONSTRAINT UserReg_H_Addr_A_Name_Is_Unique UNIQUE (Host_Addr, App_Name)
);
COMMIT;

CREATE VIEW vObs AS
SELECT o.Ident AS ObsId, RefTimeUT, RefLat_deg, RefLon_deg, refAlt_km, WavePeriod_s, WaveLength_m, Amplitude_fact, WaveDir_deg, InitPhase_deg, 
       Confidence, Uncertainty, 
       rs.UrsiCode AS UrsiCode, rs.Name AS RxName, rs.lat AS RxLat, rs.lon AS RxLon, tx.TypeId AS TxTypeId, 
       IIF(tx.TypeId=1, ts.UrsiCode, b.Code) AS TxCode, IIF(tx.TypeId=1, NULL, bt.Name) AS BeaconTypeName,   
       IIF(tx.TypeId=1, ts.Name, b.Name) AS TxName, IIF(tx.TypeId=1, ts.lat, b.lat) AS TxLat, IIF(tx.TypeId=1, ts.lon, b.lon) AS TxLon, 
       p.Ident AS ProcId, p.Name AS ProcName, p.Version AS ProcVersion, p.Rank, opo.OptionSetId AS OptionSetId, th.Ident AS TrackId, 
       pf.Ident AS PolFreqId, Polarization, Freq_kHz, StartUT, Cadence_s, trackLength, NumGaps, SubmissionUT, ELayerGroupPath_km, 
       ELayerGPathFactor, MedianPath_km, AvgFitErr_deg, AvgSNR_dB, AvgAmp_dB,
       IIF(ELayerGroupPath_km IS NULL OR MedianPath_km > ELayerGPathFactor * ELayerGroupPath_km, 
           128.0 - CAST(NumGaps AS FLOAT)/TrackLength + CAST(TrackLength AS FLOAT)/64.0, 64.0 - CAST(NumGaps AS FLOAT)/TrackLength + CAST(TrackLength AS FLOAT)/64.0) AS Quality
    FROM Observation o 
       JOIN Station rs ON o.ReceiverId=rs.Ident
       JOIN ObsProcOpt opo ON opo.ObservId=o.Ident
       JOIN Processor p ON opo.ProcId=p.Ident
       JOIN PolFreq pf ON pf.ObsProcOptId=opo.Ident
       JOIN TrackHeader th ON th.PolFreqId=pf.Ident
       JOIN TIDChars tc ON tc.TrackId=th.Ident
       JOIN TxStation tx ON tx.Ident=o.TransmitterId 
       LEFT JOIN Station ts ON tx.TypeId=1 AND ts.Ident=tx.StationId
       LEFT JOIN Beacon b ON tx.TypeId=2 AND b.Ident=tx.BeaconId
       LEFT JOIN BeaconType bt ON b.TypeId=bt.Ident;

/*
 * Global Tempoprary Table: GTT_ObsProc, Owner SYSDBA
 * The instance of content is per transaction.
 * This Global Temporary Table is used in procedure that read all (part of) processings that have different options for particular 
 * observation-processor and have to choose the best ones/ones.
 */
CREATE GLOBAL TEMPORARY TABLE GTT_ObsProc
(
	RefLat_deg	FLOAT NOT NULL,
	RefLon_deg	FLOAT NOT NULL,
	RefAlt_km	FLOAT NOT NULL,
	WavePeriod_s	FLOAT NOT NULL,
	WaveLength_m	FLOAT NOT NULL,
	Amplitude_fact	FLOAT NOT NULL,
	WaveDir_deg		FLOAT NOT NULL,
	InitPhase_deg	FLOAT NOT NULL, 
	Confidence		FLOAT NOT NULL,
	Uncertainty		FLOAT NOT NULL,
	OptionSetId	INTEGER NOT NULL,
	Polarization	Polarization NOT NULL,
	Freq_kHz	SMALLINT NOT NULL,
	TrackId		INTEGER NOT NULL,
	StartUT		TIMESTAMP NOT NULL,
	Cadence_s	SMALLINT NOT NULL,
	trackLength	SMALLINT NOT NULL,
	NumGaps		SMALLINT NOT NULL,
	SubmissionUT	TIMESTAMP NOT NULL,
	ELayerGroupPath_km	SMALLINT NOT NULL,
	ELayerGPathFactor	FLOAT NOT NULL,
	MedianPath_km	SMALLINT NOT NULL,
	AvgFitErr_deg	FLOAT NOT NULL,
	AvgSNR_dB	FLOAT NOT NULL,
	AvgAmp_dB	FLOAT NOT NULL,
	Quality		FLOAT NOT NULL,
    	CONSTRAINT GTT_vObs_TrackIdUnique UNIQUE (TrackId)
);
COMMIT;
       
       
CREATE EXCEPTION Ex_SaveOptions 'Error in SaveOptions Procedure';       
CREATE EXCEPTION Ex_ParseOptions 'Error in ParseOptions Procedure';       
CREATE EXCEPTION Ex_GetTransitterId 'Error in GetTransitterId Procedure';       
CREATE EXCEPTION Ex_AddOrUpdateProcessing 'Error in AddOrUpdateProcessing Procedure';       
CREATE EXCEPTION Ex_AddTrackSignals 'Error in AddTrackSignals Procedure';       
       
/* Create procedures */

SET TERM !! ;

/*
 * Calculates hash code of given string as integer number. The algorithm exactly the same as presently in Java String. 
 */
RECREATE PROCEDURE HashCode(str VARCHAR(8191))
RETURNS (hash INTEGER)
AS
DECLARE VARIABLE len INTEGER;
DECLARE VARIABLE p INTEGER;
DECLARE VARIABLE hashlong BIGINT;
BEGIN
    hashlong = 0;
    p = 1;
    len = CHAR_LENGTH(str);
    WHILE (p <= len)
    DO
    BEGIN
        hashlong = 31 * hashlong + ASCII_VAL(SUBSTRING(str FROM p FOR 1));
        hashlong = BIN_SHR(BIN_SHL(hashlong, 32), 32);
        p = p + 1;
    END
    hash = hashlong;
    SUSPEND;
END!!

/*
 * Returns Processor ID for given timeUT and receiver station ursi code than has maximum Rank.
 */
RECREATE PROCEDURE GetBestProcId(timeUT TIMESTAMP, ursiCode VARCHAR(10))
    RETURNS (procId INTEGER)
AS
BEGIN
	SELECT DISTINCT Ident 
	    FROM Processor 
	    WHERE Rank = (SELECT MAX(p.Rank) 
	                      FROM Processor p 
			              JOIN ObsProcOpt opo ON opo.ProcId=p.Ident 
			              JOIN Observation o ON opo.ObservId=o.Ident 
			              JOIN Station s ON s.Ident=o.ReceiverId 
		                  WHERE o.RefTimeUT=:timeUT AND s.UrsiCode=:ursiCode)
		INTO :procId;
	SUSPEND;
END!!

/*
 * Returns 1 if Processor with given name and version exists for given timeUT and receiver station ursi code.
 * Otherwise returns 0.
 */
RECREATE PROCEDURE existProcessor(timeUT TIMESTAMP, ursiCode VARCHAR(10), procName VARCHAR(50), procVersion VARCHAR(20))
    RETURNS (existFlag SMALLINT)
AS
DECLARE VARIABLE ident INTEGER;
BEGIN
    existFlag = 0;
    ident = -1;
	SELECT p.Ident FROM Processor p 
			JOIN ObsProcOpt opo ON opo.ProcId=p.Ident 
			JOIN Observation o ON opo.ObservId=o.Ident 
			JOIN Station s ON s.Ident=o.ReceiverId 
		WHERE o.RefTimeUT=:timeUT AND s.UrsiCode=:ursiCode AND p.Name=:procName AND p.Version=:procVersion
		INTO :ident;
    IF (ident >= 0)
    THEN existFlag = 1;
    SUSPEND;
END!!

/*
 * Returns exactly one row per observation from view vObs that corresponds to the best track of signals of highest Processing applied to observation
 */
RECREATE PROCEDURE getBestForHighestProc
    RETURNS (RefTimeUT TIMESTAMP, UrsiCode VARCHAR(20), RxName VARCHAR(50), 
             RxLat FLOAT, RxLon FLOAT, TxTypeId SMALLINT, TxCode VARCHAR(20), BeaconTypeName VARCHAR(50), TxName VARCHAR(50), TxLat FLOAT, TxLon FLOAT, 
             RefLat_deg FLOAT, RefLon_deg FLOAT, RefAlt_km FLOAT, WavePeriod_s FLOAT, WaveLength_m FLOAT,         
             Amplitude_fact FLOAT, WaveDir_deg FLOAT, InitPhase_deg FLOAT, Confidence FLOAT, Uncertainty FLOAT, ProcId SMALLINT,        
             ProcName VARCHAR(50), ProcVersion VARCHAR(20), OptionSetId INTEGER, polarization Polarization, Freq_kHz SMALLINT, TrackId INTEGER, 
             StartUT TIMESTAMP, Cadence_s SMALLINT, trackLength SMALLINT, NumGaps SMALLINT, ELayerGroupPath_km SMALLINT, 
             ELayerGPathFactor SMALLINT, MedianPath_km SMALLINT, AvgFitErr_deg FLOAT, AvgSNR_dB FLOAT, AvgAmp_dB FLOAT, Quality FLOAT             
    )
AS
BEGIN
    FOR SELECT RefTimeUT, UrsiCode, RxName, RxLat, RxLon, TxTypeId, TxCode, BeaconTypeName, TxName, TxLat, TxLon, RefLat_deg, RefLon_deg, refAlt_km, WavePeriod_s, 
               WaveLength_m, Amplitude_fact, WaveDir_deg, InitPhase_deg, Confidence, Uncertainty, ProcId, ProcName, ProcVersion, OptionSetId, 
               Polarization, Freq_kHz, TrackId, StartUT, Cadence_s, trackLength, NumGaps, ELayerGroupPath_km, ELayerGPathFactor, MedianPath_km, 
               AvgFitErr_deg, AvgSNR_dB, AvgAmp_dB, Quality
        FROM vObs v
        WHERE 
            v.TrackId = (SELECT MAX(v1.TrackId) FROM vObs v1 
                           WHERE v1.RefTimeUT=v.RefTimeUT AND v1.UrsiCode=v.UrsiCode AND 
                                 v1.ProcId=(SELECT ProcId FROM GetBestProcId(v.RefTimeUT, v.UrsiCode)) AND
                                 v1.Quality = 
                                     (SELECT MAX(v2.Quality) FROM vObs v2 
                                         WHERE v2.RefTimeUT=v.RefTimeUT AND v2.UrsiCode=v.UrsiCode AND 
                                               v2.ProcId=(SELECT ProcId FROM GetBestProcId(v.RefTimeUT, v.UrsiCode))
                                     )
                        )
	INTO :RefTimeUT, :UrsiCode, :RxName, :RxLat, :RxLon, :TxTypeId, :TxCode, :BeaconTypeName, :TxName, :TxLat, :TxLon, 
	     :RefLat_deg, :RefLon_deg, :refAlt_km, :WavePeriod_s, :WaveLength_m, :Amplitude_fact, :WaveDir_deg, :InitPhase_deg, :Confidence, 
	     :Uncertainty, :ProcId, :ProcName, :ProcVersion, :OptionSetId, :polarization, :Freq_kHz,   
         :TrackId, :startUT, :Cadence_s, :trackLength, :NumGaps, :ELayerGroupPath_km, :ELayerGPathFactor, :MedianPath_km, 
         :avgFitErr_deg,:avgSNR_dB, :avgAmp_dB, :Quality
    DO SUSPEND;
END!!

/*
 * Returns no more than one row per observation from view vObs that corresponds to the best track of signals of given Processing applied to observation.
 * Processing is given by Processing name and version.
 */
RECREATE PROCEDURE getBestForGivenProc(ProcName VARCHAR(50), ProcVersion VARCHAR(20))
    RETURNS (RefTimeUT TIMESTAMP, UrsiCode VARCHAR(20), RxName VARCHAR(50), RxLat FLOAT, RxLon FLOAT, TxTypeId SMALLINT, 
             TxCode VARCHAR(20), BeaconTypeName VARCHAR(50), TxName VARCHAR(50), TxLat FLOAT, TxLon FLOAT, 
             RefLat_deg FLOAT, RefLon_deg FLOAT, RefAlt_km FLOAT, WavePeriod_s FLOAT, WaveLength_m FLOAT, Amplitude_fact FLOAT, 
             WaveDir_deg FLOAT, InitPhase_deg FLOAT, Confidence FLOAT, Uncertainty FLOAT, ProcId SMALLINT, OptionSetId INTEGER, 
             polarization Polarization, Freq_kHz SMALLINT, TrackId INTEGER, StartUT TIMESTAMP, Cadence_s SMALLINT, trackLength SMALLINT, 
             NumGaps SMALLINT, ELayerGroupPath_km SMALLINT, ELayerGPathFactor SMALLINT, MedianPath_km SMALLINT, AvgFitErr_deg FLOAT, 
             AvgSNR_dB FLOAT, AvgAmp_dB FLOAT, Quality FLOAT             
    )
AS
BEGIN
    FOR SELECT RefTimeUT, UrsiCode, RxName, RxLat, RxLon, TxTypeId, TxCode, BeaconTypeName, TxName, TxLat, TxLon, 
    	       RefLat_deg, RefLon_deg, refAlt_km, WavePeriod_s, WaveLength_m, Amplitude_fact, WaveDir_deg, InitPhase_deg, Confidence, 
    	       Uncertainty, ProcId, OptionSetId, Polarization, Freq_kHz,  
               TrackId, StartUT, Cadence_s, trackLength, NumGaps, ELayerGroupPath_km, ELayerGPathFactor, MedianPath_km, 
               AvgFitErr_deg, AvgSNR_dB, AvgAmp_dB, Quality 
        FROM vObs v
        WHERE 
        v.TrackId = (SELECT MAX(v1.TrackId) FROM vObs v1 
                     WHERE v1.RefTimeUT=v.RefTimeUT AND v1.UrsiCode=v.UrsiCode AND v1.ProcName=:ProcName AND v1.ProcVersion=:ProcVersion AND 
                         v1.Quality = (SELECT MAX(v2.Quality) FROM vObs v2 WHERE v2.RefTimeUT=v.RefTimeUT AND v2.UrsiCode=v.UrsiCode AND v2.ProcName=:ProcName AND v2.ProcVersion=:ProcVersion)
                    )
	INTO :RefTimeUT, :UrsiCode, :RxName, :RxLat, :RxLon, :TxTypeId, :TxCode, :BeaconTypeName, :TxName, :TxLat, :TxLon, 
	     :RefLat_deg, :RefLon_deg, :refAlt_km, :WavePeriod_s, :WaveLength_m, :Amplitude_fact, :WaveDir_deg, :InitPhase_deg, :Confidence, 
	     :Uncertainty, :ProcId, :OptionSetId, :polarization, :Freq_kHz,  
         :TrackId, :startUT, :Cadence_s, :trackLength, :NumGaps, :ELayerGroupPath_km, :ELayerGPathFactor, :MedianPath_km, 
         :avgFitErr_deg, :avgSNR_dB, :avgAmp_dB, :Quality
    DO SUSPEND;
END!!

/*
 * Returns exactly one row per observation from view vObs that corresponds to the best track of signals of given Processing applied to observation 
 * or, if given Processing is not applied, of the highest Processing applied to observation.
 * Processing is given by Processing name and version.
 */
RECREATE PROCEDURE getBestForGivenOrHighestProc(ProcName VARCHAR(50), ProcVersion VARCHAR(20))
    RETURNS (RefTimeUT TIMESTAMP, UrsiCode VARCHAR(20), RxName VARCHAR(50), RxLat FLOAT, RxLon FLOAT, TxTypeId SMALLINT, 
             TxCode VARCHAR(20), BeaconTypeName VARCHAR(50), TxName VARCHAR(50), TxLat FLOAT, TxLon FLOAT, 
             RefLat_deg FLOAT, RefLon_deg FLOAT, RefAlt_km FLOAT, WavePeriod_s FLOAT, WaveLength_m FLOAT, Amplitude_fact FLOAT, 
             WaveDir_deg FLOAT, InitPhase_deg FLOAT, Confidence FLOAT, Uncertainty FLOAT, OutProcId SMALLINT, OutProcName VARCHAR(50), 
             OutProcVersion VARCHAR(20), OptionSetId INTEGER, polarization Polarization, Freq_kHz SMALLINT, TrackId INTEGER, 
             StartUT TIMESTAMP, Cadence_s SMALLINT, trackLength SMALLINT, NumGaps SMALLINT, ELayerGroupPath_km SMALLINT, ELayerGPathFactor SMALLINT, 
             MedianPath_km SMALLINT, AvgFitErr_deg FLOAT, AvgSNR_dB FLOAT , AvgAmp_dB FLOAT, Quality FLOAT             
    )
AS
BEGIN
    FOR SELECT RefTimeUT, UrsiCode, RxName, RxLat, RxLon, TxTypeId, TxCode, BeaconTypeName, TxName, TxLat, TxLon, 
               RefLat_deg, RefLon_deg, refAlt_km, WavePeriod_s, WaveLength_m, Amplitude_fact, WaveDir_deg, InitPhase_deg, Confidence, 
               Uncertainty, ProcId AS OutProcId, ProcName AS OutProcName, ProcVersion AS OutProcVersion, OptionSetId, POlarization, Freq_kHz, 
               TrackId, StartUT, Cadence_s, trackLength, NumGaps, ELayerGroupPath_km, ELayerGPathFactor, MedianPath_km, AvgFitErr_deg, 
               AvgSNR_dB, AvgAmp_dB, Quality
    FROM vObs v
    WHERE 
       v.TrackId = (SELECT MAX(v1.TrackId) FROM vObs v1 
                    WHERE v1.RefTimeUT=v.RefTimeUT AND v1.UrsiCode=v.UrsiCode 
                      AND v1.ProcId = IIF((SELECT existFlag FROM existProcessor(v.RefTimeUT, v.UrsiCode, :ProcName, :ProcVersion)) = 1,
                               (SELECT Ident FROM Processor WHERE Name=:ProcName AND Version=:ProcVersion), 
                               (SELECT ProcId FROM GetBestProcId(v.RefTimeUT, v.UrsiCode))
                        )
                      AND v1.Quality = (SELECT MAX(v2.Quality) FROM vObs v2 
                          WHERE v2.RefTimeUT=v.RefTimeUT AND v2.UrsiCode=v.UrsiCode
                             AND v2.ProcId = IIF((SELECT existFlag FROM existProcessor(v.RefTimeUT, v.UrsiCode, :ProcName, :ProcVersion)) = 1,
                                      (SELECT Ident FROM Processor WHERE Name=:ProcName AND Version=:ProcVersion), 
                                      (SELECT ProcId FROM GetBestProcId(v.RefTimeUT, v.UrsiCode))
                               )
                      )
                  )
	INTO :RefTimeUT, :UrsiCode, :RxName, :RxLat, :RxLon, :TxTypeId, :TxCode, :BeaconTypeName, :TxName, :TxLat, :TxLon, 
	     :RefLat_deg, :RefLon_deg, :refAlt_km, :WavePeriod_s, :WaveLength_m, :Amplitude_fact, :WaveDir_deg, :InitPhase_deg, :Confidence, 
	     :Uncertainty, :OutProcId, OutProcName, :OutProcVersion, :OptionSetId, :polarization, :Freq_kHz, :TrackId, :startUT, :Cadence_s, 
	     :trackLength, :NumGaps, :ELayerGroupPath_km, :ELayerGPathFactor, :MedianPath_km, :avgFitErr_deg, :avgSNR_dB, :avgAmp_dB, :Quality
    DO SUSPEND;
END!!

/*
 * Returns best per frequency row from view vObs that corresponds to the best per frequency track of signals of highest Processing applied 
 * to observation.
 * It will be at least one row for each observation, and as many rows as many different frequencies exist for highest Processing applied to
 * given observation. 
 */
RECREATE PROCEDURE getBestPerFreqForHighestProc
    RETURNS (refTimeUT TIMESTAMP, ursiCode VARCHAR(20), rxName VARCHAR(50), rxLat FLOAT, rxLon FLOAT, txTypeId SMALLINT, 
             txCode VARCHAR(20), beaconTypeName VARCHAR(50), txName VARCHAR(50), txLat FLOAT, txLon FLOAT, 
             refLat_deg FLOAT, refLon_deg FLOAT, refAlt_km FLOAT, wavePeriod_s FLOAT, waveLength_m FLOAT,         
             amplitude_fact FLOAT, waveDir_deg FLOAT, initPhase_deg FLOAT, confidence FLOAT, uncertainty FLOAT, procId SMALLINT,        
             procName VARCHAR(50), procVersion VARCHAR(20), optionSetId INTEGER, polarization Polarization, freq_kHz SMALLINT, 
             trackId INTEGER, startUT TIMESTAMP, cadence_s SMALLINT, trackLength SMALLINT, numGaps SMALLINT, submissionUT TIMESTAMP, 
             eLayerGroupPath_km SMALLINT, eLayerGPathFactor SMALLINT, medianPath_km SMALLINT, avgFitErr_deg FLOAT, avgSNR_dB FLOAT, 
             avgAmp_dB FLOAT, quality FLOAT             
    )
AS
DECLARE curRefTimeUT TYPE OF COLUMN Observation.RefTimeUT;
DECLARE curUrsiCode TYPE OF COLUMN Station.UrsiCode;
DECLARE curRxName TYPE OF COLUMN Station.Name;
DECLARE curRxLat TYPE OF COLUMN Station.Lat;
DECLARE curRxLon TYPE OF COLUMN Station.Lon;
DECLARE curTxTypeId TYPE OF COLUMN TxStation.TypeId;
DECLARE curTxCode TYPE OF COLUMN Station.UrsiCode;
DECLARE curBeaconTypeName TYPE OF COLUMN BeaconType.Name; 
DECLARE curTxName TYPE OF COLUMN Station.Name;
DECLARE curTxLat TYPE OF COLUMN Station.Lat;
DECLARE curTxLon TYPE OF COLUMN Station.Lon;
DECLARE curRefLat_deg TYPE OF COLUMN TIDChars.RefLat_deg;
DECLARE curRefLon_deg TYPE OF COLUMN TIDChars.RefLon_deg;
DECLARE curRefAlt_km TYPE OF COLUMN TIDChars.RefAlt_km;
DECLARE curWavePeriod_s TYPE OF COLUMN TIDChars.WavePeriod_s;
DECLARE curWaveLength_m TYPE OF COLUMN TIDChars.WaveLength_m;
DECLARE curAmplitude_fact TYPE OF COLUMN TIDChars.Amplitude_fact;
DECLARE curWaveDir_deg TYPE OF COLUMN TIDChars.WaveDir_deg;
DECLARE curInitPhase_deg TYPE OF COLUMN TIDChars.InitPhase_deg;
DECLARE curConfidence TYPE OF COLUMN TIDChars.Confidence;
DECLARE curUncertainty TYPE OF COLUMN TIDChars.Uncertainty;
DECLARE curProcId TYPE OF COLUMN Processor.Ident;
DECLARE curProcName TYPE OF COLUMN Processor.Name;
DECLARE curProcVersion TYPE OF COLUMN Processor.Version;
DECLARE curOptionSetId TYPE OF COLUMN OptionSet.OptionSetId;
DECLARE curPolarization Polarization;
DECLARE curFreq_kHz TYPE OF COLUMN PolFreq.Freq_kHz;
DECLARE curTrackId TYPE OF COLUMN TrackHeader.Ident;
DECLARE curStartUT TYPE OF COLUMN TrackHeader.StartUT; 
DECLARE curCadence_s TYPE OF COLUMN TrackHeader.Cadence_s;
DECLARE curTrackLength TYPE OF COLUMN TrackHeader.TrackLength;
DECLARE curNumGaps TYPE OF COLUMN TrackHeader.NumGaps; 
DECLARE curSubmissionUT TYPE OF COLUMN ObsProcOpt.SubmissionUT; 
DECLARE curELayerGroupPath_km TYPE OF COLUMN ObsProcOpt.ELayerGroupPath_km; 
DECLARE curELayerGPathFactor TYPE OF COLUMN ObsProcOpt.ELayerGPathFactor;
DECLARE curMedianPath_km TYPE OF COLUMN TrackHeader.MedianPath_km;
DECLARE curAvgFitErr_deg TYPE OF COLUMN TrackHeader.AvgFitErr_deg; 
DECLARE curAvgSNR_dB TYPE OF COLUMN TrackHeader.AvgSNR_dB;
DECLARE curAvgAmp_dB TYPE OF COLUMN TrackHeader.AvgAmp_dB;
DECLARE curQuality TYPE OF COLUMN vObs.Quality;
DECLARE bestRefLat_deg TYPE OF COLUMN TIDChars.RefLat_deg;
DECLARE bestRefLon_deg TYPE OF COLUMN TIDChars.RefLon_deg;
DECLARE bestRefAlt_km TYPE OF COLUMN TIDChars.RefAlt_km;
DECLARE bestWavePeriod_s TYPE OF COLUMN TIDChars.WavePeriod_s;
DECLARE bestWaveLength_m TYPE OF COLUMN TIDChars.WaveLength_m;
DECLARE bestAmplitude_fact TYPE OF COLUMN TIDChars.Amplitude_fact;
DECLARE bestWaveDir_deg TYPE OF COLUMN TIDChars.WaveDir_deg;
DECLARE bestInitPhase_deg TYPE OF COLUMN TIDChars.InitPhase_deg;
DECLARE bestConfidence TYPE OF COLUMN TIDChars.Confidence;
DECLARE bestUncertainty TYPE OF COLUMN TIDChars.Uncertainty;
DECLARE bestOptionSetId TYPE OF COLUMN OptionSet.OptionSetId;
DECLARE bestPolarization Polarization;
DECLARE bestFreq_kHz TYPE OF COLUMN PolFreq.Freq_kHz;
DECLARE bestTrackId TYPE OF COLUMN TrackHeader.Ident;
DECLARE bestStartUT TYPE OF COLUMN TrackHeader.StartUT; 
DECLARE bestCadence_s TYPE OF COLUMN TrackHeader.Cadence_s;
DECLARE bestTrackLength TYPE OF COLUMN TrackHeader.TrackLength;
DECLARE bestNumGaps TYPE OF COLUMN TrackHeader.NumGaps; 
DECLARE bestSubmissionUT TYPE OF COLUMN ObsProcOpt.SubmissionUT; 
DECLARE bestELayerGroupPath_km TYPE OF COLUMN ObsProcOpt.ELayerGroupPath_km; 
DECLARE bestELayerGPathFactor TYPE OF COLUMN ObsProcOpt.ELayerGPathFactor;
DECLARE bestMedianPath_km TYPE OF COLUMN TrackHeader.MedianPath_km;
DECLARE bestAvgFitErr_deg TYPE OF COLUMN TrackHeader.AvgFitErr_deg; 
DECLARE bestAvgSNR_dB TYPE OF COLUMN TrackHeader.AvgSNR_dB;
DECLARE bestAvgAmp_dB TYPE OF COLUMN TrackHeader.AvgAmp_dB;
DECLARE bestQuality TYPE OF COLUMN vObs.Quality;
DECLARE prevOptionSetId TYPE OF COLUMN OptionSet.OptionSetId;
DECLARE prevFreq_kHz TYPE OF COLUMN PolFreq.Freq_kHz;
DECLARE maxQuality TYPE OF COLUMN vObs.Quality;
DECLARE maxTrackId TYPE OF COLUMN TrackHeader.Ident;
DECLARE maxAvgQuality TYPE OF COLUMN vObs.Quality;
BEGIN
	DELETE FROM GTT_ObsProc;
	ursiCode = '';
    FOR SELECT RefTimeUT, UrsiCode, RxName, RxLat, RxLon, TxTypeId, TxCode, BeaconTypeName, TxName, TxLat, TxLon, 
               RefLat_deg, RefLon_deg, refAlt_km, WavePeriod_s, WaveLength_m, Amplitude_fact, WaveDir_deg, InitPhase_deg, Confidence, 
               Uncertainty, ProcId, ProcName, ProcVersion, OptionSetId, Polarization, Freq_kHz,  
               TrackId, StartUT, Cadence_s, trackLength, NumGaps, SubmissionUT, ELayerGroupPath_km, ELayerGPathFactor, MedianPath_km, 
               AvgFitErr_deg, AvgSNR_dB, AvgAmp_dB, Quality 
    FROM vObs v
    WHERE ProcId=(SELECT ProcId FROM GetBestProcId(RefTimeUT, UrsiCode))
    ORDER BY UrsiCode, RefTimeUT, ProcId, OptionSetId, Freq_kHz
	INTO :curRefTimeUT, :curUrsiCode, :curRxName, :curRxLat, :curRxLon, :curTxTypeId, :curTxCode, :curBeaconTypeName, :curTxName, :curTxLat, 
	     :curTxLon, :curRefLat_deg, :curRefLon_deg, :curRefAlt_km, :curWavePeriod_s, :curWaveLength_m, :curAmplitude_fact, :curWaveDir_deg, 
	     :curInitPhase_deg, :curConfidence, :curUncertainty, :curProcId, :curProcName, :curProcVersion, :curOptionSetId, :curPolarization, 
	     :curFreq_kHz, :curTrackId, :curStartUT, :curCadence_s, :curTrackLength, :curNumGaps, :curSubmissionUT, :curELayerGroupPath_km, 
	     :curELayerGPathFactor, :curMedianPath_km, :curAvgFitErr_deg, :curAvgSNR_dB, :curAvgAmp_dB, :curQuality
    DO
    BEGIN
	    IF (curUrsiCode != ursiCode OR curRefTimeUT != refTimeUT OR curProcId != procId) THEN
	    BEGIN
		    IF (ursiCode != '') THEN
		    BEGIN
			    INSERT INTO GTT_ObsProc 
			    	(RefLat_deg,RefLon_deg,RefAlt_km,WavePeriod_s,WaveLength_m,Amplitude_fact,WaveDir_deg,InitPhase_deg,
						Confidence,Uncertainty,OptionSetId,Polarization,Freq_kHz,TrackId,StartUT,Cadence_s,trackLength,NumGaps,SubmissionUT,
						ELayerGroupPath_km,ELayerGPathFactor,MedianPath_km,AvgFitErr_deg,AvgSNR_dB,AvgAmp_dB,Quality) 
					VALUES
					(:bestRefLat_deg,:bestRefLon_deg,:bestRefAlt_km,:bestWavePeriod_s,:bestWaveLength_m,:bestAmplitude_fact,:bestWaveDir_deg,
						:bestInitPhase_deg,:bestConfidence,:bestUncertainty,:bestOptionSetId,:bestPolarization,:bestFreq_kHz,:bestTrackId,
						:bestStartUT,:bestCadence_s,:bestTrackLength,:bestNumGaps,:bestSubmissionUT,:bestELayerGroupPath_km,:bestELayerGPathFactor,
						:bestMedianPath_km,:bestAvgFitErr_deg,:bestAvgSNR_dB,:bestAvgAmp_dB,:bestQuality);
			    SELECT FIRST 1 AVG(quality) FROM GTT_ObsProc GROUP BY OptionSetId ORDER BY 1 DESC INTO :maxAvgQuality; 
			    FOR SELECT RefLat_deg, RefLon_deg, RefAlt_km, WavePeriod_s, WaveLength_m, Amplitude_fact, WaveDir_deg, InitPhase_deg, Confidence,
						Uncertainty, OptionSetId, Polarization, Freq_kHz, TrackId, StartUT, Cadence_s, trackLength, NumGaps,
						SubmissionUT, ELayerGroupPath_km, ELayerGPathFactor, MedianPath_km, AvgFitErr_deg, AvgSNR_dB, AvgAmp_dB, Quality
					FROM GTT_ObsProc
					WHERE OptionSetId = (SELECT MAX(v1.OptionSetId) 
											FROM GTT_ObsProc v1
											WHERE (SELECT SUM(v2.Quality) FROM GTT_ObsProc v2 WHERE v2.OptionSetId=v1.OptionSetId) = :maxAvgQuality)
					INTO :refLat_deg, :refLon_deg, :refAlt_km, :wavePeriod_s, :waveLength_m, :amplitude_fact, :waveDir_deg, :initPhase_deg,
						:confidence, :uncertainty, :optionSetId, :polarization, :freq_kHz, :trackId, :startUT, :cadence_s, :trackLength, :numGaps, 
						:submissionUT, :eLayerGroupPath_km, :eLayerGPathFactor, :medianPath_km, :avgFitErr_deg, :avgSNR_dB, :avgAmp_dB, :quality
				DO
    			SUSPEND;
		    END 
			DELETE FROM GTT_ObsProc;
			ursiCode = curUrsiCode;
			refTimeUT = curRefTimeUT;
			procId = curProcId;
			procName = curProcName;
			procVersion = curProcVersion;
			RxName = curRxName;
			RxLat = curRxLat;
			RxLon = curRxLon;
			TxTypeId = curTxTypeId;
			TxCode = curTxCode;
			BeaconTypeName = curBeaconTypeName;
			TxName = curTxName;
			TxLat = curTxLat;
			TxLon = curTxLon;
			prevOptionSetId = -1;
			prevFreq_kHz = -1;
	    END
	    IF (curOptionSetId != prevOptionSetId OR curFreq_kHz != prevFreq_kHz) THEN
	    BEGIN
		    IF (prevOptionSetId != -1) THEN
		    BEGIN
			    INSERT INTO GTT_ObsProc 
			    	(RefLat_deg,RefLon_deg,RefAlt_km,WavePeriod_s,WaveLength_m,Amplitude_fact,WaveDir_deg,InitPhase_deg,
						Confidence,Uncertainty,OptionSetId,Polarization,Freq_kHz,TrackId,StartUT,Cadence_s,trackLength,NumGaps,SubmissionUT,
						ELayerGroupPath_km,ELayerGPathFactor,MedianPath_km,AvgFitErr_deg,AvgSNR_dB,AvgAmp_dB,Quality) 
					VALUES
					(:bestRefLat_deg,:bestRefLon_deg,:bestRefAlt_km,:bestWavePeriod_s,:bestWaveLength_m,:bestAmplitude_fact,:bestWaveDir_deg,
						:bestInitPhase_deg,:bestConfidence,:bestUncertainty,:bestOptionSetId,:bestPolarization,:bestFreq_kHz,:bestTrackId,
						:bestStartUT,:bestCadence_s,:bestTrackLength,:bestNumGaps,:bestSubmissionUT,:bestELayerGroupPath_km,:bestELayerGPathFactor,
						:bestMedianPath_km,:bestAvgFitErr_deg,:bestAvgSNR_dB,:bestAvgAmp_dB,:bestQuality);
		    END
		    prevOptionSetId = curOptionSetId;
		    prevFreq_kHz = curFreq_kHz;
			maxQuality = -1;
	    END
	    IF (curQuality > maxQuality OR (curQuality = maxQuality AND curTrackId > maxTrackId)) THEN
	    BEGIN
		    maxQuality = curQuality;
		    maxTrackId = curTrackId;
			bestRefLat_deg = curRefLat_deg;
			bestRefLon_deg = curRefLon_deg;
			bestRefAlt_km = curRefAlt_km;
			bestWavePeriod_s = curWavePeriod_s;
			bestWaveLength_m = curWaveLength_m;
			bestAmplitude_fact = curAmplitude_fact;
			bestWaveDir_deg = curWaveDir_deg;
			bestInitPhase_deg = curInitPhase_deg;
			bestConfidence = curConfidence;
			bestUncertainty = curUncertainty;
			bestOptionSetId = curOptionSetId;
			bestPolarization = curPolarization;
			bestFreq_kHz = curFreq_kHz;
			bestTrackId = curTrackId;
			bestStartUT = curStartUT;
			bestCadence_s = curCadence_s;
			bestTrackLength = curTrackLength;
			bestNumGaps = curNumGaps;
			bestSubmissionUT = curSubmissionUT;
			bestELayerGroupPath_km = curELayerGroupPath_km;
			bestELayerGPathFactor = curELayerGPathFactor;
			bestMedianPath_km = curMedianPath_km;
			bestAvgFitErr_deg = curAvgFitErr_deg;
			bestAvgSNR_dB = curAvgSNR_dB;
			bestAvgAmp_dB = curAvgAmp_dB;
			bestQuality = curQuality;
	    END 
    END
	IF (ursiCode != '') THEN
	BEGIN
	    INSERT INTO GTT_ObsProc 
	    	(RefLat_deg,RefLon_deg,RefAlt_km,WavePeriod_s,WaveLength_m,Amplitude_fact,WaveDir_deg,InitPhase_deg,
				Confidence,Uncertainty,OptionSetId,Polarization,Freq_kHz,TrackId,StartUT,Cadence_s,trackLength,NumGaps,SubmissionUT,
				ELayerGroupPath_km,ELayerGPathFactor,MedianPath_km,AvgFitErr_deg,AvgSNR_dB,AvgAmp_dB,Quality) 
			VALUES
			(:bestRefLat_deg,:bestRefLon_deg,:bestRefAlt_km,:bestWavePeriod_s,:bestWaveLength_m,:bestAmplitude_fact,:bestWaveDir_deg,
				:bestInitPhase_deg,:bestConfidence,:bestUncertainty,:bestOptionSetId,:bestPolarization,:bestFreq_kHz,:bestTrackId,
				:bestStartUT,:bestCadence_s,:bestTrackLength,:bestNumGaps,:bestSubmissionUT,:bestELayerGroupPath_km,:bestELayerGPathFactor,
				:bestMedianPath_km,:bestAvgFitErr_deg,:bestAvgSNR_dB,:bestAvgAmp_dB,:bestQuality);
	    SELECT FIRST 1 SUM(quality) FROM GTT_ObsProc GROUP BY OptionSetId ORDER BY 1 DESC INTO :maxAvgQuality; 
	    FOR SELECT RefLat_deg, RefLon_deg, RefAlt_km, WavePeriod_s, WaveLength_m, Amplitude_fact, WaveDir_deg, InitPhase_deg, Confidence,
				Uncertainty, OptionSetId, Polarization, Freq_kHz, TrackId, StartUT, Cadence_s, trackLength, NumGaps,
				SubmissionUT, ELayerGroupPath_km, ELayerGPathFactor, MedianPath_km, AvgFitErr_deg, AvgSNR_dB, AvgAmp_dB, Quality
			FROM GTT_ObsProc
			WHERE OptionSetId = (SELECT MAX(v1.OptionSetId) 
									FROM GTT_ObsProc v1
									WHERE (SELECT SUM(v2.Quality) FROM GTT_ObsProc v2 WHERE v2.OptionSetId=v1.OptionSetId) = :maxAvgQuality)
			INTO :refLat_deg, :refLon_deg, :refAlt_km, :wavePeriod_s, :waveLength_m, :amplitude_fact, :waveDir_deg, :initPhase_deg,
				:confidence, :uncertainty, :optionSetId, :polarization, :freq_kHz, :trackId, :startUT, :cadence_s, :trackLength, :numGaps, 
				:submissionUT, :eLayerGroupPath_km, :eLayerGPathFactor, :medianPath_km, :avgFitErr_deg, :avgSNR_dB, :avgAmp_dB, :quality
		DO
    	SUSPEND;
	END 
END!!

/*RECREATE PROCEDURE getBestPerFreqForHighestProc
    RETURNS (RefTimeUT TIMESTAMP, UrsiCode VARCHAR(20), RxName VARCHAR(50), RxLat FLOAT, RxLon FLOAT, TxTypeId SMALLINT, 
             TxCode VARCHAR(20), BeaconTypeName VARCHAR(50), TxName VARCHAR(50), TxLat FLOAT, TxLon FLOAT, 
             RefLat_deg FLOAT, RefLon_deg FLOAT, RefAlt_km FLOAT, WavePeriod_s FLOAT, WaveLength_m FLOAT,         
             Amplitude_fact FLOAT, WaveDir_deg FLOAT, InitPhase_deg FLOAT, Confidence FLOAT, Uncertainty FLOAT, ProcId SMALLINT,        
             ProcName VARCHAR(50), ProcVersion VARCHAR(20), OptionSetId INTEGER, polarization Polarization, Freq_kHz SMALLINT, 
             TrackId INTEGER, StartUT TIMESTAMP, Cadence_s SMALLINT, trackLength SMALLINT, NumGaps SMALLINT, ELayerGroupPath_km SMALLINT, 
             ELayerGPathFactor SMALLINT, MedianPath_km SMALLINT, AvgFitErr_deg FLOAT, AvgSNR_dB FLOAT, AvgAmp_dB FLOAT, Quality FLOAT             
    )
AS
BEGIN
    FOR SELECT RefTimeUT, UrsiCode, RxName, RxLat, RxLon, TxTypeId, TxCode, BeaconTypeName, TxName, TxLat, TxLon, 
               RefLat_deg, RefLon_deg, refAlt_km, WavePeriod_s, WaveLength_m, Amplitude_fact, WaveDir_deg, InitPhase_deg, Confidence, 
               Uncertainty, ProcId, ProcName, ProcVersion, OptionSetId, Polarization, Freq_kHz,  
               TrackId, StartUT, Cadence_s, trackLength, NumGaps, ELayerGroupPath_km, ELayerGPathFactor, MedianPath_km, 
               AvgFitErr_deg, AvgSNR_dB, AvgAmp_dB, Quality 
    FROM vObs v
    WHERE 
       v.TrackId = (SELECT MAX(v1.TrackId) FROM vObs v1 
                    WHERE v1.RefTimeUT=v.RefTimeUT AND v1.UrsiCode=v.Ursicode 
                      AND v1.ProcId=(SELECT ProcId FROM GetBestProcId(v.RefTimeUT, v.UrsiCode))
                      AND v1.OptionSetId=v.OptionSetId AND v1.PolFreqId=v.PolFreqId
                      AND v1.Quality = (SELECT MAX(v2.Quality) FROM vObs v2 
                                           WHERE v2.RefTimeUT=v.RefTimeUT AND v2.UrsiCode=v.UrsiCode 
                                             AND v2.ProcId=(SELECT ProcId FROM GetBestProcId(v.RefTimeUT, v.UrsiCode))
                                             AND v2.OptionSetId=v1.OptionSetId AND v2.PolFreqId=v1.PolFreqId 
                                        )
                  ) 
	INTO :RefTimeUT, :UrsiCode, :RxName, :RxLat, :RxLon, :TxTypeId, :TxCode, :BeaconTypeName, :TxName, :TxLat, :TxLon, 
	     :RefLat_deg, :RefLon_deg, :refAlt_km, :WavePeriod_s, :WaveLength_m, :Amplitude_fact, :WaveDir_deg, :InitPhase_deg, :Confidence, 
	     :Uncertainty, :ProcId, :ProcName, :ProcVersion, :OptionSetId, :polarization, :freq_kHz, 
         :TrackId, :startUT, :Cadence_s, :trackLength, :NumGaps, :ELayerGroupPath_km, :ELayerGPathFactor, :MedianPath_km, 
         :avgFitErr_deg, :avgSNR_dB, :avgAmp_dB, :Quality
    DO SUSPEND;
END!!*/

/*
 * Returns best per frequency row from view vObs that corresponds to the best per frequency track of signals of given Processing applied to observation.
 * Processing is given by Processing name and version. Obviously, there will be nothing returned for observations that does not have given
 * Processing applied to them.
 */
RECREATE PROCEDURE getBestPerFreqForGivenProc(ProcName TYPE OF COLUMN Processor.Name, ProcVersion TYPE OF COLUMN Processor.Version)
    RETURNS (RefTimeUT TIMESTAMP, UrsiCode VARCHAR(20), RxName VARCHAR(50), RxLat FLOAT, RxLon FLOAT, TxTypeId SMALLINT, 
             TxCode VARCHAR(20), BeaconTypeName VARCHAR(50), TxName VARCHAR(50), TxLat FLOAT, TxLon FLOAT, 
             RefLat_deg FLOAT, RefLon_deg FLOAT, RefAlt_km FLOAT, WavePeriod_s FLOAT, WaveLength_m FLOAT, Amplitude_fact FLOAT, 
             WaveDir_deg FLOAT, InitPhase_deg FLOAT, Confidence FLOAT, Uncertainty FLOAT, ProcId SMALLINT, OptionSetId INTEGER, 
             polarization Polarization, Freq_kHz SMALLINT, TrackId INTEGER, StartUT TIMESTAMP, Cadence_s SMALLINT, trackLength SMALLINT, 
             NumGaps SMALLINT, ELayerGroupPath_km SMALLINT, ELayerGPathFactor SMALLINT, MedianPath_km SMALLINT, AvgFitErr_deg FLOAT, 
             AvgSNR_dB FLOAT, AvgAmp_dB FLOAT, Quality FLOAT             
    )
AS
BEGIN
    FOR SELECT RefTimeUT, UrsiCode, RxName, RxLat, RxLon, TxTypeId, TxCode, BeaconTypeName, TxName, TxLat, TxLon, 
               RefLat_deg, RefLon_deg, refAlt_km, WavePeriod_s, WaveLength_m, Amplitude_fact, WaveDir_deg, InitPhase_deg, Confidence, 
               Uncertainty, ProcId, OptionSetId, Polarization, Freq_kHz,   
               TrackId, StartUT, Cadence_s, trackLength, NumGaps, ELayerGroupPath_km, ELayerGPathFactor, MedianPath_km, 
               AvgFitErr_deg, AvgSNR_dB, AvgAmp_dB, Quality
    FROM vObs v
    WHERE 
       v.TrackId = (SELECT MAX(v1.TrackId) FROM vObs v1 
                      WHERE v1.RefTimeUT=v.RefTimeUT AND v1.UrsiCode=v.UrsiCode AND v1.ProcName=:ProcName AND v1.ProcVersion=:ProcVersion AND 
                            v1.OptionSetId=v.OptionSetId AND v1.PolFreqId=v.PolFreqId AND
                            v1.Quality = (SELECT MAX(v2.Quality) FROM vObs v2 
                                            WHERE v2.RefTimeUT=v.RefTimeUT AND v2.UrsiCode=v.UrsiCode AND v2.ProcName=:ProcName AND 
                                                  v2.ProcVersion=:ProcVersion AND v2.OptionSetId=v1.OptionSetId AND v2.PolFreqId=v1.PolFreqId
                                         )
                  )
	INTO :RefTimeUT, :UrsiCode, :RxName, :RxLat, :RxLon, :TxTypeId, :TxCode, :BeaconTypeName, :TxName, :TxLat, :TxLon, 
	     :RefLat_deg, :RefLon_deg, :refAlt_km, :WavePeriod_s, :WaveLength_m, :Amplitude_fact, :WaveDir_deg, :InitPhase_deg, :Confidence, 
	     :Uncertainty, :ProcId, :OptionSetId, :polarization, :freq_kHz,
         :TrackId, :startUT, :Cadence_s, :trackLength, :NumGaps, :ELayerGroupPath_km, :ELayerGPathFactor, :MedianPath_km, 
         :avgFitErr_deg, :avgSNR_dB, :avgAmp_dB, :Quality
    DO SUSPEND;
END!!

/*
 * Returns best per frequency row from view vObs that corresponds to the best per frequency track of signals of given Processing applied to observation 
 * or, if given Processing is not applied, of the highest Processing applied to observation.
 * It will be at least one row for each observation, and as many rows as many different frequencies exist for given or highest Processing 
 * applied to given observation. 
 * Processing is given by Processing name and version.
 */
RECREATE PROCEDURE getBestPFforGivenOrHighestProc(ProcName TYPE OF COLUMN Processor.Name, ProcVersion TYPE OF COLUMN Processor.Version)
    RETURNS (RefTimeUT TIMESTAMP, UrsiCode VARCHAR(20), RxName VARCHAR(50), RxLat FLOAT, RxLon FLOAT, TxTypeId SMALLINT, 
             TxCode VARCHAR(20), BeaconTypeName VARCHAR(50), TxName VARCHAR(50), TxLat FLOAT, TxLon FLOAT, 
             RefLat_deg FLOAT, RefLon_deg FLOAT, RefAlt_km FLOAT, WavePeriod_s FLOAT, WaveLength_m FLOAT, Amplitude_fact FLOAT, 
             WaveDir_deg FLOAT, InitPhase_deg FLOAT, Confidence FLOAT, Uncertainty FLOAT, OutProcId SMALLINT, OutProcName VARCHAR(50), 
             OutProcVersion VARCHAR(20), OptionSetId INTEGER, polarization Polarization, Freq_kHz SMALLINT, TrackId INTEGER, 
             StartUT TIMESTAMP, Cadence_s SMALLINT, trackLength SMALLINT, NumGaps SMALLINT, ELayerGroupPath_km SMALLINT, ELayerGPathFactor SMALLINT, 
             MedianPath_km SMALLINT, AvgFitErr_deg FLOAT, AvgSNR_dB FLOAT, AvgAmp_dB FLOAT, Quality FLOAT             
    )
AS
BEGIN
    FOR SELECT RefTimeUT, UrsiCode, RxName, RxLat, RxLon, TxTypeId, TxCode, BeaconTypeName, TxName, TxLat, TxLon, 
           RefLat_deg, RefLon_deg, refAlt_km, WavePeriod_s, WaveLength_m, Amplitude_fact, WaveDir_deg, InitPhase_deg, Confidence, 
           Uncertainty, ProcId AS OutProcId, ProcName AS OutProcName, ProcVersion AS OutProcVersion, OptionSetId, Polarization, Freq_kHz, 
           TrackId, StartUT, Cadence_s, trackLength, NumGaps, ELayerGroupPath_km, ELayerGPathFactor, MedianPath_km, AvgFitErr_deg, AvgSNR_dB, 
           AvgAmp_dB, Quality 
    FROM vObs v
    WHERE v.TrackId = (SELECT MAX(v1.TrackId) FROM vObs v1 
                         WHERE v1.RefTimeUT=v.RefTimeUT AND v1.UrsiCode=v.UrsiCode 
                           AND v1.ProcId = IIF((SELECT existFlag FROM existProcessor(v.RefTimeUT, v.UrsiCode, :ProcName, :ProcVersion)) = 1,
                                               (SELECT Ident FROM Processor WHERE Name=:ProcName AND Version=:ProcVersion), 
                                               (SELECT ProcId FROM GetBestProcId(v.RefTimeUT, v.UrsiCode)))
                           AND v1.OptionSetId=v.OptionSetId AND v1.PolFreqId=v.PolFreqId
                           AND v1.Quality = (SELECT MAX(v2.Quality) FROM vObs v2 
                                                WHERE v2.RefTimeUT=v.RefTimeUT AND v2.UrsiCode=v.UrsiCode
                                                  AND v2.ProcId = IIF((SELECT existFlag FROM existProcessor(v.RefTimeUT, v.UrsiCode, :ProcName, :ProcVersion)) = 1,
                                                                      (SELECT Ident FROM Processor WHERE Name=:ProcName AND Version=:ProcVersion), 
                                                                      (SELECT ProcId FROM GetBestProcId(v.RefTimeUT, v.UrsiCode)))
                                                  AND v2.OptionSetId=v1.OptionSetId AND v2.Freq_kHz=v1.Freq_Khz
                                            )
                      )
	INTO :RefTimeUT, :UrsiCode, :RxName, :RxLat, :RxLon, :TxTypeId, :TxCode, :BeaconTypeName, :TxName, :TxLat, :TxLon, 
	     :RefLat_deg, :RefLon_deg, :refAlt_km, :WavePeriod_s, :WaveLength_m, :Amplitude_fact, :WaveDir_deg, :InitPhase_deg, :Confidence, 
	     :Uncertainty, :OutProcId, OutProcName, :OutProcVersion, :OptionSetId, :polarization, :Freq_kHz, :TrackId, :startUT, :Cadence_s, 
	     :trackLength, :NumGaps, :ELayerGroupPath_km, :ELayerGPathFactor, :MedianPath_km, :avgFitErr_deg, :avgSNR_dB, :avgAmp_dB, :Quality
    DO SUSPEND;
END!!

/*
 * Returns all rows from view vObs that correspond to track of signals of highest Processing applied 
 * to observation.
 * It will be at least one row for each observation for any observation. 
 */
RECREATE PROCEDURE getAllForHighestProc
    RETURNS (RefTimeUT TIMESTAMP, UrsiCode VARCHAR(20), RxName VARCHAR(50), RxLat FLOAT, RxLon FLOAT, TxTypeId SMALLINT, 
             TxCode VARCHAR(20), BeaconTypeName VARCHAR(50), TxName VARCHAR(50), TxLat FLOAT, TxLon FLOAT, 
             RefLat_deg FLOAT, RefLon_deg FLOAT, RefAlt_km FLOAT, WavePeriod_s FLOAT, WaveLength_m FLOAT, Amplitude_fact FLOAT, 
             WaveDir_deg FLOAT, InitPhase_deg FLOAT, Confidence FLOAT, Uncertainty FLOAT, ProcId SMALLINT,       
             ProcName VARCHAR(50), ProcVersion VARCHAR(20), OptionSetId INTEGER, polarization Polarization, Freq_kHz SMALLINT, 
             TrackId INTEGER, StartUT TIMESTAMP, Cadence_s SMALLINT, trackLength SMALLINT, NumGaps SMALLINT, ELayerGroupPath_km SMALLINT, 
             ELayerGPathFactor SMALLINT, MedianPath_km SMALLINT, AvgFitErr_deg FLOAT, AvgSNR_dB FLOAT, AvgAmp_dB FLOAT, Quality FLOAT             
    )
AS
BEGIN
    FOR SELECT RefTimeUT, UrsiCode, RxName, RxLat, RxLon, TxTypeId, TxCode, BeaconTypeName, TxName, TxLat, TxLon, 
           RefLat_deg, RefLon_deg, refAlt_km, WavePeriod_s, WaveLength_m, Amplitude_fact, WaveDir_deg, InitPhase_deg, Confidence, Uncertainty, 
           ProcId, ProcName, ProcVersion, OptionSetId, Polarization, Freq_kHz,   
           TrackId, StartUT, Cadence_s, trackLength, NumGaps, ELayerGroupPath_km, ELayerGPathFactor, MedianPath_km, 
           AvgFitErr_deg, AvgSNR_dB, AvgAmp_dB, Quality
    FROM vObs
    WHERE ProcId = (SELECT ProcId FROM GetBestProcId(RefTimeUT, UrsiCode))
	INTO :RefTimeUT, :UrsiCode, :RxName, :RxLat, :RxLon, :TxTypeId, :TxCode, :BeaconTypeName, :TxName, :TxLat, :TxLon, 
	     :RefLat_deg, :RefLon_deg, :refAlt_km, :WavePeriod_s, :WaveLength_m, :Amplitude_fact, :WaveDir_deg, :InitPhase_deg, :Confidence, 
	     :Uncertainty, :ProcId, :ProcName, :ProcVersion, :OptionSetId, :polarization, :Freq_kHz,  
         :TrackId, :startUT, :Cadence_s, :trackLength, :NumGaps, :ELayerGroupPath_km, :ELayerGPathFactor, :MedianPath_km, 
         :avgFitErr_deg, :avgSNR_dB, :avgAmp_dB, :Quality
    DO SUSPEND;
END!!

/*
 * Returns all rows from view vObs that correspond to track of signals of given Processing applied to observation 
 * or, if given Processing is not applied, of the highest Processing applied to observation.
 * It will be at least one row for each observation. 
 * Processing is given by Processing name and version.
 */
RECREATE PROCEDURE getAllForGivenOrHighestProc(ProcName VARCHAR(50), ProcVersion VARCHAR(20))
    RETURNS (RefTimeUT TIMESTAMP, UrsiCode VARCHAR(20), RxName VARCHAR(50), RxLat FLOAT, RxLon FLOAT, TxTypeId SMALLINT, 
             TxCode VARCHAR(20), BeaconTypeName VARCHAR(50), TxName VARCHAR(50), TxLat FLOAT, TxLon FLOAT, 
             RefLat_deg FLOAT, RefLon_deg FLOAT, RefAlt_km FLOAT, WavePeriod_s FLOAT, WaveLength_m FLOAT, Amplitude_fact FLOAT, 
             WaveDir_deg FLOAT, InitPhase_deg FLOAT, Confidence FLOAT, Uncertainty FLOAT, OutProcId SMALLINT, OutProcName VARCHAR(50), 
             OutProcVersion VARCHAR(20), OptionSetId INTEGER, polarization Polarization, Freq_kHz SMALLINT, TrackId INTEGER,  
             StartUT TIMESTAMP, Cadence_s SMALLINT, trackLength SMALLINT, NumGaps SMALLINT, ELayerGroupPath_km SMALLINT, 
             ELayerGPathFactor SMALLINT, MedianPath_km SMALLINT, AvgFitErr_deg FLOAT, AvgSNR_dB FLOAT, AvgAmp_dB FLOAT, Quality FLOAT             
    )
AS
BEGIN
    FOR SELECT RefTimeUT, UrsiCode, RxName, RxLat, RxLon, TxTypeId, TxCode, BeaconTypeName, TxName, TxLat, TxLon, 
           RefLat_deg, RefLon_deg, refAlt_km, WavePeriod_s, WaveLength_m, Amplitude_fact, WaveDir_deg, InitPhase_deg, Confidence, Uncertainty, 
           ProcId AS OutProcId, ProcName AS OutProcName, ProcVersion AS OutProcVersion, OptionSetId, Polarization, Freq_kHz, TrackId, StartUT, 
           Cadence_s, trackLength, NumGaps, ELayerGroupPath_km, ELayerGPathFactor, MedianPath_km, AvgFitErr_deg, AvgSNR_dB, AvgAmp_dB, Quality
    FROM vObs
    WHERE ProcId = IIF((SELECT existFlag FROM existProcessor(RefTimeUT, UrsiCode, :ProcName, :ProcVersion)) = 1,
                               (SELECT Ident FROM Processor WHERE Name=:ProcName AND Version=:ProcVersion), 
                               (SELECT ProcId FROM GetBestProcId(RefTimeUT, UrsiCode))
                      )
	INTO :RefTimeUT, :UrsiCode, :RxName, :RxLat, :RxLon, :TxTypeId, :TxCode, :BeaconTypeName, :TxName, :TxLat, :TxLon, 
	     :RefLat_deg, :RefLon_deg, :refAlt_km, :WavePeriod_s, :WaveLength_m, :Amplitude_fact, :WaveDir_deg, :InitPhase_deg, :Confidence, 
	     :Uncertainty, :OutProcId, OutProcName, :OutProcVersion, :OptionSetId, :polarization, :Freq_kHz, :TrackId, :startUT, :Cadence_s, 
	     :trackLength, :NumGaps, :ELayerGroupPath_km, :ELayerGPathFactor, :MedianPath_km, :avgFitErr_deg, :avgSNR_dB, :avgAmp_dB, :Quality
    DO SUSPEND;
END!!

/*
 * Takes digisonde station parameters,
 * ursi code, name, latitude, and longitude
 * and tries to find this station in table Station,
 * 1. if station is found, returns value of field Ident
 * 2. if station is not found, insert this station into the table Station and returns value of field Ident   
 */
RECREATE PROCEDURE GetOrCreateStationId(ursiCode TYPE OF COLUMN Station.UrsiCode, name TYPE OF COLUMN Station.Name,
    lat TYPE OF COLUMN Station.Lat, lon TYPE OF COLUMN Station.Lon)
RETURNS (ident SMALLINT)
AS
BEGIN
    ident = -1;
    SELECT Ident FROM Station WHERE UrsiCode=:ursiCode INTO :ident;
    IF (ident = -1) THEN
        INSERT INTO Station (UrsiCode,Name,Lat,Lon) VALUES (UPPER(TRIM(:ursiCode)),TRIM(:name),:lat,:lon) RETURNING Ident INTO :ident;
    SUSPEND;
END!!

/*
 * Takes Transmitter station parameters,
 * Type of Tx station (can be either 'digisonde' or 'beacon'), code of Tx station (for Digisonde or Ionosonde, it is Ursi Code),
 * Type of beacon (which is NULL for digisonde, otherwise it is one of the following, 'radar' 'ionosonde', or 'other'), name, latitude, and longitude.
 * 
 * Then gets Ident from for Tx station (either from table Station or Beacon) and tries to find this Tx Station in the table TxStation,
 * 1. if Tx station is found, returns value of field Ident from corresponding row of table TxSatation
 * 2. if Tx station is not found, insert this Tx statiion into the table TxStation and returns value of field Ident 
 */
RECREATE PROCEDURE GetOrCreateTransmitterId(txTypeName TYPE OF COLUMN TxStationType.Name, code TYPE OF COLUMN Station.UrsiCode, 
    beaconTypeName TYPE OF COLUMN BeaconType.Name, name TYPE OF COLUMN Station.Name, lat TYPE OF COLUMN Station.Lat, lon TYPE OF COLUMN Station.Lon)
RETURNS (txStationId SMALLINT)
AS
DECLARE TYPE_STATION TYPE OF COLUMN TxStation.TypeId = 1;
DECLARE TYPE_BEACON TYPE OF COLUMN TxStation.TypeId = 2;
DECLARE txTypeId TYPE OF COLUMN TxStation.TypeId;
DECLARE stationId TYPE OF COLUMN Station.Ident;
DECLARE beaconTypeId TYPE OF COLUMN Beacon.TypeId;
DECLARE beaconId TYPE OF COLUMN Beacon.Ident;
BEGIN
    txTypeId = -1;
    SELECT Ident FROM TxStationType WHERE Name=:txTypeName INTO :txTypeId;
    IF (txTypeId = -1) THEN
        EXCEPTION Ex_GetTransitterId 'Illegal input parameter TxTypeName, ' || txTypeName;
    txStationId = -1;
    IF (txTypeId = TYPE_STATION) THEN
    BEGIN
	    SELECT Ident FROM GetOrCreateStationId(:code,:name,:lat,:lon) INTO :stationId;
        SELECT Ident FROM TxStation WHERE TypeId=:txTypeId AND StationId IS NOT NULL AND StationId=:stationId INTO :txStationId;
        IF (txStationId = -1) THEN
            INSERT INTO TxStation (TypeId,StationId) VALUES (:txTypeId,:stationId) RETURNING Ident INTO :txStationId;
    END 
    ELSE
    BEGIN
	    beaconTypeId = -1;
	    SELECT Ident FROM BeaconType WHERE Name=:beaconTypeName INTO :beaconTypeId;
	    IF (beaconTypeId = -1) THEN
            EXCEPTION Ex_GetTransitterId 'Illegal input parameter beaconTypeName, ' || beaconTypeName;
        beaconId = -1;
        SELECT Ident FROM Beacon WHERE TypeId=:beaconTypeId AND Code=:code INTO :beaconId;
        IF (beaconId = -1) THEN
            INSERT INTO Beacon (TypeId,Code,Name,Lat,Lon) VALUES (:beaconTypeId,TRIM(:code),TRIM(:name),:lat,:lon) RETURNING Ident INTO :beaconId;
        SELECT Ident FROM TxStation WHERE TypeId=:txTypeId AND BeaconId IS NOT NULL AND BeaconId=:beaconId INTO :txStationId;
        IF (txStationId = -1) THEN
            INSERT INTO TxStation (TypeId,BeaconId) VALUES (:txTypeId,:beaconId) RETURNING Ident INTO :txStationId;
    END 
    SUSPEND;
END!!

/*
 * Takes parameters for TID observation table, which are, Reference UT time, Rx station characteristics, and Tx station characteristics.
 * Then gets Idents for given Rx and Tx stations (this will include adding them if they are not in database) and tries to find this observation 
 * in the table Observation,
 * 1. If observation is found, returns value of field Ident from corresponding row of table Observation
 * 2. If observation is not found, insert this observation into the table Observation and resturns value of field Ident
 */
RECREATE PROCEDURE GetOrCreateObservation(refTimeUT TYPE OF COLUMN Observation.RefTimeUT, rxCode TYPE OF COLUMN Station.UrsiCode, 
    rxName TYPE OF COLUMN Station.Name, rxLat TYPE OF COLUMN Station.Lat, rxLon TYPE OF COLUMN Station.Lon,
    txTypeName TYPE OF COLUMN TxStationType.Name, txCode TYPE OF COLUMN Station.UrsiCode, beaconTypeName TYPE OF COLUMN BeaconType.Name, 
    txName TYPE OF COLUMN Station.Name, txLat TYPE OF COLUMN Station.Lat, txLon TYPE OF COLUMN Station.Lon)
RETURNS (observationId INT, rxStationId SMALLINT, txStationId SMALLINT)
AS
BEGIN
	SELECT Ident FROM GetOrCreateStationId(:rxCode,:rxName,:rxLat,:rxLon) INTO :rxStationId;
	SELECT TxStationId FROM GetOrCreateTransmitterId(:txTypeName,:txCode,:beaconTypeName,:txName,:txLat,:txLon) INTO :txStationId;
	observationId = -1;
	SELECT Ident FROM Observation WHERE RefTimeUT=:refTimeUT AND ReceiverId=:rxStationId INTO :observationId;
	IF (observationId = -1) THEN
	    INSERT INTO Observation (RefTimeUT,ReceiverId,TransmitterId) VALUES (:refTimeUT,:rxStationId,:txStationId) RETURNING Ident INTO :observationId;
	SUSPEND;
END!!

/*
 * This procedure is called from saveOptions procedure at the very beginning. It parses given string optionsAsStr and save them in 
 * global temporary table GTT-OptionSet.  
 * The format of the only parameter optionsAsStr is the following:
 * 
 * {options_group_name<<G>>{option_name<<N>>option_value<<V>>...}...}
 * 
 * where 
 *     options_group_name is a group name for subset of options, and <<G>> is the tag that signifies the end of options_group_name  
 *     option_name is the name of an option, and <<N>> is the tag that signifies the end of option_name  
 *     option_value is the value of the option which name is followed by this value, and <<V>> is the tag that signifies the end of option_value
 * 
 * Note that all options_group_name, and option_name are non-empty. Some or all option_value might be empty.   
 */
RECREATE PROCEDURE parseOptions(optionsAsStr VARCHAR(8190))
AS
DECLARE CR CHAR(1);
DECLARE LF CHAR(1);
DECLARE EQUAL CHAR(1) = '=';
DECLARE OPEN_BRACKET CHAR(1) = '[';
DECLARE CLOSE_BRACKET CHAR(1) = ']';
DECLARE MSG_START VARCHAR(32);
DECLARE strLength SMALLINT;
DECLARE groupName VARCHAR(120);
DECLARE optionName VARCHAR(50);
DECLARE optionValue VARCHAR(50);
DECLARE numberOfOptions SMALLINT = 0;
DECLARE currentPos SMALLINT = 1;
DECLARE startLinePos SMALLINT;
DECLARE line VARCHAR(300);
DECLARE equalPos SMALLINT;
DECLARE ch CHAR(1);
BEGIN
	CR = ASCII_CHAR(13);
	LF = ASCII_CHAR(10);
    MSG_START = 'proc parseOptions: ';
	optionsAsStr = TRIM(REPLACE(optionsAsStr, ASCII_CHAR(9), ' '));
    strLength = CHAR_LENGTH(optionsAsStr);
	/* Ideally, it has to be one call per transaction, so that instance of Global Temporary table would automatically GTT_OptionSet recreated
	 * here. But it might be some cases when it is several calls per transaction, hopefully not too many, so we delete rows here to anticipate
	 * those cases. 
	 */ 
    DELETE FROM GTT_OptionSet;
    /* Reads line by line and insert, when line "name=value" encountered, into GTT_OptionSet */
    startLinePos = currentPos;
    WHILE (startLinePos <= strLength)
    DO
    BEGIN
	    IF (currentPos <= strLength) THEN
	    	ch = SUBSTRING(optionsAsStr FROM currentPos FOR 1);
	    ELSE
		    ch = CR;
    	currentPos = currentPos + 1;
	    IF (ch = CR OR ch = LF) THEN
	    BEGIN
		    line = TRIM(REPLACE(SUBSTRING(optionsAsStr FROM startLinePos FOR currentPos - startLinePos - 1), ASCII_CHAR(9), ' '));
		    IF (line != '' AND SUBSTRING(line FROM 1 FOR 1) != '#' AND SUBSTRING(line FROM 1 FOR 1) != ';') THEN
		    BEGIN
			    ch  = SUBSTRING(line FROM 1 FOR 1);
				IF (ch = OPEN_BRACKET) THEN
				BEGIN
					IF (SUBSTRING(line FROM CHAR_LENGTH(line) FOR 1) != CLOSE_BRACKET) THEN
		                EXCEPTION Ex_ParseOptions MSG_START || 'line starts with [ but not ends with ]';
		            groupName = TRIM(REPLACE(SUBSTRING(line FROM 2 FOR CHAR_LENGTH(line) - 2), ASCII_CHAR(9), ' '));
		            IF (groupName = '') THEN
		                EXCEPTION Ex_ParseOptions MSG_START || 'empty group';
				END
				ELSE
				BEGIN
					IF (groupName IS NULL) THEN
			            EXCEPTION Ex_ParseOptions MSG_START || 'no group specified';
				    equalPos = POSITION(EQUAL, line, 1);
				    IF (equalPos = 0) THEN
			            EXCEPTION Ex_ParseOptions MSG_START || 'no equal sign';
					optionName = TRIM(REPLACE(SUBSTRING(line FROM 1 FOR equalPos - 1), ASCII_CHAR(9), ' '));
					IF (optionName = '') THEN
			            EXCEPTION Ex_ParseOptions MSG_START || 'Option name is empty';
					optionValue = TRIM(REPLACE(SUBSTRING(line FROM equalPos + 1 FOR CHAR_LENGTH(LINE) - equalPos), ASCII_CHAR(9), ' '));
			        INSERT INTO GTT_OPTIONSET (GName, Name, Val) VALUES (:groupName, :optionName, :optionValue);
			        numberOfOptions = numberOfOptions + 1;
			    END
		    END
		    startLinePos = currentPos;
		END
    END
    IF (numberOfOptions = 0) THEN
        EXCEPTION Ex_ParseOptions MSG_START || 'No options specified';
END!!

/*
 * Calculates the hash code of Map<String, Map<String,String>> in the same manner as in Java
 */
RECREATE PROCEDURE calcOptionsSetHashCode
   RETURNS (hashCode INTEGER)
AS
DECLARE groupName VARCHAR(120);
DECLARE optionName VARCHAR(50);
DECLARE optionValue VARCHAR(50);
DECLARE longHashCode BIGINT = 0;
DECLARE groupNameHashCode BIGINT = 0;
DECLARE groupValueHashCode BIGINT = 0;
DECLARE prevGroupName VARCHAR(120) = '';
DECLARE optionNameHashCode INTEGER;
DECLARE optionValueHashCode INTEGER;
BEGIN
	FOR SELECT GName, Name, Val FROM GTT_OptionSet ORDER BY GName INTO :groupName, :optionName, :optionValue
	DO
	BEGIN
		IF (groupName != prevGroupName) THEN
		BEGIN
		    longHashCode = longHashCode + BIN_XOR(groupNameHashCode, groupValueHashCode);
            longHashCode = BIN_SHR(BIN_SHL(longHashCode, 32), 32);
            SELECT hash FROM HashCode(:groupName) INTO :groupNameHashCode;
            prevGroupName = groupName;
            groupValueHashCode = 0;
        END
        SELECT hash FROM HashCode(:optionName) INTO :optionNameHashCode;
        SELECT hash FROM HashCode(:optionValue) INTO :optionValueHashCode;
        groupValueHashCode = groupValueHashCode + BIN_XOR(optionNameHashCode, optionValueHashCode);
        groupValueHashCode = BIN_SHR(BIN_SHL(groupValueHashCode, 32), 32);
	END 
    longHashCode = longHashCode + BIN_XOR(groupNameHashCode, groupValueHashCode);
    hashCode = BIN_SHR(BIN_SHL(longHashCode, 32), 32);
	SUSPEND;
END!!

/*
 * This procedure saves options in database and returns its Option Set internal identifier, which is the field OptionSetId of table OptionSet.
 */
RECREATE PROCEDURE saveOptions(optionsAsStr VARCHAR(8190))
    RETURNS (optionSetId INTEGER)
AS
DECLARE MSG_START VARCHAR(32);
DECLARE hashCode INTEGER;
DECLARE numberOfOptions INTEGER;
DECLARE groupName VARCHAR(120);
DECLARE optionName VARCHAR(50);
DECLARE optionValue VARCHAR(50);
DECLARE setId INTEGER;
DECLARE testedGroupName VARCHAR(120);
DECLARE testedOptionName VARCHAR(50);
DECLARE testedOptionValue VARCHAR(50);
DECLARE prevSetId INTEGER = -1;
DECLARE cur CURSOR FOR (SELECT GName, Name, Val FROM GTT_OptionSet ORDER BY GName, Name);
DECLARE cnt INTEGER = 0;
DECLARE skip SMALLINT = 0;
DECLARE prevGroupName VARCHAR(120) = '';
DECLARE groupId INTEGER;
BEGIN
    MSG_START = 'proc saveOptions: ';
    EXECUTE PROCEDURE parseOptions(:optionsAsStr);
    SELECT hashCode FROM calcOptionsSetHashCode INTO :hashCode;
    SELECT COUNT(*) FROM GTT_OptionSet INTO :numberOfOptions;
    optionSetId = -1;
	OPEN cur;
    FOR SELECT os.OptionSetId, og.Name, os.Name, os.Val FROM OptionSet os JOIN OptionsGroup og ON os.OptionsGroupId=og.Ident
          JOIN OptionSetHeader oh ON oh.Ident=os.OptionSetId
        WHERE oh.SetSize=:numberOfOptions AND oh.HashCode=:hashCode
        ORDER BY os.OptionSetId, og.Name, os.Name
        INTO :setId, :testedGroupName, :testedOptionName, :testedOptionValue
    DO
    BEGIN
	    IF (setId != prevSetId) THEN
	    BEGIN
		    IF (prevSetId != -1) THEN
		    BEGIN
			    IF (cnt < numberOfOptions) THEN
			        EXCEPTION Ex_SaveOptions MSG_START || 'cnt < numberOfOptions';
			    IF (skip = 0) THEN
			    BEGIN
				    optionSetId = prevSetId;
				    LEAVE;
			    END 
			END
		    CLOSE cur;
			OPEN cur;
			prevSetId = setId;
			cnt = 0;
			skip = 0;
	    END
	    cnt = cnt + 1;
	    IF (cnt > numberOfOptions) THEN
	        EXCEPTION Ex_SaveOptions MSG_START || 'cnt > numberOfOptions';
	    IF (skip = 0) THEN
	    BEGIN
	        FETCH cur INTO groupName, optionName, optionValue;
	        IF (groupName != testedGroupName OR optionName != testedOptionName OR optionValue != testedOptionValue) THEN
	            skip = 1;
	    END 
    END
    CLOSE cur;
    IF (optionSetId = -1) THEN
    BEGIN
	    IF (cnt != 0 AND cnt != numberOfOptions) THEN
	        EXCEPTION Ex_SaveOptions MSG_START || 'cnt != 0 AND cnt != numberOfOptions';
	    IF (skip != 0 OR cnt = 0) THEN
	    BEGIN
		    INSERT INTO OptionSetHeader (HashCode, SetSize) VALUES (:hashCode, :numberOfOptions) RETURNING Ident INTO :optionSetId;
		    OPEN cur;
		    WHILE (1=1)
		    DO
		    BEGIN
	            FETCH cur INTO groupName, optionName, optionValue;
	            IF (ROW_COUNT = 0) THEN
	                LEAVE;
	            IF (groupName != prevGroupName) THEN
	            BEGIN
		            groupId = -1;
		            SELECT Ident FROM OptionsGroup WHERE Name=:groupName INTO :groupId;
		            IF (groupId = -1) THEN
		                INSERT INTO OptionsGroup (Name) VALUES (:groupName) RETURNING Ident INTO :groupId;
		            prevGroupName = groupName;
	            END 
	            INSERT INTO OptionSet (OptionSetId, OptionsGroupId, Name, Val) VALUES (:optionSetId, :groupId, :optionName, :optionValue);
		    END
		    CLOSE cur;
	    END 
	    ELSE
	      optionSetId = setId;
    END 
    SUSPEND;
END!!

/*
 * Adds track signals to the table Signal.
 * Each Track signal contains about 10 characteristics and stored as one row in table Signal.
 * All track signals are given here as one big string where track signals divided by :: (two colons).
 * Individual chacracteristics of track signal are seperated with , (comma). 
 */
RECREATE PROCEDURE AddSignals(trackId TYPE OF COLUMN TrackHeader.Ident, trackLength TYPE OF COLUMN TrackHeader.TrackLength, signalsAsStr VARCHAR(7000))
AS
DECLARE SIGNAL_DELIM CHAR(2) = '::';
DECLARE DELIM_LENGTH SMALLINT;
DECLARE strLength SMALLINT;
DECLARE currentPos SMALLINT = 1;
DECLARE nextPos SMALLINT;
DECLARE seqNumber TYPE OF COLUMN Signal.SeqNumber = 0; 
DECLARE signalAsList VARCHAR(200);
BEGIN
	DELIM_LENGTH = CHAR_LENGTH(SIGNAL_DELIM);
	signalsAsStr = TRIM(REPLACE(REPLACE(REPLACE(signalsAsStr, ASCII_CHAR(9), ' '), ASCII_CHAR(10), ' '), ASCII_CHAR(13), ' '));
	strLength = CHAR_LENGTH(signalsAsStr);
    /* Reads row by row from input string signalsAsStr and insert each row into table Signal */
    WHILE (currentPos <= strLength)
    DO
    BEGIN
	    nextPos = POSITION(SIGNAL_DELIM, signalsAsStr, currentPos);
	    IF (nextPos = 0) THEN
	        nextPos = strLength + 1;
	    signalAsList = TRIM(SUBSTRING(signalsAsStr FROM currentPos FOR nextPos - currentPos));
	    IF (CHAR_LENGTH(signalAsList) > 0) THEN
	    BEGIN
	        seqNumber = seqNumber + 1;
	        EXECUTE STATEMENT 'INSERT INTO Signal (TrackId,SeqNumber,Path_km,DoppShift_Hz,Elev_deg,Azim_deg,FitErr_deg,SNR_dB,Amp_dB,IsInterpolated) VALUES (' ||
	            trackId || ',' || seqNumber || ',' || signalAsList || ')';
	    END
	    currentPos = nextPos + DELIM_LENGTH;
    END
    IF (seqNumber = 0) THEN
        EXCEPTION Ex_AddTrackSignals 'empty signalsAsStr string';
    IF (seqNumber != trackLength) THEN
        EXCEPTION Ex_AddTrackSignals 'TrackLength is ' || trackLength || ' but number of tracks in signalsAsStr is ' || seqNumber;
END!!

/*
 * Adds one track for given TID ProcId, polarization (ordinary=1, extraordinary=1), and freq_kHz. If given polarization and freq_kHz already 
 * exist then it is supposed to be another track for this polarization and frequency.
 */
RECREATE PROCEDURE AddOneTrackProcessing(obsProcOptId TYPE OF COLUMN ObsProcOpt.Ident,
    polarization Polarization, freq_kHz TYPE OF COLUMN PolFreq.Freq_kHz, startUT TYPE OF COLUMN TrackHeader.StartUT, 
    cadence_s TYPE OF COLUMN TrackHeader.Cadence_s, trackLength TYPE OF COLUMN TrackHeader.TrackLength, numGaps TYPE OF COLUMN TrackHeader.NumGaps, 
    medianPath_km TYPE OF COLUMN TrackHeader.MedianPath_km, avgFitErr_deg TYPE OF COLUMN TrackHeader.AvgFitErr_deg, 
    avgSNR_dB TYPE OF COLUMN TrackHeader.AvgSNR_dB, avgAmp_dB TYPE OF COLUMN TrackHeader.AvgAmp_dB,
    refLat_deg TYPE OF COLUMN TIDChars.RefLat_deg, refLon_deg TYPE OF COLUMN TIDChars.RefLon_deg, refAlt_km TYPE OF COLUMN TIDChars.RefAlt_km,
    wavePeriod_s TYPE OF COLUMN TIDChars.WavePeriod_s, waveLength_m TYPE OF COLUMN TIDChars.WaveLength_m,
    amplitude_fact TYPE OF COLUMN TIDChars.Amplitude_fact, waveDir_deg TYPE OF COLUMN TIDChars.WaveDir_deg,
    initPhase_deg TYPE OF COLUMN TIDChars.InitPhase_deg, confidence TYPE OF COLUMN TIDChars.Confidence,
    uncertainty TYPE OF COLUMN TIDChars.Uncertainty, signalsAsStr VARCHAR(8190))
AS
DECLARE polFreqId TYPE OF COLUMN PolFreq.Ident;
DECLARE trackId TYPE OF COLUMN TrackHeader.Ident;
DECLARE tidCharsId TYPE OF COLUMN TIDChars.Ident;
BEGIN
	polFreqId = -1;
	SELECT Ident FROM PolFreq WHERE ObsProcOptId=:obsProcOptId AND Polarization=:polarization AND Freq_kHz=:freq_kHz INTO :polFreqId;
	IF (polFreqId = -1) THEN
	    INSERT INTO PolFreq (ObsProcOptId,Polarization,Freq_kHz) VALUES (:obsProcOptId,:polarization,:freq_kHz) RETURNING Ident INTO :polFreqId;
	INSERT INTO TrackHeader (PolFreqId,StartUT,Cadence_s,TrackLength,NumGaps,MedianPath_km,AvgFitErr_deg,AvgSNR_dB,AvgAmp_dB) VALUES 
	                        (:polFreqId,:startUT,:cadence_s,:trackLength,:numGaps,:medianPath_km,:avgFitErr_deg,:avgSNR_dB,:avgAmp_dB)
	    RETURNING Ident INTO :trackId;
	INSERT INTO TIDChars (TrackId,RefLat_deg,RefLon_deg,RefAlt_km,WavePeriod_s,WaveLength_m,Amplitude_fact,WaveDir_deg,InitPhase_deg,Confidence,Uncertainty)
	    VALUES (:trackId,:refLat_deg,:refLon_deg,:refAlt_km,:wavePeriod_s,:waveLength_m,:amplitude_fact,:waveDir_deg,:initPhase_deg,:confidence,:uncertainty)
	    RETURNING Ident INTO :tidCharsId;
	UPDATE TrackHeader SET TIDCharsId=:tidCharsId WHERE Ident=:trackId;
	EXECUTE PROCEDURE AddSignals(trackId, trackLength, signalsAsStr);
END!!

/*
 * Adds or update (if ecists) TID Processing.
 * Note that TID Observation will be created if needed, as well as Rx or/and Tx station/beacon.
 * If TID Processing consists only of one track then this procedure fullfill everything you need.
 * IF TID Processing consists of several tacks, then you have to add others one by one using stored procedure AddOneTrackProcessing and
 * value of TID ProcId which is returned by this procedure. 
 */
RECREATE PROCEDURE AddOrUpdateProcessing(refTimeUT TYPE OF COLUMN Observation.RefTimeUT, rxCode TYPE OF COLUMN Station.UrsiCode, 
    rxName TYPE OF COLUMN Station.Name, rxLat TYPE OF COLUMN Station.Lat, rxLon TYPE OF COLUMN Station.Lon,
    txTypeName TYPE OF COLUMN TxStationType.Name, txCode TYPE OF COLUMN Station.UrsiCode, beaconTypeName TYPE OF COLUMN BeaconType.Name, 
    txName TYPE OF COLUMN Station.Name, txLat TYPE OF COLUMN Station.Lat, txLon TYPE OF COLUMN Station.Lon,
    procName TYPE OF COLUMN Processor.Name, procVersion TYPE OF COLUMN Processor.Version, compareOptions SMALLINT, options VARCHAR(8190),
    submissionUT TYPE OF COLUMN ObsProcOpt.SubmissionUT,
    eLayerGroupPath_km TYPE OF COLUMN ObsProcOpt.ELayerGroupPath_km, eLayerGPathFactor TYPE OF COLUMN ObsProcOpt.ELayerGPathFactor,
    polarization Polarization, freq_kHz TYPE OF COLUMN PolFreq.Freq_kHz, startUT TYPE OF COLUMN TrackHeader.StartUT, 
    cadence_s TYPE OF COLUMN TrackHeader.Cadence_s, trackLength TYPE OF COLUMN TrackHeader.TrackLength, numGaps TYPE OF COLUMN TrackHeader.NumGaps, 
    medianPath_km TYPE OF COLUMN TrackHeader.MedianPath_km, avgFitErr_deg TYPE OF COLUMN TrackHeader.AvgFitErr_deg, 
    avgSNR_dB TYPE OF COLUMN TrackHeader.AvgSNR_dB, avgAmp_dB TYPE OF COLUMN TrackHeader.AvgAmp_dB,
    refLat_deg TYPE OF COLUMN TIDChars.RefLat_deg, refLon_deg TYPE OF COLUMN TIDChars.RefLon_deg, refAlt_km TYPE OF COLUMN TIDChars.RefAlt_km,
    wavePeriod_s TYPE OF COLUMN TIDChars.WavePeriod_s, waveLength_m TYPE OF COLUMN TIDChars.WaveLength_m,
    amplitude_fact TYPE OF COLUMN TIDChars.Amplitude_fact, waveDir_deg TYPE OF COLUMN TIDChars.WaveDir_deg,
    initPhase_deg TYPE OF COLUMN TIDChars.InitPhase_deg, confidence TYPE OF COLUMN TIDChars.Confidence,
    uncertainty TYPE OF COLUMN TIDChars.Uncertainty, trackAsStr VARCHAR(7500))
RETURNS (obsProcOptId INTEGER, updated SMALLINT)
AS
DECLARE procId TYPE OF COLUMN Processor.Ident;
DECLARE observId TYPE OF COLUMN Observation.Ident;
DECLARE optionSetId TYPE OF COLUMN OptionSetHeader.Ident;
DECLARE tidCharsId TYPE OF COLUMN TIDChars.Ident;
BEGIN
	procId = -1;
	procName = TRIM(procName);
	procVersion = TRIM(procVersion);
	SELECT Ident FROM Processor WHERE Name=:procName AND Version=:procVersion INTO :procId;
	IF (procId = -1) THEN
	    EXCEPTION Ex_AddOrUpdateProcessing 'unknown processor name, ' || procName || ', and/or version, ' || procVersion;
	SELECT ObservationId FROM GetOrCreateObservation(:refTimeUT, :rxCode, :rxName, :rxLat, :rxLon, :txTypeName, :txCode, :beaconTypeName, :txName, :txLat, :txLon)
	    INTO :observId;
	SELECT OptionSetId FROM SaveOptions(:options) INTO :optionSetId;
	obsProcOptId = -1;
	updated = 0;
	IF (compareOptions = 0) THEN
    	DELETE FROM ObsProcOpt WHERE ObservId=:observId AND ProcId=:procId;
	ELSE
		DELETE FROM ObsProcOpt WHERE ObservId=:observId AND ProcId=:procId AND OptionSetId=:optionSetId;
   	updated = ROW_COUNT;
	INSERT INTO ObsProcOpt (ObservId,ProcId,OptionSetId,SubmissionUT,ELayerGroupPath_km,ELayerGPathFactor) 
	    VALUES (:observId,:procId,:optionSetId,:submissionUT,:eLayerGroupPath_km,:eLayerGPathFactor) RETURNING Ident INTO :obsProcOptId;
	EXECUTE PROCEDURE AddOneTrackProcessing(obsProcOptId, polarization, freq_kHz, startUT, cadence_s, trackLength, numGaps, medianPath_km, 
	    avgFitErr_deg, avgSNR_dB, avgAmp_dB, refLat_deg, refLon_deg, 
	    refAlt_km, wavePeriod_s, waveLength_m, amplitude_fact, waveDir_deg, initPhase_deg, confidence, uncertainty, trackAsStr);
	SUSPEND;
END!!

SET TERM ; !!

/* Create various generators */

/* for sequence unique Ident value for new record in Station table */
CREATE SEQUENCE Station_Ident_Sequence;
/* for sequence unique Ident value for new record in Beacon table */
CREATE SEQUENCE Beacon_Ident_Sequence;
/* for sequence unique Ident value for new record in TxStation table */
CREATE SEQUENCE TxStation_Ident_Sequence;
/* for sequence unique Ident value for new record in OptionsGroup table */
CREATE SEQUENCE OptionsGroup_Ident_Sequence;
/* for sequence unique Ident value for new record in OptionSetHeader table */
CREATE SEQUENCE OptionSetHeader_Ident_Sequence;
/* for sequence unique Ident value for new record in Observation table */
CREATE SEQUENCE Observation_Ident_Sequence;
/* for sequence unique Ident value for new record in ObsProcOpt table */
CREATE SEQUENCE ObsProcOpt_Ident_Sequence;
/* for sequence unique Ident value for new record in PolFreq table */
CREATE SEQUENCE PolFreq_Ident_Sequence;
/* for sequence unique Ident value for new record in TrackHeader table */
CREATE SEQUENCE TrackHeader_Ident_Sequence;
/* for sequence unique Ident value for new record in TIDChars table */
CREATE SEQUENCE TIDChars_Ident_Sequence;
/* for sequence unique Ident value for new record in UserReg table */
CREATE SEQUENCE UserReg_Ident_Sequence;

/* Create indexes */

/* for table SetHashCode */
CREATE ASCENDING  INDEX OptionSetHeader_HashCode ON OptionSetHeader (SetSize,HashCode);

/* Create various triggers and procedures */

SET TERM !! ;

/* Before-Insert trigger for Station table */
CREATE TRIGGER Tr_Before_Ins_Station FOR Station BEFORE INSERT POSITION 0
	AS BEGIN
		NEW.Ident = NEXT VALUE FOR Station_Ident_Sequence;
	END!!

/* Insert trigger for Beacon table to automatically generate unique Ident value */
CREATE TRIGGER Tr_Before_Ins_Beacon FOR Beacon
    BEFORE INSERT POSITION 0
    AS BEGIN
        NEW.IDENT = NEXT VALUE FOR Beacon_Ident_Sequence; 
    END!!

/* Insert trigger for TxStation table to automatically generate unique Ident value */
CREATE TRIGGER Tr_Before_Ins_TxStation FOR TxStation
    BEFORE INSERT POSITION 0
    AS BEGIN
        NEW.IDENT = NEXT VALUE FOR  TxStation_Ident_Sequence; 
    END!!

/* Before-insert trigger for OptionsGroup table */
CREATE TRIGGER Tr_Before_Ins_OptionsGroup FOR OptionsGroup BEFORE INSERT POSITION 0
	AS BEGIN
		NEW.Ident = NEXT VALUE FOR OptionsGroup_Ident_Sequence;
	END!!

/* Before-insert trigger for OptionSetHeader table */
CREATE TRIGGER Tr_Before_Ins_OptionSetHeader FOR OptionSetHeader BEFORE INSERT POSITION 0
	AS BEGIN
		NEW.Ident = NEXT VALUE FOR OptionSetHeader_Ident_Sequence;
	END!!

/* Insert trigger for Observation table to automatically generate unique Ident value */
CREATE TRIGGER Tr_Before_Ins_Observation FOR Observation
    BEFORE INSERT POSITION 0
    AS BEGIN
        NEW.IDENT = NEXT VALUE FOR Observation_Ident_Sequence; 
    END!!

/* Insert trigger for ObsProcOpt table to automatically generate unique Ident value */
CREATE TRIGGER Tr_Before_Ins_ObsProcOpt FOR ObsProcOpt
    BEFORE INSERT POSITION 0
    AS BEGIN
        NEW.IDENT = NEXT VALUE FOR ObsProcOpt_Ident_Sequence; 
    END!!

/* Insert trigger for PolFreq table to automatically generate unique Ident value */
CREATE TRIGGER Tr_Before_Ins_PolFreq FOR PolFreq
    BEFORE INSERT POSITION 0
    AS BEGIN
        NEW.IDENT = NEXT VALUE FOR PolFreq_Ident_Sequence; 
    END!!

/* Insert trigger for TrackHeader table to automatically generate unique Ident value */
CREATE TRIGGER Tr_Before_Ins_TrackHeader FOR TrackHeader
    BEFORE INSERT POSITION 0
    AS BEGIN
        NEW.IDENT = NEXT VALUE FOR TrackHeader_Ident_Sequence; 
    END!!

/* Insert trigger for TIDChars table to automatically generate unique Ident value */
CREATE TRIGGER Tr_Before_Ins_TIDChars FOR TIDChars
    BEFORE INSERT POSITION 0
    AS BEGIN
        NEW.IDENT = NEXT VALUE FOR TIDChars_Ident_Sequence; 
    END!!

/* Insert trigger for USERREG table to automatically generate unique Ident value */
CREATE TRIGGER Create_UserReg_Ident FOR UserReg
    BEFORE INSERT POSITION 0
    AS BEGIN
        NEW.IDENT = NEXT VALUE FOR UserReg_Ident_Sequence; 
    END!!

SET TERM ; !!

/* Create Roles */
CREATE ROLE RealTimer;
CREATE ROLE Expert;
CREATE ROLE Common;  /* regular user role */

/* Create Grants */

/* !!!don't delete this grant - it's for connection availability checking!!! */
GRANT SELECT, INSERT, UPDATE ON USERREG TO UCONNECT;

GRANT SELECT                           ON vObs TO RealTimer, Expert, Common;
/*
GRANT SELECT, REFERENCES(Ident)        ON Observation TO RealTimer, Expert, Common;
GRANT SELECT, REFERENCES(Ident)        ON ObsProcOpt TO RealTimer, Expert, Common;
GRANT SELECT, REFERENCES(Ident)        ON PolFreq TO RealTimer, Expert, Common;
GRANT SELECT, REFERENCES(Ident)        ON TrackHeader TO RealTimer, Expert, Common;
GRANT SELECT, REFERENCES(Ident)        ON TIDChars TO RealTimer, Expert, Common;
*/
GRANT SELECT, REFERENCES(Ident)        ON OptionsGroup TO RealTimer, Expert, Common;
GRANT SELECT, REFERENCES(Ident)        ON OptionSetHeader TO RealTimer, Expert, Common;
GRANT SELECT, REFERENCES(Ident)        ON Station TO RealTimer, Expert, Common;
GRANT SELECT, REFERENCES(Ident)        ON TxStation TO RealTimer, Expert, Common;
GRANT SELECT, REFERENCES(Ident)        ON Beacon TO RealTimer, Expert, Common;
GRANT SELECT                           ON Signal TO RealTimer, Expert, Common;
GRANT SELECT, REFERENCES(OptionSetId)  ON OptionSet TO RealTimer, Expert, Common;
GRANT INSERT, UPDATE, DELETE           ON Observation TO RealTimer, Expert;
GRANT INSERT, UPDATE, DELETE           ON ObsProcOpt TO RealTimer, Expert;
GRANT INSERT, UPDATE, DELETE           ON PolFreq TO RealTimer, Expert;
GRANT INSERT, UPDATE, DELETE           ON TrackHeader TO RealTimer, Expert;
GRANT INSERT, UPDATE, DELETE           ON TIDChars TO RealTimer, Expert;
GRANT INSERT, UPDATE, DELETE           ON Signal TO RealTimer, Expert;
GRANT INSERT, UPDATE, DELETE           ON Beacon TO RealTimer, Expert;
GRANT INSERT, UPDATE, DELETE           ON TxStation TO RealTimer, Expert;
GRANT INSERT, UPDATE, DELETE           ON OptionsGroup TO RealTimer, Expert;
GRANT INSERT, UPDATE, DELETE           ON OptionSetHeader TO RealTimer, Expert;
GRANT INSERT, UPDATE, DELETE           ON Station    TO RealTimer, Expert;
GRANT INSERT, UPDATE, DELETE           ON OptionSet  TO RealTimer, Expert;

/* CREATE USER TidRealTimer PASSWORD 'password' FIRSTNAME 'TID' LASTNAME 'Real Timer'; */
GRANT RealTimer TO TidRealTimer;
/* CREATE USER TidExplorer PASSWORD 'password' FIRSTNAME 'TID' LASTNAME 'Explorer'; */
GRANT COMMON TO TIDExplorer;

COMMIT;
